/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.jsxp.framework.dom.ViewParserDom;
import org.jsxp.framework.fileupload.adapter.FileUpload;
import org.jsxp.framework.logging.Log;
import org.jsxp.framework.resources.ResourceResolver;
import org.jsxp.framework.util.adapter.AdapterFinder;
import org.jsxp.framework.util.text.StringUtil;
import org.jsxp.framework.util.text.StringUtils;
import org.jsxp.framework.webroot.jsxpframework.components.ErrorXhtmlController;

/**
 * The Class Application. Default implementation of an application. It is recommend
 * to create a subclass for your webapps.
 * 
 * The application is automatically instantiated by the jsxp framework. You can access it with
 * {@link Application#getApplication()}. Do not create an instance yourself.
 */
public class Application implements ResourceResolver, ViewControllerLifeCycleExceptionHandler {
	private static Application app = new Application();
	protected static final String DIRECT_DELIVERY_ALLOWED_FILE_REGULAR_EXPRESSIONS = "org.jsxp.framework.directDeliveryAllowedFileRegExpressions";
	protected static final String DIRECT_DELIVERY_DENIED_FILE_REGULAR_EXPRESSIONS = "org.jsxp.framework.directDeliveryDeniedFileRegExpressions";
	protected static final String FILE_UPLOAD_ADAPTER = "org.jsxp.framework.fileUploadAdapter";
	private static final String GENERATED_SUFFIX = "Generated";
	protected static final String PROJECT_STAGE_NAME = "org.jsxp.framework.projectStage";

	protected static final String SUPPORTED_LOCALES = "org.jsxp.framework.SupportedLocales";
	protected static final String VIEW_CONTROL_BASE_PACKAGE = "org.jsxp.framework.ViewControllerBasePackage";

	private static final String VIEW_CONTROL_SUFFIX = "Controller";
	/**
	 * Gets the application.
	 * 
	 * @return the application
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Application> T getApplication() {
		return (T) app;
	}
	/**
	 * Sets the application. Used internally by the jsxp framework.
	 * 
	 * @param application the new application
	 */
	static void setApplication(final Application application) {
		app = application;
	}
	private String[] basePackages = null;
	private LinkedList<String> directDeliveryAllowedFileRegExpressions = null;

	private LinkedList<String> directDeliveryDeniedFileRegExpressions = null;
	private List<Locale> locales = null;

	private ProjectStage projectStage = null;

	private final ViewParser viewParser = new ViewParserDom();

	/**
	 * Instantiates a new application.
	 */
	protected Application() {
	}

	/**
	 * Called by the jsxp framework if the Context was initialized. Override this method
	 * to do anything with the context.
	 * 
	 * @param context the context
	 */
	protected void contextInitialized(final Context context) { }

	/**
	 * Creates the context. Override to create instances from special Context classes.
	 * 
	 * @return the context
	 */
	protected Context createContext() {
		return new Context();
	}

	/**
	 * Create a project stage for a given project stage name.
	 * Override this method to define which project stages your application knows and how they are configured.
	 * 
	 * @param name the name of the project stage.
	 * 
	 * @return the project stage instance
	 */
	protected ProjectStage createProjectStage(final String name) {
		final ProjectStage projectStage = new ProjectStage(name);

		if (name.equals("production")) {
			projectStage.setUseDebugMessagesViewController(false);
			projectStage.setUseDeveloperWarningsViewController(false);
			projectStage.setUseScopeMessagesViewController(false);
		} else if (name.equals("development")) {
			projectStage.setUseDebugMessagesViewController(true);
			projectStage.setUseDeveloperWarningsViewController(true);
			projectStage.setUseScopeMessagesViewController(true);
		}

		return projectStage;
	}

	/**
	 * Creates the user context. Override to create instances from special UserContext classes.
	 * 
	 * @return the user context
	 */
	protected UserContext createUserContext() {
		return new UserContext();
	}

	/**
	 * Should set the locale that should be used for this request to the context
	 * and return an uri that may be delocalized (e.g. index.html instead of en/index.html
	 * in the default implementation) regarding to the way localization should be done
	 * (if it is done via the uri). Override to change translation behaviour.
	 * @param uri the incoming uri from the client
	 * @return the uri that is without locale information
	 */
	public String delocalizeUriAndInitContextLocale(String uri) {
		final List<Locale> supportedLocales = Application.getApplication().getSupportedApplicationLocales();
		for(final Locale l : supportedLocales) {
			String rootPath = uri;
			int slashIndex = rootPath.lastIndexOf("/");
			slashIndex = rootPath.lastIndexOf("/", slashIndex-1);
			if (slashIndex > -1) {
				rootPath = rootPath.substring(0, slashIndex);
			}
			final String lastUriPart = uri.substring(slashIndex+1);

			if(lastUriPart.startsWith(l.toString()+"/")) {
				Context.getContext().setRequestLocale(l);
				uri = StringUtils.combine("/", true, rootPath,lastUriPart.substring(lastUriPart.indexOf("/")));
				return uri;
			}
		}
		return uri;
	}

	/**
	 * Gets the default application locale.
	 * 
	 * @return the default application locale
	 */
	public Locale getDefaultApplicationLocale() {
		return Locale.getDefault();
	}

	/**
	 * Gets default the character encoding. Override this method if you want to change the type
	 * for all the views in your application. Or override {@link ViewController#getCharacterEncoding()} to
	 * set it for one view only. This will give your view a special character
	 * encoding. Default is UTF-8.
	 * 
	 * @return the character encoding
	 */
	public String getDefaultCharacterEncoding() {
		return "UTF-8";
	}

	/**
	 * Gets the default content type of the rendered output. Override it to change the type for all the views in your application.
	 * Default is "text/html". Can also be overridden in a view controller for that special
	 * view controller.
	 * 
	 * @return the content type
	 */
	public String getDefaultContentType() {
		return "text/html";
	}

	/**
	 * Gets the direct delivery allowed list. Override this method to allow direct delivery for
	 * special file types in your webapp. Per default it allows the direct delivery of stylesheet, images, ...
	 * (".*\\.css", ".*\\.js",".*\\.jpg", ".*\\.jpeg",".*\\.gif", ".*\\.png");
	 * The allow list is a list of regular expression strings that can be matched to urls.
	 * This is used by the default implementation
	 * of the {@link #isDirectDeliveryAllowed(String)}.
	 * 
	 * @return the direct delivery allowed list (null in default application)
	 * @see #isDirectDeliveryAllowed(String)
	 */
	protected List<String> getDirectDeliveryAllowedList() {
		return directDeliveryAllowedFileRegExpressions;
	}

	/**
	 * Gets the direct delivery denied list. Like the direct delivery allow list it can
	 * be used to disable the direct delivery of requested urls. This is used by the default implementation
	 * of the {@link #isDirectDeliveryAllowed(String)}.
	 * 
	 * The deny list is a list of regular expression strings that can be matched to urls.
	 * 
	 * @return the direct delivery denied list (null in default application)
	 * @see #isDirectDeliveryAllowed(String)
	 */
	protected List<String> getDirectDeliveryDeniedList() {
		return directDeliveryDeniedFileRegExpressions;
	}

	/**
	 * Gets the parameter. Override to use application parameter (key/value pairs) for your
	 * application. Per default the parameters from web.xml (the servlet config parameters) are used.
	 * 
	 * @param key the key
	 * @param defaultValue the default value
	 * 
	 * @return the parameter
	 */
	public String getParameter(final String key, final String defaultValue) {
		String value = null;

		try {
			value = Context.getContext().getServletConfig().getInitParameter(key);
		}
		catch (final Throwable ex)
		{
			// sometimes the case, if used via the ViewControllerGenerator
			// and the servlet library is not in the classpath,
			// or servlet values are not set
			// so we can just ignore it
		}

		if (value == null) {
			return defaultValue;
		} else {
			return value;
		}
	}

	/**
	 * Gets the current project stage. Override this method to define in which project stage your application is.
	 * 
	 * Default implementation uses {@link #getParameter(String, String)} with
	 * key {@link #PROJECT_STAGE_NAME} and the "development" string as defaultValue
	 * to get the project stage name and will than use the method {@link #createProjectStage(String)} to
	 * get the project stage when initialized.
	 * 
	 * @return the current project stage
	 */
	public ProjectStage getProjectStage() {
		return projectStage;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.ResourceResolver#getResourceInputStream(java.lang.String, java.lang.Object)
	 */
	@Override
	public InputStream getResourceInputStream(final String uri, final Object objectForClassLoader) {
		InputStream is = getResourceInputStream(uri, objectForClassLoader, true);
		if (is == null) {
			is = getResourceInputStream(uri, objectForClassLoader, false);
			if (is == null) {
				is = getResourceInputStreamFile(uri, objectForClassLoader, true);
				if (is == null) {
					is = getResourceInputStreamFile(uri, objectForClassLoader, false);
				}
			}
		}
		return is;
	}

	private InputStream getResourceInputStream(String uri, final Object objectForClassLoader, final boolean translateUri) {
		if (translateUri) {
			uri = localizeUri(uri);
		}

		InputStream is = null;
		try {
			is = Context.getContext().getServletContext().getResourceAsStream(uri);
		} catch (final Throwable t) {
		}
		return is;
	}

	private InputStream getResourceInputStreamFile(String uri, final Object objectForClassLoader, final boolean translateUri) {
		if (translateUri) {
			uri = localizeUri(uri);
		}

		final File f = new File(uri);
		if (f != null && f.exists() && f.isFile()) {
			InputStream is;
			try {
				is = new FileInputStream(f);
			} catch (final FileNotFoundException e) {
				is = null;
			}
			if (is != null) {
				return is;
			}
		}

		return objectForClassLoader.getClass().getClassLoader().getResourceAsStream(uri);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.ResourceResolver#getResourceModifiedTimeStamp(java.lang.String, java.lang.Object)
	 */
	@Override
	public long getResourceModifiedTimeStamp(final String uri, final Object objectForClassLoader) {
		long modifiedTimeStamp = getResourceModifiedTimeStamp(uri, objectForClassLoader, true);
		if (modifiedTimeStamp == 0) {
			modifiedTimeStamp = getResourceModifiedTimeStamp(uri, objectForClassLoader, false);
			if (modifiedTimeStamp == 0) {
				modifiedTimeStamp = getResourceModifiedTimeStampFile(uri, objectForClassLoader, true);
				if (modifiedTimeStamp == 0) {
					modifiedTimeStamp = getResourceModifiedTimeStampFile(uri, objectForClassLoader, false);
				}
			}
		}
		return modifiedTimeStamp;
	}

	private long getResourceModifiedTimeStamp(String uri, final Object objectForClassLoader, final boolean translateUri) {
		if (translateUri) {
			uri = localizeUri(uri);
		}

		final File f = new File((Context.getContext()).getServletContext().getRealPath(uri));
		if (f == null || !f.exists() || !f.isFile()) {
			return 0;
		}
		return f.lastModified();
	}

	private long getResourceModifiedTimeStampFile(String uri, final Object objectForClassLoader, final boolean translateUri) {
		if (translateUri) {
			uri = localizeUri(uri);
		}

		File f = new File(uri);
		if (f != null && f.exists() && f.isFile()) {
			return f.lastModified();
		}

		final URL url = objectForClassLoader.getClass().getClassLoader().getResource(uri);
		if (url != null) {
			f = new File(url.getFile());
			return f.lastModified();
		}

		return 0;
	}

	/**
	 * Override to specify another session timeout duration
	 * @return the default session timeout in ms (300)
	 */
	public int getSessionTimeout() {
		return 300;
	}

	/**
	 * Gets the supported application locales.
	 * 
	 * @return the supported application locales
	 */
	protected List<Locale> getSupportedApplicationLocales() {
		if (locales == null) {
			locales = new ArrayList<Locale>(1);
			locales.add(getDefaultApplicationLocale());

			final String localeString = getParameter(SUPPORTED_LOCALES, "");
			final String[] localStrings = localeString.split(",");
			for (int i=0; i < localStrings.length; i++) {
				localStrings[i] = localStrings[i].trim();
				if (!locales.contains(localStrings[i])) {
					locales.add(new Locale(localStrings[i]));
				}
			}
		}

		return locales;
	}

	/**
	 * Gets the application template. If your application should use an application wide template for
	 * your views, you have to override this method and return the template {@link ViewController}.
	 * A view controller can have it own template. Templates can be used hierarchical.
	 * 
	 * @return the template
	 * 
	 * @see ViewController
	 */
	protected ViewController getTemplate() {
		return null;
	}

	/**
	 * Gets the {@link ViewAlias} alias for the current call.
	 * 
	 * @param sourcePath the source path (origin/referer of the request)
	 * @param targetPath the target path (requested target)
	 * @param parameter the parameter (request parameter)
	 * 
	 * @return the alias for
	 */
	public ViewAlias getViewAliasFor(final String sourcePath, final String targetPath, final Map<String, Object[]> parameter) {
		return null;
	}



	/**
	 * Gets the view control for an uri. You can override this method to implement a very
	 * special behavior but you should not need to do this.
	 * <br/><br/>
	 * Gets a view if no view is matching the uri in {@link #getViewController(String)}.
	 * As default if no view is matching the uri
	 * it tries to append  first "index.xhtml" than "index.html" to the uri to find
	 * a view control for this new uri.
	 * <br/><br/>
	 * Uses {@link #resolveViewController(String, boolean)} internally.
	 * @param uri the uri
	 * 
	 * @return the view control
	 */
	public ViewController getViewController(final String uri) {
		ViewController ViewController = resolveViewController(uri, true);
		if (ViewController == null) {
			ViewController = resolveViewController(uri, false);
		}

		// test index.xhtml of a directory
		if (ViewController == null) {
			final String uriIndexXHtml = StringUtils.combine("/", true, uri, "index.xhtml");
			ViewController = resolveViewController(uriIndexXHtml, true);

			if (ViewController == null) {
				ViewController = resolveViewController(uriIndexXHtml, false);
			}

			// test index.html of a directory
			if (ViewController == null) {
				final String uriIndexHtml = StringUtils.combine("/", true, uri, "index.html");
				ViewController = resolveViewController(uriIndexHtml, true);
				if (ViewController == null) {
					ViewController = resolveViewController(uriIndexHtml, false);
				}

				if (ViewController == null)
				{
					return Application.getApplication().getViewNotFoundViewController();
				}
			}

			// necessary to redirect to find direct delivery files when they are requested via the browser afterwards
			if (!uri.isEmpty() && !uri.endsWith("/")) {
				Context.getContext().redirect(uri+"/", false, Context.getContext().getRequestInputParameter());
			}
		}

		return ViewController;
	}

	/**
	 * These returns the packages where the
	 * framework looks for die {@link ViewController} classes.
	 * 
	 * Uses {@link #getParameter(String, String)} with
	 * key {@link #VIEW_CONTROL_BASE_PACKAGE} and the empty string as defaultValue
	 * to load the view control base packages (as a comma separated string) when initialized.
	 * 
	 * You can override this method to define more detailed how the
	 * framework should look for
	 * your {@link ViewController} classes.
	 * @return the view control base packages
	 */
	public String[] getViewControllerBasePackages() {
		Log.info(this, StringUtils.combineAll("using base packages: ", StringUtils.arrayAsString(basePackages)));
		return basePackages;
	}

	@SuppressWarnings("unchecked")
	private Class<? extends ViewController> getViewControllerClassForName(final String packageName, final String className, final String generatedSuffix) throws ClassNotFoundException {
		for(final String basePackage : Application.getApplication().getViewControllerBasePackages()) {
			try {
				return (Class<? extends ViewController>) Class.forName(basePackage+packageName+className);
			} catch(final ClassNotFoundException e) {
			}
		}
		for(final String basePackage : Application.getApplication().getViewControllerBasePackages()) {
			try {
				return (Class<? extends ViewController>) Class.forName(basePackage+packageName+className+generatedSuffix);
			} catch(final ClassNotFoundException e) {
			}
		}
		throw new ClassNotFoundException("Could not find class \""+className+"\" in any of the given base packages. "
				+"(package="+packageName+", generated suffix="+generatedSuffix+")");
	}

	/**
	 * Gets the view forbidden view control. In the default implementation this returns
	 *  a view to render the http error
	 * 403. You can override this method to implement your own behavior.
	 * 
	 * @return the view forbidden view control
	 */
	public ViewController getViewForbiddenViewController() {
		return new ErrorXhtmlController(403, "Forbidden", "You are not allowed to view the page you requested.");
	}

	/**
	 * Gets the view not found view control. In the default implementation this returns
	 *  a view to render the http error
	 * 404. You can override this method to implement your own behavior.
	 * 
	 * @return the view not found view control
	 */
	public ViewController getViewNotFoundViewController() {
		return new ErrorXhtmlController(404, "Page not found", "The page you requested was not found on this server.");
	}

	/**
	 * Gets the factory for the process of view design files. If you want to use
	 * another implementation instead of the current DOM-implementation you have to
	 * override this method.
	 * 
	 * @return the factory
	 */
	public ViewParser getViewParser() {
		return viewParser;
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * 
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleExecutingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleExecutingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleInitException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleInitException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleInputProcessingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleInputProcessingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handlePreparingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handlePreparingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleRenderingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleRenderingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleTemplatingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleTemplatingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleValidatingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleValidatingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/**
	 * Initializer. This method is called by the jsxp framework to initialize the application.
	 * The default implementation initializes the default. You can override it in your application class
	 * if you want to initialize the application instance. It is recommend to call this (super) method
	 * then.
	 */
	public void initialize() {
		directDeliveryAllowedFileRegExpressions = new LinkedList<String>();
		final String directDeliveryAllowedFileRegExpressionsString = getParameter(DIRECT_DELIVERY_ALLOWED_FILE_REGULAR_EXPRESSIONS, null);
		if (directDeliveryAllowedFileRegExpressionsString != null) {
			final String[] directDeliveryAllowedFileRegExpressionsStrings = directDeliveryAllowedFileRegExpressionsString.split(",");
			for (int i=0; i < directDeliveryAllowedFileRegExpressionsStrings.length; i++) {
				directDeliveryAllowedFileRegExpressionsStrings[i] = directDeliveryAllowedFileRegExpressionsStrings[i].trim();
				directDeliveryAllowedFileRegExpressions.add(directDeliveryAllowedFileRegExpressionsStrings[i]);
			}
		} else {
			// default for direct delivery allowed
			directDeliveryAllowedFileRegExpressions.add(".*\\.css");
			directDeliveryAllowedFileRegExpressions.add(".*\\.js");
			directDeliveryAllowedFileRegExpressions.add(".*\\.jpg");
			directDeliveryAllowedFileRegExpressions.add(".*\\.jpeg");
			directDeliveryAllowedFileRegExpressions.add(".*\\.gif");
			directDeliveryAllowedFileRegExpressions.add(".*\\.png");
		}

		final String directDeliveryDeniedFileRegExpressionsString = getParameter(DIRECT_DELIVERY_DENIED_FILE_REGULAR_EXPRESSIONS, null);
		if (directDeliveryDeniedFileRegExpressionsString != null) {
			directDeliveryDeniedFileRegExpressions = new LinkedList<String>();
			final String[] directDeliveryDeniedFileRegExpressionsStrings = directDeliveryDeniedFileRegExpressionsString.split(",");
			for (int i=0; i < directDeliveryDeniedFileRegExpressionsStrings.length; i++) {
				directDeliveryDeniedFileRegExpressionsStrings[i] = directDeliveryDeniedFileRegExpressionsStrings[i].trim();
				directDeliveryDeniedFileRegExpressions.add(directDeliveryDeniedFileRegExpressionsStrings[i]);
			}
		}

		final String basePackageName = getParameter(VIEW_CONTROL_BASE_PACKAGE, "");
		basePackages = basePackageName.split(",");
		for (int i=0; i < basePackages.length; i++) {
			basePackages[i] = basePackages[i].trim();
		}

		final String projectStageName = getParameter(PROJECT_STAGE_NAME, "development");
		projectStage = createProjectStage(projectStageName);
	}

	/**
	 * Process input parameter. Defines how the application will handle input.
	 * Per default it processes multipart form data and uploaded files.
	 * 
	 * @return the map< string, string[]>
	 */
	@SuppressWarnings("unchecked")
	public Map<String, Object[]> initInputParameter() {
		final HttpServletRequest request = Context.getContext().getHttpServletRequest();
		final FileUpload fileUpload = AdapterFinder.find(FileUpload.class, getParameter(FILE_UPLOAD_ADAPTER, null));

		if (fileUpload != null && fileUpload.isFileUploadSupported(request)) {
			return fileUpload.parseUploadRequest(request);
		}

		final Map<String, Object[]> parameters = new HashMap<String, Object[]>();
		for (final Object key : request.getParameterMap().keySet()) {
			parameters.put((String)key, (Object[])request.getParameterMap().get(key));
		}

		return request.getParameterMap();
	}

	/**
	 * Checks if is allowed to view page. In the default implementation this
	 * calls {@link UserContext#isAllowedToViewPage(ViewController, String)}
	 * 
	 * @param view the view (can be null if only a uri is checked)
	 * @param uri the uri
	 * 
	 * @return true, if is allowed to view page
	 */
	public boolean isAllowedToViewPage(final ViewController view, final String uri) {
		return Context.getContext().getUserContext().isAllowedToViewPage(view, uri);
	}

	/**
	 * Checks if direct delivery is allowed. Uses {@link #getDirectDeliveryAllowedList()} and {@link #getDirectDeliveryDeniedList()} to match with the url.
	 * The denied rules are stronger than the allowed rules.
	 * 
	 * @param url the url
	 * 
	 * @return true, if is direct delivery allowed
	 */
	protected final boolean isDirectDeliveryAllowed(final String url) {
		if(url == null) {
			return false;
		}
		final List<String> directDeliveryDeniedPatterns = getDirectDeliveryDeniedList();
		if(directDeliveryDeniedPatterns != null) {
			for(final String deny : directDeliveryDeniedPatterns) {
				if(deny != null && (deny.equals(url) || url.matches(deny))) {
					return false;
				}
			}
		}
		final List<String> directDeliveryAllowedPatterns = getDirectDeliveryAllowedList();
		if(directDeliveryAllowedPatterns != null) {
			for(final String allow : directDeliveryAllowedPatterns) {
				if(allow != null && (allow.equals(url) || url.matches(allow))) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Called by the jsxp framework each time a view controller is added to the
	 * life cycle manager. E.g. you can add a listener to the life cycle phases
	 * of a view control if you want need to do something application wide.
	 * 
	 * @param ViewController the view control
	 */
	public void lifeCycleStartsForViewController(final ViewController ViewController) {
	}

	/**
	 * Override to change translation behaviour. This will result in an uri that is used
	 * to find stuff on the server. Use a combination of this method and the {@link #delocalizeUriAndInitContextLocale(String)} method
	 * to implement how uri that are requested from the client should result in uri used on the server.
	 * @param uri the uri that is used after using {@link #delocalizeUriAndInitContextLocale(String)} before
	 * @return uri that has all necessary information in it to find the information for the current locale
	 */
	protected String localizeUri(String uri) {
		final Locale locale = Context.getContext().getCurrentLocale();
		if (!locale.equals(getDefaultApplicationLocale())) {
			String rootPath = uri;
			final int slashIndex = rootPath.lastIndexOf("/");
			if (slashIndex > -1) {
				rootPath = rootPath.substring(0, slashIndex);
			}
			uri = uri.substring(slashIndex);
			uri = StringUtils.combine("/", true, rootPath, StringUtil.getJavaPackageName(locale.toString()),uri);
		}
		return uri;
	}

	/**
	 * resolves a view control. creates a package and classname from the given uri to
	 * get an instance of the corresponding view controller.

	 * @param uri
	 * @param translateUri will use the {@link #localizeUri(String)} method to localize the uri if true
	 * 
	 * @return the corresponding view controller or null (if no view controller class was found) or {@link #getViewForbiddenViewController()}
	 * if the found view controller is forbidden
	 */
	protected ViewController resolveViewController(String uri, final boolean translateUri) {
		if (translateUri) {
			uri = localizeUri(uri);
		}

		String packageName = ".";
		String className = uri;

		if(uri.indexOf("/") != -1) {
			packageName = "." + uri.substring(0, uri.lastIndexOf("/"))+".";
			packageName = StringUtil.getJavaPackageName(packageName);
			className = uri.substring(uri.lastIndexOf("/")+1);
		}

		packageName = packageName.replaceAll("/", ".");
		className = StringUtil.getJavaClassName(className)+VIEW_CONTROL_SUFFIX;

		ViewController ViewController = null;
		try {
			final Class<? extends ViewController> ViewControllerClass = getViewControllerClassForName(packageName, className, GENERATED_SUFFIX);

			if (translateUri && ViewControllerClass.getName().endsWith(GENERATED_SUFFIX)) {
				return null;
			}

			if (ViewControllerClass == null) {
				throw new RuntimeException("no view control class for view '"+uri+"' found");
			}
			ViewController = ViewControllerClass.newInstance();
		} catch (final InstantiationException e) {
			throw new RuntimeException("view control instantiation failed for view '"+uri+"'", e);
		} catch (final IllegalAccessException e) {
			throw new RuntimeException("could not access view control class for view '"+uri+"'", e);
		} catch (final ClassNotFoundException e) {
			return null;
		}

		return ViewController;
	}

	/**
	 * Called by the jsxp framework if a Session was restored. Override this method
	 * to do anything with the session.
	 * 
	 * @param sessionScope the session scope
	 */
	protected void sessionRestored(final Map<String, Object> sessionScope) { }
}
