/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.jsxp.framework.ViewControllerLifeCycle.LifeCyclePhase;
import org.jsxp.framework.elementprocessing.ElementProcessorCommand;
import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.internal.ProxyElement;
import org.jsxp.framework.logging.Log;
import org.jsxp.framework.util.text.StringUtil;
import org.jsxp.framework.webroot.jsxpframework.components.RedirectViewXhtmlController;


/**
 * Servlet implementation class BaseServlet. The core of a jsxp web application.
 * This servlet needs to be configured in your application web.xml
 */
public class BaseServlet extends HttpServlet {
	private static Map<String, Object> applicationAttributes = new Hashtable<String, Object>();
	private static Map<String, Object> applicationScope = new Hashtable<String, Object>();
	private static final String FLASH_ATTRIBUTES = "org.jsxp.framework.framework.flashAttributes";
	private static final String FLASH_SCOPE = "org.jsxp.framework.framework.flashScope";
	private static final String HISTORY_MAP = "org.jsxp.framework.framework.historyMap";
	private static final long serialVersionUID = 1L;
	private static final String SESSION_ATTRIBUTES = "org.jsxp.framework.framework.sessionAttributes";
	private static final String SESSION_SCOPE = "org.jsxp.framework.framework.sessionScope";
	private static final String USER_CONTEXT = "org.jsxp.framework.framework.userSession";
	private static final String VIEW_ATTRIBUTES = "org.jsxp.framework.framework.viewAttributes";
	private static final String VIEW_SCOPE = "org.jsxp.framework.framework.viewScope";

	/**
	 * The Constructor.
	 * 
	 * @see HttpServlet#HttpServlet()
	 */
	public BaseServlet() {
		super();
	}

	private void clearNextFlashScope() {
		((FlashMap<String, Object>)Context.getContext().getScopes().getFlashScope()).flash();
		((FlashMap<String, Object>)Context.getContext().getScopes().getFlashAttributes()).flash();
	}

	private HttpSession createSession(final HttpServletRequest request, final Context c) {
		HttpSession session = request.getSession(false);
		if (session == null) {
			session = request.getSession();
			final int sessionTimeout = Application.getApplication().getSessionTimeout();
			c.addDebugMessage("session created with timeout: "+sessionTimeout);
			session.setMaxInactiveInterval(sessionTimeout);
		}

		return session;
	}

	private void createSessionScope(final Context c) {
		final Map<String, Object> sessionScope = new Hashtable<String, Object>();
		final Map<String, Object> sessionAttributes = new Hashtable<String, Object>();
		Context.getContext().addDebugMessage("new session scope: "+sessionScope);

		final UserContext userContext = Application.getApplication().createUserContext();
		sessionAttributes.put(USER_CONTEXT, userContext);

		final Hashtable<String, ViewAlias> historyMap = new Hashtable<String, ViewAlias>();
		sessionAttributes.put(HISTORY_MAP, historyMap);

		c.setUserContext(userContext);
		c.getScopes().setSessionScope(sessionScope);
		c.getScopes().setSessionAttributes(sessionAttributes);
	}

	private void createViewScope(final Context c) {
		final Map<String, Object> viewScope = new Hashtable<String, Object>();
		final Map<String, Object> viewAttributes = new Hashtable<String, Object>();
		Context.getContext().addDebugMessage("new view scope: "+viewScope);
		c.getScopes().setViewScope(viewScope);
		c.getScopes().setViewAttributes(viewAttributes);
	}

	private void deliverFileDirectly(final HttpServletResponse response, final String uri)
			throws FileNotFoundException, IOException {
		Context.getContext().addDebugMessage("direct file delivery: "+uri);
		final InputStream is = Application.getApplication().getResourceInputStream(uri, this);

		if(is != null) {
			final byte[] buffer = new byte[1024];
			int numBytes = is.read(buffer);
			while(numBytes != -1) {
				response.getOutputStream().write(buffer, 0, numBytes);
				numBytes = is.read(buffer);
			}
		}
	}

	/* (non-Javadoc)
	 * @see javax.servlet.GenericServlet#init(javax.servlet.ServletConfig)
	 */
	@Override
	public void init() throws ServletException {
		initApplication();
		initContext();
		initUserContext();
		Application.getApplication().initialize();
	}

	/**
	 * Initialize the application object
	 */
	@SuppressWarnings("unchecked")
	private void initApplication() {
		final String applicationClassName = getServletConfig().getInitParameter("applicationClassName");

		if (applicationClassName != null) {
			try {
				final Class<? extends Application> applicationClass = (Class<? extends Application>) Class.forName(applicationClassName);
				final Application application = applicationClass.newInstance();
				Application.setApplication(application);
			} catch (final Exception e) {
				throw new RuntimeException("application object init failed for '"+applicationClassName+"'", e);
			}
		} else {
			Application.setApplication(new Application());
		}

		Log.info(this, "using application object: "+Application.getApplication());
	}

	private void initCommandID(final HttpServletRequest request,
			final Context c) {
		final String commandId = request.getHeader("www.jsxp.org-spec-1.0-command-id");
		if (commandId != null) {
			c.setCommandRequest(true);
			c.setCommandId(commandId);
		}
	}

	/**
	 * init a empty context object with only the servlet,
	 *  servletconfig and applicationscope in it
	 */
	private Context initContext() {
		Context.setContext(Application.getApplication().createContext());
		final Context c = Context.getContext();
		c.setServlet(this);
		c.setServletContext(getServletContext());
		c.setServletConfig(getServletConfig());
		c.getScopes().setApplicationScope(applicationScope);
		c.getScopes().setApplicationAttributes(applicationAttributes);
		return c;
	}

	private Context initContext(final HttpServletRequest request, final HttpServletResponse response, final String uri) {
		final Context c = Context.getContext();
		c.setHttpServletRequest(request);
		c.setHttpServletResponse(response);

		final String referer = initReferer(request);
		initUpdateAndRemoveIDs(request, c);
		initCommandID(request, c);
		initRequestScope(c);

		final HttpSession session = request.getSession(false);
		restoreSessionScope(session, c);

		final ViewAlias lastView = initLastViewFromSessionAttributes(c, referer);
		restoreViewScope(referer, session, c, lastView, uri);
		restoreFlashScope(referer, session, c);

		if (session != null) {
			c.addDebugMessage("using session: "+session.getId()+ " with timeout "+session.getMaxInactiveInterval()+" seconds");
		} else {
			c.addDebugMessage("no session before request");
		}

		return c;
	}

	@SuppressWarnings("unchecked")
	private ViewAlias initLastViewFromSessionAttributes(final Context c, final String referer) {
		if (Context.getContext().getScopes().getSessionAttributes() == null) {
			return null;
		}
		final Hashtable<String, ViewAlias> historyMap = (Hashtable<String, ViewAlias>)Context.getContext().getScopes().getSessionAttributes().get(HISTORY_MAP);

		if (historyMap == null) {
			return null;
		}

		final ViewAlias lastView = historyMap.get(referer);
		c.setLastView(lastView);
		return lastView;
	}

	private String initReferer(final HttpServletRequest request) {
		String referer = request.getHeader("referer");
		Context.getContext().addDebugMessage("referer: "+referer);

		if (referer == null) {
			referer = "";
		} else {
			referer = relativeRefererUri(referer, request);
		}

		Context.getContext().addDebugMessage("relative referer: "+referer);
		return referer;
	}

	private String initRequest(final HttpServletRequest request) {
		String uri = request.getRequestURI();
		uri = relativeUri(uri, request);

		initContext();
		Context.getContext().setLocalizedRequestUri(uri);
		uri = Application.getApplication().delocalizeUriAndInitContextLocale(uri);
		Context.getContext().setDelocalizedRequestUri(uri);
		return uri;
	}

	private void initRequestScope(final Context c) {
		c.getScopes().setRequestScope(new HashMap<String, Object>());
		c.getScopes().setRequestAttributes(new HashMap<String, Object>());
		Context.getContext().addDebugMessage("new request scope: "+c.getScopes().getRequestScope());
	}

	private void initUpdateAndRemoveIDs(final HttpServletRequest request,
			final Context c) {
		if (request.getHeader("www.jsxp.org-spec-1.0-partial-request") != null) {
			c.setPartialRequest(true);
		}

		final String ajaxIdsHeader = request.getHeader("www.jsxp.org-spec-1.0-update-ids");
		if (ajaxIdsHeader != null) {
			final String[] ajaxIds = ajaxIdsHeader.split(",");
			final List<String> listAjaxIds = new LinkedList<String>();
			for (final String s : ajaxIds) {
				listAjaxIds.add(s.trim());
			}
			c.setUpdateIds(listAjaxIds);
			c.setPartialRequest(true);
		} else {
			c.setUpdateIds(new LinkedList<String>());
		}

		final List<String> removeIds = new ArrayList<String>();
		c.setRemoveIds(removeIds);
	}

	private void initUserContext() {
		Context.getContext().setUserContext(Application.getApplication().createUserContext());
	}

	private Context preprocessRequest(final HttpServletRequest request,
			final HttpServletResponse response, final String uri) {
		final Context c = initContext(request, response, uri);
		Application.getApplication().contextInitialized(Context.getContext());
		Application.getApplication().sessionRestored(Context.getContext().getScopes().getSessionScope());

		// add debug messages
		Context.getContext().addDebugMessage("using application class '"+Application.getApplication().getClass()+"'");
		Context.getContext().addDebugMessage("Using view control base packages:");
		for(final String packageName : Application.getApplication().getViewControllerBasePackages()) {
			Context.getContext().addDebugMessage(" -> '"+packageName+"'");
		}

		// init request input parameters
		final Map<String, Object[]> requestInputParameters = Application.getApplication().initInputParameter();
		Context.getContext().setRequestInputParameter(requestInputParameters);

		if (Context.getContext().isPartialRequest()) {
			Context.getContext().addDebugMessage("partial request");
			Context.getContext().addDebugMessage("updateIds: "+Context.getContext().getUpdateIds());
			Context.getContext().addDebugMessage("removeIds: "+Context.getContext().getRemoveIds());
		}
		return c;
	}

	@SuppressWarnings("unchecked")
	private void processCommandRequest(final HttpServletRequest request,
			final HttpServletResponse response, final String uri, final Context c,
			final Map<String, Object[]> requestInputParameters)
					throws ServletException, IOException {
		final HashMap<String, ElementProcessorCommand> elementProcessorCommands =
				(HashMap<String, ElementProcessorCommand>)
				Context.getContext().getScopes().getSessionAttributes().get(ElementProcessorCommand.class.getName());

		if (elementProcessorCommands == null) {
			returnNotFound(request, response, uri, c, requestInputParameters);
			return;
		}

		final ElementProcessorCommand elementProcessorCommand = elementProcessorCommands.remove(Context.getContext().getCommandId());

		if (elementProcessorCommand == null) {
			returnNotFound(request, response, uri, c, requestInputParameters);
			return;
		}

		Element element = elementProcessorCommand.process();

		if (element == null) {
			returnNotFound(request, response, uri, c, requestInputParameters);
			return;
		}

		if (element instanceof ProxyElement) {
			element = ((ProxyElement)element).getRealElement();
		}

		setPartialRequestResponseInfos(response);

		// send response to client
		Application.getApplication().getViewParser().write((InternalElement) element, response.getOutputStream(), response.getContentType());
		return;
	}

	@SuppressWarnings("unchecked")
	private ViewController processViewController(final ViewController viewController, final Context c, final String uri, final Map<String, Object[]> parameters, final HttpServletRequest request, final HttpServletResponse response) throws ServletException, IOException {
		try {
			final ViewControllerLifeCycleManager lifeCycle = new ViewControllerLifeCycleManager();
			ViewController renderViewController = lifeCycle.processLifeCycle(parameters, uri, viewController, true);
			renderViewController.setNeedToSaveScopesToSession(viewController.needToSaveScopesToSession());

			final HashMap<String, Object[]> historyParameters = new HashMap<String, Object[]>(parameters);
			final Hashtable<String, ViewAlias> historyMap = (Hashtable<String, ViewAlias>)Context.getContext().getScopes().getSessionAttributes().get(HISTORY_MAP);

			if (Context.getContext().isRedirect()) {
				final ViewController redirectViewController = Context.getContext().getRedirectViewController();
				redirectViewController.setNeedToSaveScopesToSession(viewController.needToSaveScopesToSession());
				renderViewController = redirectViewController;

				if (renderViewController.getLifeCycle().getLifeCyclePhase().equals(LifeCyclePhase.CREATED)) {
					renderViewController = lifeCycle.processLifeCycle(null, null, renderViewController, false);
				}

				historyMap.put("", new ViewAlias(viewController, historyParameters));
			}

			final StringBuffer completeUri = new StringBuffer(request.getRequestURI());
			if(request.getQueryString() != null) {
				completeUri.append("?").append(request.getQueryString());
			}
			historyMap.put(relativeUri(completeUri.toString(), request), new ViewAlias(renderViewController, historyParameters));

			if(renderViewController.getContentType() != null) {
				response.setContentType(renderViewController.getContentType());
			}

			setPartialRequestResponseInfos(response);

			response.setStatus(renderViewController.getStatusCode());
			return renderViewController;
		} catch (final Exception e) {
			throw new RuntimeException(e);
		}
	}

	private void processViewRequest(final HttpServletRequest request,
			final HttpServletResponse response, final String uri, final Context c,
			final Map<String, Object[]> requestInputParameters)
					throws ServletException, IOException {
		// get the view control
		ViewController viewController = null;
		try {
			viewController = resolveView(c, uri, requestInputParameters, request, response);
			viewController = processViewController(viewController, c, uri, requestInputParameters, request, response);
			saveScopes(viewController, request, c, uri);
		} catch (final RuntimeException t) {
			saveScopes(null, request, c, uri);
			throw t;
		}

		// send response to client
		Application.getApplication().getViewParser().write((InternalElement) viewController.getRootElement(), response.getOutputStream(), response.getContentType());
		return;
	}

	private String relativeRefererUri(final String referer, final HttpServletRequest request) {
		final String contextPath = request.getContextPath();

		URI refererUri;
		URI requestUri;
		try {
			refererUri = new URI(referer);
			requestUri = new URI(request.getRequestURL().toString());
		} catch (final URISyntaxException e) {
			throw new RuntimeException(e);
		}

		if (StringUtil.compare(refererUri.getScheme(), requestUri.getScheme()) &&
				StringUtil.compare(refererUri.getUserInfo(), requestUri.getUserInfo()) &&
				StringUtil.compare(refererUri.getHost(), requestUri.getHost()) &&
				refererUri.getPort() == requestUri.getPort()) {

			String relativeReferer = StringUtil.concateStrings(null, refererUri.getPath(), "?", refererUri.getQuery(), refererUri.getFragment());

			if (relativeReferer.endsWith("?")) {
				relativeReferer = relativeReferer.substring(0, relativeReferer.length()-1);
			}

			if (relativeReferer.startsWith(contextPath)) {
				relativeReferer = relativeReferer.substring(contextPath.length());
				if (relativeReferer.startsWith("/")) {
					if (relativeReferer.length()>1) {
						relativeReferer = relativeReferer.substring(1);
					} else {
						relativeReferer = "";
					}
				}
				return relativeReferer;
			}
		}

		return referer;
	}

	/**
	 * delete context and servlet path from request uri
	 * 
	 * @param uri
	 * @param request
	 * @return the relative uri
	 */
	private String relativeUri(String uri, final HttpServletRequest request) {
		final String contextPath = request.getContextPath();

		if (uri.startsWith(contextPath)) {
			uri = uri.substring(contextPath.length());
		}

		if (uri.startsWith("/")) {
			if (uri.length()>1) {
				uri = uri.substring(1);
			} else {
				uri = "";
			}
		}

		return uri;
	}

	private ViewController resolveView(final Context c, final String uri, Map<String, Object[]> parameters, final HttpServletRequest request, final HttpServletResponse response) throws ServletException, IOException {
		if(parameters == null) {
			parameters = new HashMap<String, Object[]>();
		}

		String lastViewUri = null;
		if(Context.getContext().getLastView() != null && Context.getContext().getLastView().getViewFilename()!=null) {
			lastViewUri = Context.getContext().getLastView().getViewFilename();
		}

		ViewController ViewController = null;

		// check for view aliases
		final ViewAlias viewAlias = Application.getApplication().getViewAliasFor(lastViewUri, uri, parameters);
		if (viewAlias != null) {
			if(!Application.getApplication().isAllowedToViewPage(null, uri)) {
				ViewController = Application.getApplication().getViewForbiddenViewController();
			} else {
				ViewController = viewAlias.getViewController();
			}
			parameters = viewAlias.getGetParameters();
			if(parameters == null) {
				parameters = new HashMap<String, Object[]>();
			}
			Context.getContext().setRequestInputParameter(parameters);
		} else {
			ViewController = Application.getApplication().getViewController(uri);
			if(!Application.getApplication().isAllowedToViewPage(ViewController, uri)) {
				ViewController = Application.getApplication().getViewForbiddenViewController();
			}
		}

		return ViewController;
	}

	@SuppressWarnings("unchecked")
	private void restoreFlashScope(String referer, final HttpSession session, final Context c) {
		Map<String, Object> flashScopePrevious = null;
		Map<String, Object> flashAttributesPrevious = null;
		if (session == null) {
			Context.getContext().addDebugMessage("no previous flash scope");
		} else {
			flashScopePrevious = (Map<String, Object>) session.getAttribute(referer+"|"+FLASH_SCOPE);
			flashAttributesPrevious = (Map<String, Object>) session.getAttribute(referer+"|"+FLASH_ATTRIBUTES);
			if (flashScopePrevious == null && referer.contains("?")) {
				referer = referer.substring(0, referer.indexOf("?"));
				flashScopePrevious = (Map<String, Object>) session.getAttribute(referer+"|"+FLASH_SCOPE);
				flashAttributesPrevious = (Map<String, Object>) session.getAttribute(referer+"|"+FLASH_ATTRIBUTES);
			}

			if (flashScopePrevious == null) {
				Context.getContext().addDebugMessage("no previous flash scope");
			} else {
				Context.getContext().addDebugMessage("using previous flash scope: "+flashScopePrevious);
			}
		}
		final FlashMap<String, Object> flashScope = new FlashMap<String, Object>(flashScopePrevious);
		final FlashMap<String, Object> flashAttributes = new FlashMap<String, Object>(flashAttributesPrevious);
		c.getScopes().setFlashScope(flashScope);
		c.getScopes().setFlashAttributes(flashAttributes);
	}

	@SuppressWarnings("unchecked")
	private void restoreSessionScope(final HttpSession session, final Context c) {
		if (session == null) {
			createSessionScope(c);
			return;
		}

		final Map<String, Object> sessionScope = (Map<String, Object>) session.getAttribute(SESSION_SCOPE);
		final Map<String, Object> sessionAttributes = (Map<String, Object>) session.getAttribute(SESSION_ATTRIBUTES);

		if (sessionScope == null || sessionAttributes == null) {
			createSessionScope(c);
			return;
		}
		Context.getContext().addDebugMessage("using session scope: "+sessionScope);
		final UserContext userSession = (UserContext) sessionAttributes.get(USER_CONTEXT);
		c.setUserContext(userSession);
		c.getScopes().setSessionScope(sessionScope);
		c.getScopes().setSessionAttributes(sessionAttributes);
	}

	@SuppressWarnings("unchecked")
	private void restoreViewScope(final String referer, final HttpSession session, final Context c, final ViewAlias lastView, final String uri) {
		if (session == null) {
			createViewScope(c);
			return;
		}

		final Map<String, Object> viewScope = (Map<String, Object>) session.getAttribute(referer+"|"+VIEW_SCOPE);
		final Map<String, Object> viewAttributes = (Map<String, Object>) session.getAttribute(referer+"|"+VIEW_ATTRIBUTES);

		if (viewScope == null || viewAttributes == null || lastView == null || !lastView.getViewFilename().equals(uri) || lastView.getViewController()==null) {
			createViewScope(c);
			return;
		}

		Context.getContext().addDebugMessage("using view scope: "+viewScope);
		c.getScopes().setViewScope(viewScope);
		c.getScopes().setViewAttributes(viewAttributes);
	}

	private void returnNotFound(final HttpServletRequest request,
			final HttpServletResponse response, final String uri, final Context c,
			final Map<String, Object[]> requestInputParameters)
					throws ServletException, IOException {
		final ViewController ViewController = processViewController(Application.getApplication().getViewNotFoundViewController(), c, uri, requestInputParameters, request, response);
		// send response to client
		Application.getApplication().getViewParser().write((InternalElement) ViewController.getRootElement(), response.getOutputStream(), response.getContentType());
		return;
	}

	private void saveFlashScope(final HttpSession session, final HttpServletRequest request, final Context c, final String uri) {
		final StringBuffer completeUri = new StringBuffer(uri);

		session.setAttribute(relativeUri(completeUri.toString(), request)+"|"+FLASH_SCOPE, c.getScopes().getFlashScope());
		session.setAttribute(relativeUri(completeUri.toString(), request)+"|"+FLASH_ATTRIBUTES, c.getScopes().getFlashAttributes());
		if(request.getQueryString() != null) {
			completeUri.append("?").append(request.getQueryString());
			session.setAttribute(relativeUri(completeUri.toString(), request)+"|"+FLASH_SCOPE, c.getScopes().getFlashScope());
			session.setAttribute(relativeUri(completeUri.toString(), request)+"|"+FLASH_ATTRIBUTES, c.getScopes().getFlashAttributes());
		}

		session.setAttribute(""+"|"+FLASH_SCOPE, c.getScopes().getFlashScope());
		session.setAttribute(""+"|"+FLASH_ATTRIBUTES, c.getScopes().getFlashAttributes());
	}

	private void saveScopes(final ViewController viewController, final HttpServletRequest request, final Context c, final String uri) {
		if (!(viewController instanceof RedirectViewXhtmlController)) {
			clearNextFlashScope();
		}

		HttpSession session = null;

		if (viewController != null && viewController.needToSaveScopesToSession()) {
			session = createSession(request, c);
			saveSessionScope(session, c);
			saveViewScope(uri, session, c);

			String completeUri = request.getRequestURI();
			if (viewController instanceof RedirectViewXhtmlController) {
				completeUri = initReferer(request);
			}
			saveFlashScope(session, request, c, completeUri);
		}
	}

	private void saveSessionScope(final HttpSession session, final Context c) {
		session.setAttribute(SESSION_SCOPE, c.getScopes().getSessionScope());
		session.setAttribute(SESSION_ATTRIBUTES, c.getScopes().getSessionAttributes());
	}

	private void saveViewScope(final String uri, final HttpSession session, final Context c) {
		session.setAttribute(uri+"|"+VIEW_SCOPE, c.getScopes().getViewScope());
		session.setAttribute(uri+"|"+VIEW_ATTRIBUTES, c.getScopes().getViewAttributes());
	}

	/* (non-Javadoc)
	 * @see javax.servlet.http.HttpServlet#service(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
	 */
	@Override
	protected void service(final HttpServletRequest request, final HttpServletResponse response) throws ServletException, IOException {

		final String uri = initRequest(request);

		if(Application.getApplication().isDirectDeliveryAllowed(uri)) {
			deliverFileDirectly(response, uri);
			return;
		} else {
			final Context c = preprocessRequest(request, response, uri);

			if (Context.getContext().isCommandRequest()) {
				processCommandRequest(request, response, uri, c, Context.getContext().getRequestInputParameter());
			} else {
				processViewRequest(request, response, uri, c, Context.getContext().getRequestInputParameter());
			}
		}
	}

	private void setPartialRequestResponseInfos(final HttpServletResponse response) {
		if (Context.getContext().isPartialRequest()) {
			response.setContentType("text/xml");
		}

		if (!Context.getContext().getUpdateIds().isEmpty()) {
			final List<String> updateIds = Context.getContext().getUpdateIds();
			response.setHeader("www.jsxp.org-spec-1.0-update-ids", StringUtil.concateStrings(',', updateIds));
		}

		if (!Context.getContext().getRemoveIds().isEmpty()) {
			final List<String> removeIds = Context.getContext().getRemoveIds();
			response.setHeader("www.jsxp.org-spec-1.0-remove-ids", StringUtil.concateStrings(',', removeIds));
		}
	}

}
