/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import javax.servlet.Servlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.jsxp.framework.logging.Log;
import org.jsxp.framework.webroot.jsxpframework.components.DebugMessagesXhtmlController;
import org.jsxp.framework.webroot.jsxpframework.components.DeveloperWarningsXhtmlController;
import org.jsxp.framework.webroot.jsxpframework.components.FlashMessagesXhtmlController;
import org.jsxp.framework.webroot.jsxpframework.components.RedirectViewXhtmlController;


/**
 * The Class Context. Default implementation of a context.
 * 
 * The context is automatically instantiated by the framework. You can access it with
 * {@link Context#getContext()}. Do not create an instance yourself.
 */
public class Context {

	private static ThreadLocal<Context> contextInstance =
			new ThreadLocal<Context>();

	/** The Constant PREVIOUS_INPUT_PARAMETERS. */
	protected static final String PREVIOUS_INPUT_PARAMETERS = "org.jsxp.framework.framework.redirect.previousInputParameters";

	/** The Constant RESTORE_PREVIOUS_INPUT. */
	protected static final String RESTORE_PREVIOUS_INPUT = "org.jsxp.framework.framework.redirect.restorePreviousInput";
	/**
	 * Gets the context. Use this method to get the current context anywhere in your code.
	 * 
	 * @return the context
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Context> T getContext() {
		return (T)contextInstance.get();
	}
	/**
	 * Sets the context. Used internally by the jsxp framework.
	 * 
	 * @param context the new context
	 */
	static void setContext(final Context context) {
		contextInstance.set(context);
	}
	private String commandId;
	private boolean commandRequest;
	private final List<String> debugMessages = new LinkedList<String>();
	private String delocalizedRequestUri;
	private final FlashMessages flashMessages = new FlashMessages(this);
	private HttpServletRequest httpServletRequest;
	private HttpServletResponse httpServletResponse;
	private ViewAlias lastView;
	private String localizedRequestUri;
	private boolean partialRequest;
	private ViewController redirectViewController;

	private List<String> removeIds;
	private Map<String, Object[]> requestInputParameter;
	private Locale requestLocale;
	private final Scopes<String, Object> scopes = new Scopes<String, Object>();
	private Servlet servlet;
	private ServletConfig servletConfig;
	private ServletContext servletContext;

	private List<String> updateIds;

	private UserContext userSession;

	private final List<String> warnings = new LinkedList<String>();

	/**
	 * Instantiates a new context. Should only be used by the jsxp framework. Use
	 * {@link #getContext()} instead.
	 */
	protected Context () {
	}

	/**
	 * Adds the debug message.
	 * 
	 * @param message the message
	 * 
	 * @see DebugMessagesXhtmlController
	 */
	public void addDebugMessage(final String message) {
		debugMessages.add(message);
		Log.debug(this, message);
	}

	/**
	 * Adds a developer warning.
	 * 
	 * @param message the message
	 * 
	 * @see DeveloperWarningsXhtmlController
	 */
	public void addDeveloperWarningMessage(final String message) {
		warnings.add(message);
	}

	/**
	 * Adds the flash message.
	 * 
	 * @param message the message
	 * 
	 * @see FlashMessagesXhtmlController
	 */
	public void addFlashErrorMessage(final String message) {
		flashMessages.addErrorMessage(message);
	}

	/**
	 * Adds the flash message.
	 * 
	 * @param message the message
	 * 
	 * @see FlashMessagesXhtmlController
	 */
	public void addFlashInformationMessage(final String message) {
		flashMessages.addInformationMessage(message);
	}

	/**
	 * Adds the flash message.
	 * 
	 * @param message the message
	 * 
	 * @see FlashMessagesXhtmlController
	 */
	public void addFlashWarningMessage(final String message) {
		flashMessages.addWarningMessage(message);
	}

	public void clearRedirect() {
		this.redirectViewController = null;
	}

	/**
	 * Gets the command id if this is a command request
	 * @return a currently setted command if, if there is one in this request
	 */
	public String getCommandId() {
		return commandId;
	}

	/**
	 * Gets the context path.
	 * 
	 * @return the context path
	 */
	public String getContextPath() {
		return this.httpServletRequest.getContextPath();
	}

	/**
	 * Gets the current locale. Either it returns the locale that is explicitly set at the context via {@link #setRequestLocale(Locale)}
	 * or it will try to find the best fitting locale regarding to the applications supported locales and the preferred locales from
	 * the user context. If nothing is found it will return the default application locale.
	 * 
	 * @return the current locale
	 */
	public Locale getCurrentLocale() {
		if(requestLocale != null) {
			return requestLocale;
		}

		// default locale
		if (getUserContext() == null || Application.getApplication().getDefaultApplicationLocale().equals(Context.getContext().getUserContext().getPreferredUserLocale())) {
			return Application.getApplication().getDefaultApplicationLocale();
		}

		// supported locales
		final int bestMatchQuality = 0;
		Locale bestMatch = null;

		final List<Locale> locales = Application.getApplication().getSupportedApplicationLocales();
		for (final Locale ul : Context.getContext().getUserContext().getPreferredUserLocales()) {
			for(final Locale al : locales) {
				int matchQuality = 0;
				if(ul.getLanguage().equals(al.getLanguage())) {
					matchQuality++;
					if(ul.getCountry().equals(al.getCountry())) {
						matchQuality++;
						if(ul.getVariant().equals(al.getVariant())) {
							matchQuality++;
						}
					}
				}

				if(matchQuality > bestMatchQuality) {
					bestMatch = al;
				}
			}
		}

		if(bestMatch != null) {
			return bestMatch;
		}

		return Application.getApplication().getDefaultApplicationLocale();
	}

	/**
	 * Gets the debug messages.
	 * 
	 * @return the debug messages
	 * 
	 * @see DebugMessagesXhtmlController
	 */
	public List<String> getDebugMessages() {
		return debugMessages;
	}

	public String getDelocalizedRequestUri() {
		return delocalizedRequestUri;
	}

	/**
	 * Gets the developer warnings.
	 * 
	 * @return the developer warnings
	 * 
	 * @see DeveloperWarningsXhtmlController
	 */
	public List<String> getDeveloperWarnings() {
		return warnings;
	}

	/**
	 * Gets the flash messages.
	 * 
	 * @return the flash messages
	 * 
	 * @see FlashMessagesXhtmlController
	 * @see FlashMessages
	 */
	public FlashMessages getFlashMessages() {
		return flashMessages;
	}

	/**
	 * Gets the http servlet request.
	 * 
	 * @return the http servlet request
	 */
	public HttpServletRequest getHttpServletRequest() {
		return httpServletRequest;
	}

	/**
	 * Gets the http servlet response.
	 * 
	 * @return the http servlet response
	 */
	public HttpServletResponse getHttpServletResponse() {
		return httpServletResponse;
	}

	/**
	 * Gets the http session.
	 * 
	 * @return the http session
	 */
	public HttpSession getHttpSession() {
		return httpServletRequest.getSession(false);
	}

	/**
	 * Gets the last view.
	 * 
	 * @return the last view
	 */
	public ViewAlias getLastView() {
		return lastView;
	}

	public String getLocalizedRequestUri() {
		return localizedRequestUri;
	}

	/**
	 * Gets the redirect view control if context was set for redirect.
	 * 
	 * @return the redirect view control
	 */
	public final ViewController getRedirectViewController() {
		return redirectViewController;
	}

	/**
	 * Gets the removed IDs. Used if the clients states that he wants to remove some elements from a view.
	 * E.g. used via AJAX in a WebApplication. You can call this method to know what elements
	 * the client want to get rid off.
	 * 
	 * 
	 * @return the removes the ids
	 */
	public List<String> getRemoveIds() {
		return this.removeIds;
	}

	/**
	 * Gets the request input parameter.
	 * 
	 * @return the request input parameter
	 */
	public Map<String, Object[]> getRequestInputParameter() {
		return requestInputParameter;
	}

	/**
	 * Gets the request locale.
	 * 
	 * @return the request locale
	 */
	public Locale getRequestLocale() {
		return requestLocale;
	}

	/**
	 * 
	 * @return the previous save input parameter
	 */
	@SuppressWarnings("unchecked")
	public Map<String, Object[]> getSavedPreviousInput() {
		return (Map<String, Object[]>) getScopes().getFlashAttributes().get(PREVIOUS_INPUT_PARAMETERS);
	}

	/**
	 * Gets the scopes.
	 * 
	 * @return the scopes
	 */
	public Scopes<String, Object> getScopes() {
		return scopes;
	}

	/**
	 * Gets the servlet.
	 * 
	 * @return the servlet
	 */
	public Servlet getServlet() {
		return servlet;
	}

	/**
	 * Gets the servlet config.
	 * 
	 * @return the servlet config
	 */
	public ServletConfig getServletConfig() {
		return servletConfig;
	}


	/**
	 * Gets the servlet context.
	 * 
	 * @return the servlet context
	 */
	public ServletContext getServletContext() {
		return servletContext;
	}

	/**
	 * Gets the update ids. Used if only special elements from a view are requested. E.g.
	 * used via AJAX in a WebApplication. You can call this method to know what elements are requested.
	 * 
	 * @return the update ids
	 */
	public List<String> getUpdateIds() {
		return this.updateIds;
	}

	/**
	 * Gets the current user context.
	 * 
	 * @return the user context
	 */
	public UserContext getUserContext() {
		return userSession;
	}

	/**
	 * 
	 * @return is this a command request (for example for AJAX)
	 */
	public boolean isCommandRequest() {
		return commandRequest;
	}

	/**
	 * @return is partial request
	 */
	public boolean isPartialRequest() {
		return partialRequest;
	}

	public boolean isPreviousInputSaved() {
		final Boolean shouldRestorePreviousInput = (Boolean) Context.getContext().getScopes().getFlashAttributes().get(Context.RESTORE_PREVIOUS_INPUT);
		return Boolean.TRUE.equals(shouldRestorePreviousInput);
	}

	/**
	 * 
	 * @return true if the context is set for redirect and the redirect is a client side redirect, otherwise false
	 */
	public boolean isRedirect()
	{
		return redirectViewController != null;
	}

	/**
	 * Marks the request for redirect to another url.
	 * 
	 * @param url the url
	 * @param clientRedirect true if the redirect is a client side redirect, otherwise it is a server side redirect
	 * @param redirectParameter the redirect parameter
	 */
	public final void redirect(final String url, final boolean clientRedirect, final Map<String, Object[]> redirectParameter) {
		this.redirectViewController = new RedirectViewXhtmlController(localizedRequestUri, url, redirectParameter, clientRedirect);
	}

	/**
	 * Marks the request for redirect to a view alias target.
	 * @param viewAlias
	 * @param clientRedirect
	 */
	public final void redirect(final ViewAlias viewAlias, final boolean clientRedirect) {
		final RedirectViewXhtmlController redirect = new RedirectViewXhtmlController(localizedRequestUri, viewAlias.getViewFilename(), viewAlias.getGetParameters(), clientRedirect);
		this.redirectViewController = redirect;
	}

	/**
	 * Marks the request for redirect to another view.
	 * 
	 * @param redirectViewController the redirect view control
	 * @param clientRedirect true if the redirect is a client side redirect, otherwise it is a server side redirect
	 * @param redirectParameter the redirect parameter
	 */
	public final void redirect(final ViewController redirectViewController, final boolean clientRedirect, final Map<String, Object[]> redirectParameter) {
		this.redirectViewController = new RedirectViewXhtmlController(localizedRequestUri, redirectViewController.getViewFilename(), redirectParameter, clientRedirect);
	}

	/**
	 * Activate/deactivate the restoration of the previous input in the next request.
	 * E.g. set this at the current context in a view controller to true, if you want to redirect
	 * to a html page in a web application and fill the form with the old values after a validation
	 * error.
	 */
	public void saveInputToRestoreLater() {
		Context.getContext().getScopes().getFlashAttributes().put(RESTORE_PREVIOUS_INPUT, true);
		//We have to clone the parameter map since the HttpRequest re-uses it in the next request
		final HashMap<String, Object[]> requestInputParameters = new HashMap<String, Object[]>(getRequestInputParameter());
		getScopes().getFlashAttributes().put(PREVIOUS_INPUT_PARAMETERS, requestInputParameters);
	}

	void setCommandId(final String commandId) {
		this.commandId = commandId;
	}

	void setCommandRequest(final boolean commandRequest) {
		this.commandRequest = commandRequest;
	}



	public void setDelocalizedRequestUri(final String delocalizedRequestUri) {
		this.delocalizedRequestUri = delocalizedRequestUri;
	}

	/**
	 * Sets the http servlet request.
	 * 
	 * @param httpServletRequest the new http servlet request
	 */
	void setHttpServletRequest(final HttpServletRequest httpServletRequest) {
		this.httpServletRequest = httpServletRequest;
	}

	/**
	 * Sets the http servlet response.
	 * 
	 * @param httpServletResponse the new http servlet response
	 */
	void setHttpServletResponse(final HttpServletResponse httpServletResponse) {
		this.httpServletResponse = httpServletResponse;
	}

	/**
	 * Sets the last view. Used internally by the jsxp framework.
	 * 
	 * @param lastView the new last view
	 */
	void setLastView(final ViewAlias lastView) {
		this.lastView = lastView;
	}

	public void setLocalizedRequestUri(final String localizedRequestUri) {
		this.localizedRequestUri = localizedRequestUri;
	}

	void setPartialRequest(final boolean partialRequest) {
		this.partialRequest = partialRequest;
	}

	/**
	 * Sets the IDs to be removed. Used internally by the jsxp framework.
	 * 
	 * @param removeIds the IDs to be removed
	 */
	void setRemoveIds(final List<String> removeIds) {
		this.removeIds = removeIds;
	}

	/**
	 * Sets the request input parameter.
	 * 
	 * @param requestInputParameter the new request input parameter
	 */
	void setRequestInputParameter(final Map<String, Object[]> requestInputParameter) {
		this.requestInputParameter = requestInputParameter;
	}

	/**
	 * Sets the request locale. Use this method if you want to set the locale for the request
	 * explicitly.
	 * 
	 * @param requestLocale the new request locale
	 */
	void setRequestLocale(final Locale requestLocale) {
		this.requestLocale = requestLocale;
	}

	/**
	 * Sets the servlet.
	 * 
	 * @param servlet the new servlet
	 */
	void setServlet(final Servlet servlet) {
		this.servlet = servlet;
	}

	/**
	 * Sets the servlet config.
	 * 
	 * @param servletConfig the new servlet config
	 */
	void setServletConfig(final ServletConfig servletConfig) {
		this.servletConfig = servletConfig;
	}

	/**
	 * Sets the servlet context.
	 * 
	 * @param servletContext the new servlet context
	 */
	void setServletContext(final ServletContext servletContext) {
		this.servletContext = servletContext;
	}

	/**
	 * Sets the update ids. Used internally by the jsxp framework.
	 * 
	 * @param updateIds the new update ids
	 */
	void setUpdateIds(final List<String> updateIds) {
		this.updateIds = updateIds;
	}

	/**
	 * Sets the user context. Used internally by the jsxp framework.
	 * 
	 * @param userSession the new user context
	 */
	void setUserContext(final UserContext userSession) {
		this.userSession = userSession;
	}
}
