/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.util.Map;

/**
 * The Interface Element. Basic interface for jsxp elements. Elements are the basic
 * parts a view is composed of.
 */
public interface Element {

	/**
	 * Adds the element.
	 * 
	 * @param e the e
	 */
	public void addElement(Element e);

	/**
	 * Adds the elements.
	 * 
	 * @param elements the elements
	 */
	public void addElements(Element[] elements);

	/**
	 * Adds the text.
	 * 
	 * @param text the text
	 */
	public void addText(String text);

	/**
	 * Adds the as elements.
	 * 
	 * @param xml the xml
	 * 
	 * @return the element
	 */
	public Element addXmlAsElements(String xml);

	/**
	 * Copy.
	 * 
	 * @return the element
	 */
	public Element copy();

	/**
	 * Copy.
	 * 
	 * @param idPrefix the id prefix
	 * @param idSuffix the id suffix
	 * 
	 * @return the element
	 */
	public Element copy(String idPrefix, String idSuffix);

	/**
	 * Exchange.
	 * 
	 * @param e the e
	 */
	public void exchange(Element e);

	/**
	 * Exchange.
	 * 
	 * @param elements the elements
	 */
	public void exchange(Element[] elements);

	/**
	 * Find element by id.
	 * 
	 * @param id the id
	 * 
	 * @return the element
	 */
	public Element findElementById(String id);

	/**
	 * Find element by name.
	 * 
	 * @param name the name
	 * @param namespaceURI the namespace uri
	 * 
	 * @return the element
	 */
	public Element findElementByName(String name, String namespaceURI);

	/**
	 * Gets the attribute.
	 * 
	 * @param key the key
	 * 
	 * @return the attribute
	 */
	public String getAttribute(String key);

	/**
	 * Gets the attribute.
	 * 
	 * @param namespace the namespace
	 * @param key the key
	 * 
	 * @return the attribute
	 */
	public String getAttribute(String namespace, String key);

	/**
	 * Gets the attributes.
	 * 
	 * @return the attributes
	 */
	public Map<String, String> getAttributes();

	/**
	 * Gets the elements.
	 * 
	 * @return the elements
	 */
	public Element[] getElements();

	/**
	 * Gets the id.
	 * 
	 * @return the id
	 */
	public String getId();

	/**
	 * Gets the id prefix.
	 * 
	 * @return the id prefix
	 */
	public String getIdPrefix();

	/**
	 * Gets the id suffix.
	 * 
	 * @return the id suffix
	 */
	public String getIdSuffix();

	/**
	 * Gets the name.
	 * 
	 * @return the name
	 */
	public String getName();

	/**
	 * Gets the namespace uri.
	 * 
	 * @return the namespace uri
	 */
	public String getNamespaceURI();

	/**
	 * Gets the parent.
	 * 
	 * @return the parent
	 */
	public Element getParent();

	/**
	 * Gets the source view controller.
	 * 
	 * @return the source view controller
	 */
	public ViewController getSourceViewController();

	/**
	 * Gets the value.
	 * 
	 * @return the value
	 */
	public String getValue();

	/**
	 * Checks for attribute.
	 * 
	 * @param key the key
	 * 
	 * @return true, if successful
	 */
	public boolean hasAttribute(String key);

	/**
	 * Checks for attribute.
	 * 
	 * @param namespace the namespace
	 * @param key the key
	 * 
	 * @return true, if successful
	 */
	public boolean hasAttribute(String namespace, String key);

	/**
	 * Checks if is text.
	 * 
	 * @return true, if is text
	 */
	public boolean isText();

	/**
	 * Removes the.
	 */
	public void remove();

	/**
	 * Removes the attribute.
	 * 
	 * @param name the name
	 */
	public void removeAttribute(String name);

	/**
	 * Removes the attribute.
	 * 
	 * @param namespace the namespace
	 * @param name the name
	 */
	public void removeAttribute(String namespace, String name);

	/**
	 * Removes the element.
	 * 
	 * @param e the e
	 */
	public void removeElement(Element e);

	/**
	 * Removes the elements.
	 */
	public void removeElements();

	/**
	 * Sets the attribute.
	 * 
	 * @param key the key
	 * @param value the value
	 */
	public void setAttribute(String key, String value);

	/**
	 * Sets the attribute.
	 * 
	 * @param namespace the namespace
	 * @param key the key
	 * @param value the value
	 */
	public void setAttribute(String namespace, String key, String value);

	/**
	 * Sets the value.
	 * 
	 * @param value the new value
	 */
	public void setValue(String value);

	/**
	 * Sets the values.
	 * 
	 * @param values the new values
	 */
	public void setValues(String[] values);

	/**
	 * Sets the variable.
	 * 
	 * @param name the name
	 * @param value the value
	 */
	void setVariable(String name, String value);

	/**
	 * Fill variables. Use a bean to match get/setter of the bean and the element-variables to fill the variables with values.
	 * 
	 * @param bean the bean used to fill the variables of an element
	 */
	public void setVariablesWithBean(Object bean);

}
