/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

import org.jsxp.framework.annotations.ElementGetter;
import org.jsxp.framework.annotations.VariableSetter;
import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.util.text.StringUtil;


/**
 * The Class ElementProxyInvocationhandler.
 */
class ElementProxyInvocationhandler implements InvocationHandler {

	private final InternalElement elem;
	private final Class<? extends Element> interfaceType;

	/**
	 * Instantiates a new element proxy invocationhandler.
	 * 
	 * @param elem the elem
	 * @param interfaceType the interface type
	 */
	ElementProxyInvocationhandler(final InternalElement elem, final Class<? extends Element> interfaceType) {
		this.elem = elem;
		this.interfaceType = interfaceType;
	}

	/* (non-Javadoc)
	 * @see java.lang.reflect.InvocationHandler#invoke(java.lang.Object, java.lang.reflect.Method, java.lang.Object[])
	 */
	@Override
	public Object invoke(final Object proxy, final Method method, final Object[] args) throws Throwable {

		String idPrefix = null;

		if("getRealElement".equals(method.getName())) {
			return elem;
		} else if ("setVariablesWithBean".equals(method.getName()) && args != null && args.length==1) {
			setVariablesWithBean(args[0], proxy);
			return null;
		} else if(method.getDeclaringClass() == Element.class && method.getDeclaringClass() == Object.class) {
			return method.invoke(elem, args);
		} else if(method.getAnnotation(ElementGetter.class) != null) {
			final String suffix = elem.getIdSuffix();
			if (suffix == null) {
				return elem.findElementById(method.getAnnotation(ElementGetter.class).name());
			} else {
				return elem.findElementById(StringUtil.concateStrings(null, method.getAnnotation(ElementGetter.class).name(), "$", suffix));
			}
		} else if(method.getAnnotation(VariableSetter.class) != null) {
			elem.setVariable(method.getAnnotation(VariableSetter.class).name(), (String) args[0]);
		} else if("createTemplate".equals(method.getName())) {
			if (args != null && args.length == 1 && args[0] != null) { idPrefix = args[0].toString(); }
			final ElementTemplate<Element> template =
					new ElementTemplate<Element>(((Element)proxy).copy(null, "template_"+toString()), interfaceType, idPrefix);
			return template;
		} else if("useAsTemplateOnly".equals(method.getName())) {
			if (args != null && args.length == 1 && args[0] != null) { idPrefix = args[0].toString(); }
			final ElementTemplate<Element> template =
					new ElementTemplate<Element>(((Element)proxy).copy(null, "template_"+toString()), interfaceType, idPrefix);
			((Element)proxy).remove();
			return template;
		} else {
			return method.invoke(elem, args);
		}

		return null;
	}

	/**
	 * Fill variables. Use a bean to match get/setter of the bean and the element-variables to fill the variables with values.
	 * 
	 * @param bean the bean used to fill the variables of an element
	 * @param proxy the proxy
	 */
	private void setVariablesWithBean(final Object bean, final Object proxy) {
		if (bean == null) {
			return;
		}

		for (final Method m : bean.getClass().getMethods()) {
			if (m.getName().startsWith("get") && !m.getReturnType().equals(void.class) && m.getParameterTypes().length == 0) {
				final String propertyName = m.getName().substring(3);
				Method mm;

				try {
					mm = proxy.getClass().getMethod("setVariable"+propertyName, new Class[] { String.class });
				} catch (final Exception e) {
					continue;
				}

				try {
					if (mm != null) {
						final Object propertyValue = m.invoke(bean, (Object[]) null);
						String strValue = null;
						if (propertyValue != null) {
							if (propertyValue instanceof String) {
								strValue = (String)propertyValue;
							} else {
								strValue = propertyValue.toString();
							}
						}
						elem.setVariable(propertyName, strValue);
					}
				} catch (final Exception e) {
					throw new RuntimeException(e);
				}
			}
		}
	}

}
