/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.lang.reflect.Proxy;
import java.util.LinkedList;

import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.internal.ProxyElement;


/**
 * The Class ElementTemplate. Use this class to create a template for a jsxp element. This way
 * you can use a jsxp element as a prototype and clone it. E.g. you can create lists by using
 * an element to define how a row in the list should look like.
 */
public class ElementTemplate<ElementType extends Element> {

	private int counter = 0;
	private final LinkedList<Element> disable = new LinkedList<Element>();
	private final LinkedList<Element> disableOnce = new LinkedList<Element>();
	private final ElementType element;
	private final String idPrefix;
	private final Class<? extends Element> interfaceType;

	/**
	 * Instantiates a new element template.
	 * 
	 * @param element the element
	 * @param interfaceType the interface type
	 * @param idPrefix the id prefix
	 */
	public ElementTemplate(final ElementType element, final Class<? extends Element> interfaceType, final String idPrefix) {
		this.element = element;
		this.interfaceType = interfaceType;
		this.idPrefix = idPrefix;
	}

	/**
	 * Creates the element.
	 * 
	 * @return the element type
	 */
	public ElementType createElement() {
		return createElement(null);
	}

	/**
	 * Creates the element.
	 * 
	 * @param bean the parameter bean to fill the element variables with values from the get-methods of the bean.
	 * 
	 * @return the element type
	 */
	public ElementType createElement(final Object bean) {
		final InternalElement newElement = (InternalElement) element.copy(idPrefix, ""+counter++);

		removeInstanceElements(newElement);
		final ElementProxyInvocationhandler handler = new ElementProxyInvocationhandler(newElement, interfaceType);

		final ElementType result = createProxyInstance(handler);
		if (bean != null) {	result.setVariablesWithBean(bean); }

		disableOnce.clear();

		return result;
	}

	@SuppressWarnings("unchecked")
	private ElementType createProxyInstance(final ElementProxyInvocationhandler handler) {
		return (ElementType) Proxy.newProxyInstance(getClass().getClassLoader(), new Class[] {interfaceType, ProxyElement.class}, handler);
	}

	/**
	 * Disable a subelement for the next calls of a createElement-method. The created element
	 * won�t have this subelement as child.
	 * 
	 * @param element the element
	 */
	public void disable(final Element element) {
		if(element.getId() != null && !"".equals(element.getId())) {
			disable.add(element);
		}
	}

	/**
	 * Disable a subelement once for the next call of a createElement-method. The created element
	 * won�t have this subelement as child.
	 * 
	 * @param element the element
	 */
	public void disableOnce(final Element element) {
		if(element.getId() != null && !"".equals(element.getId())) {
			disableOnce.add(element);
		}
	}

	/**
	 * Enable a subelement that you have disabled earlier.
	 * 
	 * @param element the element
	 */
	public void enable(final Element element) {
		disableOnce.remove(element);
		disable.remove(element);

	}

	private boolean removeElementIfNecessary(final Element element) {
		final String elementId = element.getId();
		for(final Element cur : disableOnce) {
			if(cur.getId().equals(elementId)) {
				element.remove();
				return true;
			}
		}
		for(final Element cur : disable) {
			if(cur.getId().equals(elementId)) {
				element.remove();
				return true;
			}
		}
		return false;
	}

	private void removeInstanceElements(final Element newElement) {
		if(newElement.getId() != null && !"".equals(newElement.getId())) {
			if (removeElementIfNecessary(newElement)) {
				return;
			}
		}
		for(final Element element : newElement.getElements()) {
			removeInstanceElements(element);
		}
	}
}
