/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * The Class FlashMap. Used as container for the flash scope.
 */
public class FlashMap<K, V> implements Map<K, V> {

	private Map<K,V> merged = null;
	private Map<K,V> previous = null;

	/**
	 * Instantiates a new flash map.
	 */
	public FlashMap() {
		this(null);
	}

	/**
	 * Instantiates a new flash map.
	 * 
	 * @param previous the previous
	 */
	public FlashMap(Map<K,V> previous) {
		if (previous == null) {
			previous = new HashMap<K, V>();
		}
		this.previous = previous;
		merged = new HashMap<K, V>();;
		merged.putAll(previous);
	}

	/* (non-Javadoc)
	 * @see java.util.Map#clear()
	 */
	@Override
	public void clear() {
		previous.clear();
		merged.clear();
	}

	/* (non-Javadoc)
	 * @see java.util.Map#containsKey(java.lang.Object)
	 */
	@Override
	public boolean containsKey(final Object key) {
		return merged.containsKey(key);
	}

	/* (non-Javadoc)
	 * @see java.util.Map#containsValue(java.lang.Object)
	 */
	@Override
	public boolean containsValue(final Object value) {
		return merged.containsValue(value);
	}

	/* (non-Javadoc)
	 * @see java.util.Map#entrySet()
	 */
	@Override
	public Set<java.util.Map.Entry<K, V>> entrySet() {
		return merged.entrySet();
	}

	/**
	 * Flash the map.
	 */
	public void flash() {
		for (final K key : previous.keySet()) {
			merged.remove(key);
		}
	}

	/* (non-Javadoc)
	 * @see java.util.Map#get(java.lang.Object)
	 */
	@Override
	public V get(final Object key) {
		return merged.get(key);
	}

	/* (non-Javadoc)
	 * @see java.util.Map#isEmpty()
	 */
	@Override
	public boolean isEmpty() {
		return merged.isEmpty();
	}

	/* (non-Javadoc)
	 * @see java.util.Map#keySet()
	 */
	@Override
	public Set<K> keySet() {
		return merged.keySet();
	}

	/* (non-Javadoc)
	 * @see java.util.Map#put(java.lang.Object, java.lang.Object)
	 */
	@Override
	public V put(final K key, final V value) {
		if (previous.containsKey(key)) {
			previous.remove(key);
		}
		return merged.put(key, value);
	}

	/* (non-Javadoc)
	 * @see java.util.Map#putAll(java.util.Map)
	 */
	@Override
	public void putAll(final Map<? extends K, ? extends V> t) {
		for (final K key : t.keySet()) {
			if (previous.containsKey(key)) {
				previous.remove(key);
			}
		}
		merged.putAll(t);
	}

	/* (non-Javadoc)
	 * @see java.util.Map#remove(java.lang.Object)
	 */
	@Override
	public V remove(final Object key) {
		previous.remove(key);
		return merged.remove(key);
	}

	/**
	 * Renew all entries to survive the next flash.
	 */
	public void renew() {
		previous.clear();
	}

	/**
	 * Renew an entry to survive the next flash.
	 * 
	 * @param key the key
	 */
	public void renew(final K key) {
		if (previous.containsKey(key)) {
			previous.remove(key);
		}
	}

	/* (non-Javadoc)
	 * @see java.util.Map#size()
	 */
	@Override
	public int size() {
		return merged.size();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return merged.toString();
	}

	/* (non-Javadoc)
	 * @see java.util.Map#values()
	 */
	@Override
	public Collection<V> values() {
		return merged.values();
	}
}
