/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.util.LinkedList;



/**
 * The Class FlashMessages. You can get access to the flash messages by {@link Context#getFlashMessages()}. Use the
 * FlashMessages to add any messages that should exists till the next request.
 */
public class FlashMessages {
	private static final String ERROR_LIST = "org.jsxp.framework.framework.flashmessages.errorList";
	private static final String MESSAGE_LIST = "org.jsxp.framework.framework.flashmessages.messageList";
	private static final String WARNING_LIST = "org.jsxp.framework.framework.flashmessages.warningList";
	private final Context context;

	/**
	 * Instantiates a new flash messages.
	 * 
	 * @param context the context
	 */
	FlashMessages(final Context context) {
		this.context = context;
	}

	/**
	 * Adds the error message.
	 * 
	 * @param message the message
	 */
	@SuppressWarnings("unchecked")
	public void addErrorMessage(final String message) {
		LinkedList<String> errorList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(ERROR_LIST);
		if(errorList == null) {
			errorList = new LinkedList<String>();
			context.getScopes().getFlashAttributes().put(ERROR_LIST, errorList);
		}
		errorList.add(message);
	}

	/**
	 * Adds the information message.
	 * 
	 * @param message the message
	 */
	@SuppressWarnings("unchecked")
	public void addInformationMessage(final String message) {
		LinkedList<String> messageList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(MESSAGE_LIST);
		if(messageList == null) {
			messageList = new LinkedList<String>();
			context.getScopes().getFlashAttributes().put(MESSAGE_LIST, messageList);
		}
		messageList.add(message);
	}

	/**
	 * Adds the warning message.
	 * 
	 * @param message the message
	 */
	@SuppressWarnings("unchecked")
	public void addWarningMessage(final String message) {
		LinkedList<String> warningList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(WARNING_LIST);
		if(warningList == null) {
			warningList = new LinkedList<String>();
			context.getScopes().getFlashAttributes().put(WARNING_LIST, warningList);
		}
		warningList.add(message);
	}

	/**
	 * Clear lists.
	 */
	@SuppressWarnings("unchecked")
	public void clearLists() {
		final LinkedList<String> messageList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(MESSAGE_LIST);
		final LinkedList<String> warningList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(WARNING_LIST);
		final LinkedList<String> errorList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(ERROR_LIST);

		if(messageList != null) {
			messageList.clear();
		}

		if(warningList != null) {
			warningList.clear();
		}

		if(errorList != null) {
			errorList.clear();
		}
	}

	/**
	 * Gets the error messages.
	 * 
	 * @return the error messages
	 */
	@SuppressWarnings("unchecked")
	public LinkedList<String> getErrorMessages() {
		return (LinkedList<String>) context.getScopes().getFlashAttributes().get(ERROR_LIST);
	}

	/**
	 * Gets the info messages.
	 * 
	 * @return the info messages
	 */
	@SuppressWarnings("unchecked")
	public LinkedList<String> getInfoMessages() {
		return (LinkedList<String>) context.getScopes().getFlashAttributes().get(MESSAGE_LIST);
	}

	/**
	 * Gets the warning messages.
	 * 
	 * @return the warning messages
	 */
	@SuppressWarnings("unchecked")
	public LinkedList<String> getWarningMessages() {
		return (LinkedList<String>) context.getScopes().getFlashAttributes().get(WARNING_LIST);
	}

	/**
	 * Checks if is empty.
	 * 
	 * @return true, if is empty
	 */
	@SuppressWarnings("unchecked")
	public boolean isEmpty() {
		final LinkedList<String> messageList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(MESSAGE_LIST);
		final LinkedList<String> warningList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(WARNING_LIST);
		final LinkedList<String> errorList = (LinkedList<String>) context.getScopes().getFlashAttributes().get(ERROR_LIST);

		boolean messageListEmpty = false;
		boolean warningListEmpty = false;
		boolean errorListEmpty = false;

		if(messageList == null || messageList.isEmpty()) {
			messageListEmpty = true;
		}
		if(warningList == null || warningList.isEmpty()) {
			warningListEmpty = true;
		}
		if(errorList == null || errorList.isEmpty()) {
			errorListEmpty = true;
		}
		return messageListEmpty && warningListEmpty && errorListEmpty;
	}

}
