/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Locale;


/**
 * The Class UserContext. Default implementation of a user context.
 * 
 * The context is automatically instantiated by the framework. You can access it with
 * {@link Context#getUserContext()}. Do not create an instance yourself.
 * 
 * @see Context
 */
public class UserContext implements Serializable {
	private static final long serialVersionUID = 1L;

	/**
	 * Instantiates a new user context.
	 */
	public UserContext() {
	}

	/**
	 * Gets the preferred user locale (from the request).
	 * 
	 * @return the preferred user locale
	 */
	public Locale getPreferredUserLocale() {
		return Context.getContext().getHttpServletRequest().getLocale();
	}

	/**
	 * Gets the preferred user locales.
	 * 
	 * @return the preferred user locales
	 */
	@SuppressWarnings("unchecked")
	public Locale[] getPreferredUserLocales() {
		final LinkedList<Locale> locales = new LinkedList<Locale>();
		final Enumeration<Locale> localesEnum = Context.getContext().getHttpServletRequest().getLocales();
		while (localesEnum.hasMoreElements()) {
			final Locale locale = localesEnum.nextElement();
			locales.add(locale);
		}

		return locales.toArray(new Locale[]{});
	}

	/**
	 * Checks if is allowed to view page. Normally called from the default implementation
	 * of {@link Application#isAllowedToViewPage(ViewController, String)}
	 * 
	 * @param view the view (can be null if only a uri is checked)
	 * @param uri the uri
	 * 
	 * @return true, if is allowed to view page
	 */
	public boolean isAllowedToViewPage(final ViewController view, final String uri) {
		return true;
	}
}
