/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.util.LinkedList;

import org.jsxp.framework.elementprocessing.ElementProcessor;




/**
 * The Class Validator. You need it, if you use
 * the jsxp frameworks basic validation features to validate input parameter.
 * You can create a Validator-instance anywhere.
 * Use the {@link ViewController#createValidator()} method. The standard way is to use it
 * in your {@link ViewController#validate()} method.
 * 
 * @see ValidationError
 */
public class Validator {

	/** The Constant VALIDATION_ERROR_LIST. */
	protected static final String VALIDATION_ERROR_LIST = "org.jsxp.framework.framework.validation.errorList";

	private boolean hasErrors;
	private final ViewController ViewController;

	/**
	 * Instantiates a new validator.
	 * 
	 * @param ViewController the view controller
	 */
	public Validator(final ViewController ViewController) {
		this.ViewController = ViewController;
	}

	/**
	 * Adds the validation error. If you add an error to the validator, the corresponding
	 * view controller will not be executed by calling {@link ViewController#skipExecution()}.
	 * Also the the ViewController will be redirected to {@link Context#getLastView()}. And the method
	 * {@link Context#saveInputToRestoreLater()} will be called, so that the next view controller
	 * can get the input parameters even after a redirect by implementing {@link ViewController#getPreviousInputParameterBean()}.
	 * 
	 * @param message
	 * @param elementProcessors
	 */
	@SuppressWarnings("unchecked")
	public void addValidationError(final String message, final ElementProcessor... elementProcessors) {
		LinkedList<ValidationError> errorsList = (LinkedList<ValidationError>) Context.getContext().getScopes().getFlashAttributes().get(VALIDATION_ERROR_LIST);
		if(errorsList == null) {
			errorsList = new LinkedList<ValidationError>();
			Context.getContext().getScopes().getFlashAttributes().put(VALIDATION_ERROR_LIST, errorsList);
		}
		errorsList.add(new ValidationError(message, elementProcessors));

		if (!hasErrors) {
			hasErrors=true;
			ViewController.skipExecution();
			if (Context.getContext().getLastView() != null) {
				Context.getContext().redirect(Context.getContext().getLastView(), true);
			}

			//After the re-direct the input values should be restored in the previous view
			Context.getContext().saveInputToRestoreLater();
		}
	}

	/**
	 * Checks for errors.
	 * 
	 * @return true, if successful
	 */
	public boolean hasErrors() {
		return hasErrors;
	}

	/**
	 * Validates a condition and add an validation error in one step
	 * 
	 * @param condition
	 * @param message
	 * @param elementProcessors
	 */
	public void validate(final boolean condition, final String message, final ElementProcessor... elementProcessors) {
		if(!condition) {
			addValidationError(message, elementProcessors);
		}
	}

}
