/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.jsxp.framework.ViewControllerLifeCycle.LifeCyclePhase;
import org.jsxp.framework.elementprocessing.ElementProcessor;
import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.internal.ProxyElement;
import org.jsxp.framework.logging.Log;
import org.jsxp.framework.util.text.StringUtil;
import org.jsxp.framework.webroot.jsxpframework.components.AbstractCommandController;


/**
 * The Class ViewController. A ViewController contains the code for any jsxp component. It has method
 * to be initialized, to validate input parameter, to define the result and so on.
 * <br/><br/>
 * Mostly it has
 * a corresponding view design file (e.g. index.xhtml), for an exclusion see {@link AbstractCommandController}.
 * The ViewController is called by the jsxp framework if the corresponding view is requested.
 * <br/><br/>
 * Normally a ViewController class will be generated by the framework for each view design file.
 * You have to derive this generated class to enter your code. The generated file will have
 * special method to set view variables and to get access to identified jsxp element of the view.
 * These methods help you to get compile time safety.
 * <br/><br/>
 * You can also derive your own ViewController classes to build specialized ones or to make a
 * new component based on an existing.
 * <br/><br/>
 * The ViewController follows a {@link ViewControllerLifeCycle} for each request. You can get access
 * to the life cycle by {@link ViewController#getLifeCycle()} and you can add a {@link PhaseListener}
 * to get more control by {@link ViewControllerLifeCycle#addPhaseListener(LifeCyclePhase, PhaseListener)}.
 * The possibles phases are listed in this enum: {@link LifeCyclePhase}.
 */
public class ViewController implements ViewControllerLifeCycleExceptionHandler {
	private boolean applicationTemplating=true;
	private final List<ElementProcessor> elementProcessors = new LinkedList<ElementProcessor>();
	private final List<Element> elements = new LinkedList<Element>();

	public Map<String, InternalElement> idElements = null;
	private String idPrefix = null;
	private final Map<String, ViewController> idPrefixViewControllers = new HashMap<String, ViewController>();
	private List<ViewController> importedViews = null;
	private ViewControllerLifeCycle lifeCycle = null;
	private boolean needToSaveScopesToSession=true;
	private ViewController parent = null;

	private InternalElement rootElement = null;
	private boolean skipExecution=false;
	private ViewController templatedView;
	private boolean usedAsImportedView = false;
	private Validator validator;
	private final Map<String, String> variableValues = new HashMap<String, String>();
	private String viewFilename = null;
	private ViewController viewTemplate = null;

	/**
	 * Instantiates a new view controller.
	 * 
	 * @param viewFilename the view filename
	 */
	public ViewController(final String viewFilename) {
		this.viewFilename = viewFilename;
		this.importedViews = new LinkedList<ViewController>();
		this.lifeCycle = new ViewControllerLifeCycle(this);
		getLifeCycle().finishPhase(LifeCyclePhase.CREATED);
	}

	/**
	 * Adds the element.
	 * 
	 * @param element the element
	 */
	public final void addElement(final Element element) {
		elements.add(element);
		if (element.getId() != null) {
			if (idElements.containsKey(element.getId())) {
				throw new DuplicateIdException("id '"+element.getId()+"' already definded");
			}
			idElements.put(element.getId(), (InternalElement)element);
		}
	}

	/**
	 * Adds an element processor to this view controller. Can be used in any phase before
	 * the phase {@link ViewControllerLifeCycle.LifeCyclePhase#ELEMENTPROCESSING}. The various {@link ElementProcessor}
	 * can add special behaviour to elements (like Ajax behaviour or change of design).
	 * @param elementProcessor
	 */
	public final void addElementProcessor(final ElementProcessor elementProcessor) {
		elementProcessors.add(elementProcessor);
	}

	private void checkElementTree() {
		final Method[] methods = this.getClass().getMethods();
		for(final Method m : methods) {
			if(m.getName().startsWith("get")  && m.getReturnType().equals(Element.class)
					&& m.getParameterTypes().length==0) {
				final Annotation[] annotations = m.getAnnotations();
				for(final Annotation a : annotations) {
					if(a instanceof org.jsxp.framework.annotations.Element) {
						Element result;
						try {
							result = (Element) m.invoke(this, (Object[])null);
						} catch (final Exception e) {
							throw new RuntimeException("could not access element '"+m.getName().substring(3)+"' from view '"+this+"("+ViewController.class+")", e);
						}
						if (result == null) {
							throw new RuntimeException("element '"+m.getName().substring(3)+"' from view '"+this+"("+ViewController.class+") not found in element tree");
						}
					}
				}
			}
		}
	}

	/**
	 * Check if the view contains the variable.
	 * 
	 * @param key the key
	 * 
	 * @return true, if successful
	 */
	public final boolean containsVariable(final String key) {
		return variableValues.containsKey(key);
	}

	/**
	 * Creates a validator that you can use in the view controller.
	 * 
	 * @return the validator
	 */
	private Validator createValidator() {
		return new Validator(this);
	}

	/**
	 * Gets the {@link ElementProcessor}s (including the ones from imported ViewController).
	 * Add a {@link ElementProcessor} in the view controller
	 * in any phase before the phase {@link ViewControllerLifeCycle.LifeCyclePhase#ELEMENTPROCESSING}
	 * with {@link #addResource(ViewResource)}. Than these processors will be processed in that phase.
	 * 
	 * @return the resources
	 */
	final List<ElementProcessor> elementProcessors() {
		final List<ElementProcessor> elementProcessors = getImportedElementProcessors();
		elementProcessors.addAll(this.elementProcessors);
		return elementProcessors;
	}

	/**
	 * This method will be called by the jsxp framework in the execution phase. It will only
	 * be called for directly requested ViewController, not for template and not for imported
	 * ViewController. Therefore you have to call their {@link #execute()} method explicitly if
	 * you want them also to be called. But recommended is to implemented code that should be executed
	 * also if the ViewController is used indirectly when imported in the {@link #init()} method.
	 * <br/><br/>
	 * In this method you should call your business logic.
	 * 
	 * @throws Exception the exception
	 */
	public void execute() throws Exception {}

	/**
	 * Gets the character encoding from the {@link Application#getDefaultCharacterEncoding()}. Override this method if your view has a special character
	 * encoding. Default is UTF-8.
	 * 
	 * @return the character encoding
	 */
	public String getCharacterEncoding() {
		return Application.getApplication().getDefaultCharacterEncoding();
	}

	/**
	 * Gets the content type from the {@link Application#getDefaultContentType()}. Override this
	 * method to get return a special content type.
	 * 
	 * @return the content type
	 */
	public String getContentType() {
		return Application.getApplication().getDefaultContentType();
	}

	/**
	 * Returns the element with the jsxp id.
	 * 
	 * @param id
	 * @return the corresponding element
	 */
	public final Element getElement(final String id) {
		return getElement(null, id, -1);
	}

	/**
	 * Gets the element with the jsxp id. Mostly used by the automatically generated view controller classes.
	 * 
	 * @param id the id
	 * @param interfaceType the interface type
	 * 
	 * @return the corresponding element
	 */
	@SuppressWarnings("unchecked")
	protected final <T extends Element> T getElement(final String id, final Class<T> interfaceType) {
		final InternalElement elem = (InternalElement) getElement(id);

		final ElementProxyInvocationhandler handler = new ElementProxyInvocationhandler(elem, interfaceType);

		return (T) Proxy.newProxyInstance(getClass().getClassLoader(), new Class[] {interfaceType, ProxyElement.class}, handler);
	}

	/**
	 * Returns the element with the jsxp id and index (used if elements are inserted via element templating).
	 * @param id
	 * @param index
	 * @return the corresponding element
	 */
	public final Element getElement(final String id, final int index) {
		return getElement(null, id, index);
	}

	/**
	 * Returns the element with the jsxp id prefix and id
	 * @param idPrefix
	 * @param id
	 * @return the corresponding element
	 */
	public final Element getElement(final String idPrefix, final String id) {
		return getElement(idPrefix, id, -1);
	}

	/**
	 * Returns the element with the jsxp id prefix and id and index (used if elements are inserted via element templating).
	 * @param idPrefix
	 * @param id
	 * @param index
	 * @return the corresponding element
	 */
	public final Element getElement(final String idPrefix, final String id, final int index) {
		if (index >= 0) {
			return idElements.get(StringUtil.concateStrings(null, getIdPrefix(), idPrefix,id, "$", ""+index));
		} else {
			return idElements.get(StringUtil.concateStrings(null, getIdPrefix(), idPrefix,id));
		}
	}

	/**
	 * Gets the id prefix. Override this method if you need to use a special idPrefix for
	 * your jsxp ids (you will see them in the rendered results). Normally you don�t need
	 * to do this.
	 * 
	 * @return the id prefix
	 */
	public String getIdPrefix() {
		return idPrefix;
	}

	/**
	 * Returns the {@link ElementProcessor}s of all imported views by calling there
	 * {@link #elementProcessors()} method.
	 * @return
	 */
	private final List<ElementProcessor> getImportedElementProcessors() {
		final List<ElementProcessor> importedElementProcessors = new LinkedList<ElementProcessor>();

		for(final ViewController view : getImportedViews()) {
			importedElementProcessors.addAll(view.elementProcessors());
		}

		return importedElementProcessors;
	}

	/**
	 * Gets the imported views.
	 * 
	 * @return the imported views
	 */
	final List<ViewController> getImportedViews() {
		return importedViews;
	}

	/**
	 * Gets an instance of the input parameter bean. You have to override this method to use a
	 * special input parameter class. The jsxp framework will
	 * call this method to get a fresh instance and fill the bean by calling the set-method
	 * on the beans properties.
	 * 
	 * @return the input parameter
	 */
	public Object getInputParameterBean() {
		return null;
	}

	/**
	 * Gets the life cycle.
	 * 
	 * @return the life cycle
	 */
	public final ViewControllerLifeCycle getLifeCycle() {
		return lifeCycle;
	}

	/**
	 * Gets the parent ViewController, e.g. when a view controller is used as a sub component from
	 * another view controller.
	 * 
	 * @return the parent
	 */
	protected final ViewController getParent() {
		return parent;
	}

	/**
	 * Gets an instance of the previous input parameter bean. As default
	 * this returns null.
	 * You can override this method
	 * if you want to support the restoration of previous input parameter in a
	 * different bean.
	 *  The framework will
	 * call this method to get a fresh instance and fill the bean by calling the set-method
	 * on the beans properties. Use that value in your execute method.
	 * 
	 * @return the previous input parameter bean
	 */
	public Object getPreviousInputParameterBean() {
		return null;
	}

	/**
	 * Gets the root element. This is the top most jsxp element of the view (e.g. the corresponding
	 * element to the html-tag in HTML).
	 * 
	 * @return the root element
	 */
	public final Element getRootElement() {
		return rootElement;
	}

	/**
	 * Gets the status code. Override this method if you want to change the status code.
	 * 
	 * @return the status code
	 */
	public int getStatusCode() {
		return 200;
	}

	/**
	 * Gets the templated view if this view controller is use as a template.
	 * 
	 * @return the templated view
	 */
	protected final ViewController getTemplatedView() {
		return templatedView;
	}

	/**
	 * Gets the template elements. Override this method if your ViewController should be used
	 * as a template. Return the elements, that will be exchange by elements with same jsxp id
	 * from the templated view.
	 * 
	 * @return the template elements
	 */
	protected Element[] getTemplateElements() {
		return null;
	}

	/**
	 * Gets the validator that you can use in the view controller..
	 * 
	 * @return the validator
	 */
	protected final Validator getValidator() {
		if(validator == null) {
			validator = createValidator();
		}
		return validator;
	}

	/**
	 * Gets the value for a view variable.
	 * 
	 * @param key the key
	 * 
	 * @return the value for variable
	 */
	public final String getValueForVariable(final String key) {
		return variableValues.get(key);
	}

	/**
	 * Gets the variables as key/value map.
	 * 
	 * @return the variables
	 */
	public Map<String, String> getVariables() {
		return variableValues;
	}

	/**
	 * Gets the view filename. The view filename is the name of the corresponding view design file.
	 * 
	 * @return the view filename
	 */
	public final String getViewFilename() {
		return viewFilename;
	}

	/**
	 * Gets the view template.
	 * 
	 * @return the view template
	 */
	public final ViewController getViewTemplate() {
		return viewTemplate;
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * 
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleExecutingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleExecutingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleInitException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleInitException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleInputProcessingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleInputProcessingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handlePreparingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handlePreparingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleRenderingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleRenderingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleTemplatingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleTemplatingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}

	/** Throws the given exception encapsulated in a runtime exception per default.
	 * @see org.jsxp.framework.ViewControllerLifeCycleExceptionHandler#handleValidatingException(java.lang.Exception, org.jsxp.framework.ViewController)
	 */
	@Override
	public void handleValidatingException(final Exception e, final ViewController source) {
		throw new RuntimeException(e);
	}
	/**
	 * Import view.
	 * 
	 * @param ViewController the view control
	 * 
	 * @throws Exception the exception
	 */
	public final <T extends ViewController> T importView(final T ViewController) throws Exception {
		return importView(ViewController, null, false, null);
	}

	/**
	 * Import view.
	 * 
	 * @param ViewController the view control
	 * @param idPrefix the id prefix
	 * @param inputParameterInjection the input parameter injection
	 * @param inputParameterPrefix the input parameter prefix
	 * 
	 * @throws Exception the exception
	 */
	public final <T extends ViewController> T importView(final T ViewController, String idPrefix, final boolean inputParameterInjection, final String inputParameterPrefix) throws Exception {
		if (idPrefix == null) {
			idPrefix = ViewController.toString();
		}

		if (idPrefixViewControllers.containsKey(idPrefix)) {
			final ViewController v = idPrefixViewControllers.get(idPrefix);
			throw new DuplicateIdException("already imported view control '"+v+"' with idPrefix '"+idPrefix+"' in view '"+this);
		}

		new ViewControllerLifeCycleManager().processLifecycleImportedView(inputParameterPrefix, inputParameterInjection, ViewController, this, idPrefix);

		importedViews.add(ViewController);
		idPrefixViewControllers.put(idPrefix, ViewController);

		Context.getContext().addDebugMessage("import view '"+ViewController.getViewFilename()+"("+ViewController+")' to view '"+this.getViewFilename()+"("+this+")'");
		Log.debug(this, "import view '"+ViewController.getViewFilename()+"("+ViewController+")' to view '"+this.getViewFilename()+"("+this+")'");

		ViewController.setUsedAsImportedView(true);

		return ViewController;
	}

	/**
	 * This method will be called by the jsxp framework in the initializing phase. It will
	 * be called for any used ViewController, also for template and for imported
	 * ViewController.
	 * <br/><br/>
	 * In this method you should do anything that a view controller needs to do if he is used
	 * in any way.
	 * 
	 * @throws Exception the exception
	 */
	protected void init() throws Exception {
	}

	/**
	 * Configures if application template is enabled. Override it to disable the use of application template
	 * or use the {@link #setApplicationTemplateEnabled(boolean)} method.
	 * 
	 * @return true, if global templating is enabled for this view.
	 */
	protected boolean isApplicationTemplateEnabled() {
		return applicationTemplating;
	}

	/**
	 * Checks if is skip execution.
	 * 
	 * @return true, if is skip execution
	 */
	protected final boolean isSkipExecution() {
		return skipExecution;
	}

	/**
	 * Checks if is used as imported view.
	 * 
	 * @return true, if is used as imported view
	 */
	public final boolean isUsedAsImportedView() {
		return usedAsImportedView;
	}


	/**
	 * Checks if is used as template.
	 * 
	 * @return true, if is used as template
	 */
	public final boolean isUsedAsTemplate() {
		return templatedView != null;
	}



	/**
	 * Need to save scopes to session. Return true if this or any imported view controller
	 * has called {@link #setNeedToSaveScopesToSession(boolean)} with true as parameter.
	 * This has the effect that the jsxp framework will create a session and save the scopes
	 * in it, so you can access them later.
	 * 
	 * @return true, if successful
	 */
	public final boolean needToSaveScopesToSession() {
		if (this.needToSaveScopesToSession) {
			return true;
		}

		for(final ViewController importedView : importedViews) {
			if (importedView.needToSaveScopesToSession()) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Prepare.
	 * 
	 * @throws Exception the exception
	 */
	final void prepare() throws Exception {
		prepare(null, null);
	}

	/**
	 * Prepare.
	 * 
	 * @param idPrefix the id prefix
	 * 
	 * @throws Exception the exception
	 */
	final void prepare(final String idPrefix) throws Exception {
		prepare(null, idPrefix);
	}

	/**
	 * Prepare.
	 * 
	 * @param parent the parent
	 * @param idPrefix the id prefix
	 * 
	 * @throws Exception the exception
	 */
	final void prepare(final ViewController parent, final String idPrefix) throws Exception {
		Context.getContext().addDebugMessage("init view '"+this.viewFilename+"("+this+")'");
		this.idPrefix = idPrefix;
		this.parent = parent;
		idElements = new HashMap<String, InternalElement>();
		if (parent != null) {
			rootElement = Application.getApplication().getViewParser().read(viewFilename, Application.getApplication(), this, idPrefix, idElements,
					(InternalElement)parent.getRootElement(), this);
		} else {
			rootElement = Application.getApplication().getViewParser().read(viewFilename, Application.getApplication(), this, idPrefix, idElements, null, this);
		}
		checkElementTree();
	}

	/**
	 * Replace variable value. Mostly used by the generated ViewController classes.
	 * 
	 * @param variableName the variable name
	 * @param newValue the new value
	 */
	protected final void replaceVariableValue(final String variableName, final String newValue) {
		variableValues.put(variableName, newValue);
	}

	public final void setApplicationTemplateEnabled(final boolean applicationTemplating) {
		this.applicationTemplating = applicationTemplating;
	}

	/**
	 * Sets the need to save scopes to session. Even if you set this to false it
	 * can be that a session is created and the scopes are saved. E.g. if an
	 * imported ViewController has the need or if this view controller is
	 * part of a view flow.
	 * 
	 * @param needToSaveScopesToSession the new need to save scopes to session
	 */
	public final void setNeedToSaveScopesToSession(final boolean needToSaveScopesToSession) {
		this.needToSaveScopesToSession = needToSaveScopesToSession;
	}

	/**
	 * Sets the root element.
	 * 
	 * @param rootElement the new root element
	 */
	final void setRootElement(final InternalElement rootElement) {
		this.rootElement = rootElement;
	}

	/**
	 * Sets the used as imported view.
	 * 
	 * @param usedAsImportedView the new used as imported view
	 */
	final void setUsedAsImportedView(final boolean usedAsImportedView) {
		this.usedAsImportedView = usedAsImportedView;
	}

	/**
	 * Sets the variable.
	 * 
	 * @param variableName the variable name
	 * @param value the value
	 */
	public final void setVariable(final String variableName, final String value) {
		replaceVariableValue(variableName, value);
	}

	/**
	 * Sets the view template.
	 * 
	 * @param viewTemplate the new view template
	 */
	public final <ViewTemplateInputParameterType> void setViewTemplate(final ViewController viewTemplate) {
		this.viewTemplate = viewTemplate;
		if (viewTemplate != null) {
			viewTemplate.templatedView = this;
		}
	}

	/**
	 * Skip execution. Marks the ViewController to skip the execution phase.
	 */
	protected final void skipExecution() {
		skipExecution = true;
	}

	/**
	 * Substitute variables.
	 */
	final void substituteVariables() {
		for (final Element element : elements) {
			((InternalElement)element).substituteVariables();
		}
	}

	/**
	 * This method will be called by the jsxp framework in the validation lifecycle phase.
	 * Implement this method to validate the input parameter. You can use any validation framework
	 * you like in this method or use {@link Validator} if you want the simple jsxp validation support.
	 * 
	 * @throws Exception the exception
	 */
	public void validate() throws Exception {}

}
