/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.logging.Log;
import org.jsxp.framework.resources.ViewResourceResolver;
import org.jsxp.framework.util.text.StringUtil;
import org.jsxp.framework.util.text.StringUtils;


/**
 * This generator will parse and process the view design files and generate
 * corresponding {@link ViewController} classes.
 */
public class ViewControllerGenerator {

	private static final HashSet<String> ALLOWED_TK_IDS = new HashSet<String>();
	/** The Constant GENERATED_CLASS_SUFFIX. */
	public static final String GENERATED_CLASS_SUFFIX = "ControllerGenerated";

	private static HashSet<String> generatedFiles = new HashSet<String>();
	private static HashSet<String> generatedSetters = new HashSet<String>();
	/** The Constant JSXP_ELEMENT_ID_FIELD_NAME. */
	public static final String JSXP_ELEMENT_ID_FIELD_NAME = "JSXP_ELEMENT_ID";
	private static String[] patterns;
	private static String[] translations;
	private static File webFileDirectory;

	static {
		ALLOWED_TK_IDS.add("jsxp:developer_warnings");
		ALLOWED_TK_IDS.add("jsxp:debug_messages");
		ALLOWED_TK_IDS.add("jsxp:flash_messages");
	}

	private static boolean checkMatch(final String[] pathElems, final String pattern) {
		final String[] patternElements = pattern.split("/");
		int i=0;
		for(int j=0; j<patternElements.length; j++) {
			String elem = patternElements[j];
			//			System.out.println("    -> Pattern element: "+elem);
			if(elem.equals("**")) {
				//Skip all directories that do not match the next pattern element

				if(j==patternElements.length-1) {
					//When the pattern ends with "**" (any directory), it never matches any files.
					return false;
				}
				j++;
				elem = patternElements[j];
				final String regex = elem.replaceAll("\\.", "\\\\.").replaceAll("\\*", ".*");
				//				System.out.println("    -> Regex="+regex);
				while(i < pathElems.length-1 && !pathElems[i].matches(regex)) {
					i++;
				}
			}
			//			System.out.println("    -> Checking if path element "+pathElems[i]+" matches pattern: "+elem);
			final String regex = elem.replaceAll("\\.", "\\\\.").replaceAll("\\*", ".*");
			//			System.out.println("    -> Regex="+regex);
			if(pathElems[i].matches(regex)) {
				//				System.out.println("      -> Matches!");
				if(i==pathElems.length-1) {
					if(j==patternElements.length-1) {
						return true;
					} else {
						return false;
					}
				} else {
					i++;
				}
			} else {
				return false;
			}
		}
		return false;
	}

	private static boolean fileMatchesPattern(final File f) {
		final String absolutePath = f.getAbsolutePath();
		if(!absolutePath.startsWith(webFileDirectory.getAbsolutePath())) {
			throw new IllegalStateException("File path \""+f.getAbsolutePath()+"\" does not start with base path \""+webFileDirectory.getAbsolutePath()+"\".");
		}
		String relativePath = f.getAbsolutePath().substring(webFileDirectory.getAbsolutePath().length());
		if(relativePath.startsWith("/")) {
			relativePath = relativePath.substring(1);
		}
		//		System.out.println("Checking if relative path matches patterns: "+relativePath);
		final String[] rPathElems = relativePath.split("/");

		for(final String pattern : patterns) {
			//			System.out.println(" -> Checking pattern match: "+pattern);
			if(checkMatch(rPathElems, pattern)) {
				return true;
			}
		}

		return false;
	}

	private static void findFilesAndGenerateControllers(final File webFileDirectory, final File generatedContentDirectory, final String basePackageName, final String relativHtmlDirectory, final boolean firstCall) {
		final File[] files = webFileDirectory.listFiles();

		if(files != null) {
			for(final File f : files) {
				if(f.isDirectory()) {
					final String dirName = f.getName();
					boolean ignore = false;
					if(firstCall && translations!=null) {
						for(final String s : translations) {
							if(s.equals(dirName)) {
								ignore = true;
								break;
							}
						}
					}
					if(!ignore && !("WEB-INF".equals(dirName) || "META-INF".equals(dirName))) {
						final File generatorDir = new File(generatedContentDirectory, StringUtil.getJavaPackageName(dirName));
						final String newPackage = basePackageName+"."+StringUtil.getJavaPackageName(dirName);
						String newRelativeDirectory = StringUtil.getJavaPackageName(dirName);
						if(relativHtmlDirectory != null) {
							newRelativeDirectory = relativHtmlDirectory+"/"+StringUtil.getJavaPackageName(dirName);
						}
						findFilesAndGenerateControllers(f, generatorDir, newPackage, newRelativeDirectory, false);
					}
				} else if(fileMatchesPattern(f)) {
					String htmlFileName = f.getName();
					if(relativHtmlDirectory != null) {
						htmlFileName = relativHtmlDirectory+"/"+f.getName();
					}
					generateViewControllerClass(f, generatedContentDirectory, basePackageName, htmlFileName);
				}
			}
		}
	}

	private static void generateElementInterface(final String elementId, final InternalElement elem, final PrintWriter w, final Map<String, InternalElement> idElements) {
		final String javaInterfaceName = StringUtil.getJavaPropertyName(elementId);
		final String accessModifier = getAccessModifier(elem);

		w.print("\t");
		w.print(accessModifier);
		w.print(" static interface ");
		w.print(javaInterfaceName);
		w.println(" extends Element {");
		w.println("\tpublic static final String "+JSXP_ELEMENT_ID_FIELD_NAME+"=\""+elementId+"\";");

		w.println("\t\tpublic org.jsxp.framework.ElementTemplate<"+javaInterfaceName+"> createTemplate();");
		w.println("\t\tpublic org.jsxp.framework.ElementTemplate<"+javaInterfaceName+"> createTemplate(String idPrefix);");
		w.println("\t\tpublic org.jsxp.framework.ElementTemplate<"+javaInterfaceName+"> useAsTemplateOnly();");
		w.println("\t\tpublic org.jsxp.framework.ElementTemplate<"+javaInterfaceName+"> useAsTemplateOnly(String idPrefix);");
		for (final String id : idElements.keySet()) {
			if(!id.startsWith("jsxp:")) {
				if(isChildOf(idElements.get(id), elem)) {
					boolean isPublic = true;
					if (elem.hasAttribute("http://www.jsxp.org/spec/1.0", "public")) {
						final String publicAttr = elem.getAttribute("http://www.jsxp.org/spec/1.0", "public");
						if (publicAttr.equals("true")) {
							isPublic = true;
						} else if (publicAttr.equals("false")) {
							isPublic = false;
						} else {
							throw new RuntimeException("unknown attribute value '"+publicAttr+"' for attribute 'public' in namespace '"+"http://www.jsxp.org/spec/1.0"+"'");
						}
					}

					if(isPublic) {
						final String javaPropertyName = StringUtil.getJavaPropertyName(id);
						w.println("\t\t@org.jsxp.framework.annotations.Element");
						w.println("\t\t@org.jsxp.framework.annotations.ElementGetter(name=\""+id+"\")");
						w.println("\t\tpublic Element getElement"+javaPropertyName+"();");
					}
				}
			}
		}
		generateSettersForVariables(elem, elem, w, elem, "\t", false, idElements);
		generatedSetters.clear();

		w.println("\t}");
		w.println();
	}

	private static void generateErrorFile(final File javaClassFile, final Exception e, final String htmlFileName) {
		try {
			final PrintWriter w = new PrintWriter(new FileWriter(javaClassFile));
			w.println("Could not generate view controller for file "+htmlFileName+":");
			w.println(e.getMessage());
			e.printStackTrace(w);
			w.close();
		} catch (final IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
	}

	private static void generateJavaCode(final InternalElement rootElement, final Map<String, InternalElement> idElements, final String javaClassName,
			final File javaClassFile, final String basePackageName, final String htmlFileName) {
		PrintWriter w;
		try {
			w = new PrintWriter(new FileWriter(javaClassFile));

			w.println("package "+basePackageName+";");
			w.println();

			w.println("import org.jsxp.framework.ViewController;");
			int elementCount = 0;
			for (final String id : idElements.keySet()) {
				if(!id.startsWith("jsxp:")) {
					elementCount++;
				}
			}
			if (elementCount>0) {
				w.println("import org.jsxp.framework.Element;");
			}
			w.println();

			w.println("public class "+javaClassName+" extends ViewController {");

			for (final String id : idElements.keySet()) {
				if(!id.startsWith("jsxp:")) {
					final InternalElement elem = idElements.get(id);
					generateElementInterface(id, elem, w, idElements);
				}
			}

			w.println("\tpublic "+javaClassName+"() {");
			w.println("\t\tsuper(\""+htmlFileName+"\");");
			w.println("\t}");

			w.println("\n");

			for (final String id : idElements.keySet()) {
				if(!id.startsWith("jsxp:")) {
					final InternalElement elem = idElements.get(id);
					final String javaName = StringUtil.getJavaPropertyName(id);
					w.println("\t@org.jsxp.framework.annotations.Element");
					w.print("\t");
					w.print(getAccessModifier(elem));
					w.println(" "+javaName+" getElement"+javaName+"() {");
					w.println("\t\treturn ("+javaName+") getElement(\""+id+"\", "+javaName+".class);");
					w.println("\t}");
					w.println();

					w.print("\t");
					w.print(getAccessModifier(elem));
					w.println(" static String getIdElement"+javaName+"() {");
					w.println("\t\treturn \""+id+"\";");
					w.println("\t}");
				} else {
					if(!ALLOWED_TK_IDS.contains(id)) {
						w.println("GENERATOR ERROR: The referenced jsxp ID \""+id+"\" is unknown.");
					}
				}
			}

			generateSettersForVariables(null, null, w, rootElement, "", true, idElements);

			w.println("}");

			w.close();
		} catch (final IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	private static void generateSetter(final Element genElement, final Element previousIdElement, final String attributeText, final PrintWriter w, final String prefix, final boolean createBody) {
		int previousIndex = 0;
		int index = attributeText.indexOf("$(");

		while(index != -1 && index >= previousIndex) {
			if(index == 0 || attributeText.charAt(index-1)!='$') {
				final int endIndex = attributeText.indexOf(")", index);
				if(endIndex > index) {
					String variableName = attributeText.substring(index+2, endIndex);
					final String fullName = variableName;
					String variableNamePrefix = null;
					if(variableName.indexOf(".") != -1) {
						variableNamePrefix=variableName.substring(0, variableName.indexOf("."));
						variableName = variableName.substring(variableName.indexOf(".")+1);
					}
					// * Generate the setter on the root view (genElement==null) if there is no
					//   previous ID element in the element tree, or if the prefix is available
					//   and empty.
					// * Generate the setter on the current element if the prefix matches the ID
					//   of genElement, or if there is no prefix and genElemen==previousIdElement
					if(genElement == null) {
						if(previousIdElement==null || "".equals(variableNamePrefix)) {
							generateSetterCode(w, prefix, createBody, variableName, fullName);
						}
					} else {
						if(!StringUtil.isEmpty(variableNamePrefix) && variableNamePrefix.equals(genElement.getId())) {
							generateSetterCode(w, prefix, createBody, variableName, fullName);
						} else if(variableNamePrefix==null && !StringUtil.isEmpty(genElement.getId()) && previousIdElement==genElement) {
							generateSetterCode(w, prefix, createBody, variableName, fullName);
						}
					}
					previousIndex = endIndex;
				} else {
					System.err.println("-> WARNING: Generating setters for variables: \"$(\" found, but there is no corresponding \")\" ("+attributeText+")");
					previousIndex = index;
				}
			} else {
				previousIndex = index;
			}
			index = attributeText.indexOf("$(", previousIndex+1);
		}
	}

	private static void generateSetterCode(final PrintWriter w, final String prefix,
			final boolean createBody, final String variableName, final String fullName) {
		final String propertyName = StringUtil.getJavaPropertyName(variableName);
		if(!generatedSetters.contains(propertyName)) {
			generatedSetters.add(propertyName);
			w.println(prefix+"\t@org.jsxp.framework.annotations.VariableSetter(name=\""+fullName+"\")");
			w.print(prefix+"\tpublic void setVariable"+propertyName+"(String value)");
			if(createBody) {
				w.println(" {");
				w.println(prefix+"\t\treplaceVariableValue(\""+fullName+"\", value);");
				w.println(prefix+"\t}");
				w.println();
			} else {
				w.println(";");
			}
		}
	}

	private static void generateSettersForVariables(final Element genElement, final Element previousIdElement, final PrintWriter w, final Element currentElement, final String prefix, final boolean createBody, final Map<String, InternalElement> idElements) {
		if(currentElement.isText()) {
			final String text = currentElement.getValue();
			generateSetter(genElement, previousIdElement, text, w, prefix, createBody);
		} else {
			final Map<String, String> attributes = currentElement.getAttributes();
			for(final String key : attributes.keySet()) {
				final String attributeText = attributes.get(key);
				generateSetter(genElement, previousIdElement, attributeText, w, prefix, createBody);
			}
			for(final Element next : currentElement.getElements()) {
				if(idElements.containsValue(next)) {
					generateSettersForVariables(genElement, next, w, next, prefix, createBody, idElements);
				} else {
					generateSettersForVariables(genElement, previousIdElement, w, next, prefix, createBody, idElements);
				}
			}
		}
	}

	private static void generateViewControllerClass(final File htmlFile, final File generatedContentDirectory, final String basePackageName, final String htmlFileName) {
		final String javaClassName = StringUtil.getJavaClassName(htmlFile.getName())+GENERATED_CLASS_SUFFIX;

		final File javaClassFile = new File(generatedContentDirectory, javaClassName+".java");
		generatedFiles.add(javaClassFile.getAbsolutePath());

		if(!javaClassFile.exists() || htmlFile.lastModified() > javaClassFile.lastModified()) {
			final Map<String, InternalElement> idElements = new HashMap<String, InternalElement>();
			final ViewParser p = Application.getApplication().getViewParser();
			System.out.println("Generating view controller for "+htmlFile.getAbsolutePath());

			if(!generatedContentDirectory.exists()) {
				generatedContentDirectory.mkdirs();
			}

			try {
				long lastModified = 0;
				try {
					lastModified = htmlFile.lastModified();
				} catch(final Exception e) {
					//We can safely ignore this exception, we really only wanted to know the time stamp of a file.
					//If there is no time stamp (i.e. the file was loaded from a jar file), we can cache it.
				}

				final InputStream inputStream = Application.getApplication().getResourceInputStream(htmlFile.getAbsolutePath(), null);

				final InternalElement rootElement = p.read(htmlFile.getAbsolutePath(), new ViewResourceResolver(lastModified, inputStream), p, null, idElements, null, null);
				generateJavaCode(rootElement, idElements, javaClassName, javaClassFile, basePackageName, htmlFileName);
			} catch(final Exception e) {
				generateErrorFile(javaClassFile, e, htmlFileName);
			} finally {
				generatedSetters.clear();
			}
		}
	}

	private static String getAccessModifier(final InternalElement elem) {
		String accessModifier = "public";
		if (elem.hasAttribute("http://www.jsxp.org/spec/1.0", "public")) {
			final String publicAttr = elem.getAttribute("http://www.jsxp.org/spec/1.0", "public");
			if (publicAttr.equals("true")) {
				accessModifier = "public";
			} else if (publicAttr.equals("false")) {
				accessModifier = "protected";
			} else {
				throw new RuntimeException("unknown attribute value '"+publicAttr+"' for attribute 'public' in namespace '"+"http://www.jsxp.org/spec/1.0"+"'");
			}
		}
		return accessModifier;
	}

	private static boolean isChildOf(Element childElement, final InternalElement elem) {
		while(childElement.getParent() != null) {
			childElement = childElement.getParent();
			if(childElement == elem) {
				return true;
			}
		}
		return false;
	}

	/**
	 * The main method.
	 * 
	 * @param args the args
	 */
	public static void main(final String[] args) {
		System.out.println("starting view controller generation");

		final Context c = new Context();
		Context.setContext(c);

		if(args.length != 4 && args.length != 5 && args.length != 6) {
			//TODO: give a better error description!
			System.err.println("ERROR: 4, 5 or 6 arguments expected.");
			System.err.println();
			System.err.println("USAGE: <baseDirectory> <webFileDirectory> <generatedContentDirectory> <basePackage> <filePatterns> <translations>");
			System.err.println("filePatterns are a comma separated list, default is: \"**/*.xhtml, **/*.html\"");
			System.err.println("translations are a comma separated list like: \"de,de_AT,es\"");
			System.exit(1);
		}

		final File baseDirectory = new File(args[0]);
		if(!baseDirectory.exists() || !baseDirectory.isDirectory()) {
			System.err.println("ERROR: Base directory does not exist ("+baseDirectory.getAbsolutePath()+")");
			System.exit(2);
		}
		webFileDirectory = new File(baseDirectory, args[1]);
		if(!webFileDirectory.exists() || !webFileDirectory.isDirectory()) {
			System.err.println("ERROR: Web file directory does not exist ("+webFileDirectory.getAbsolutePath()+")");
			System.exit(3);
		}
		final String basePackageDirectory = StringUtil.getJavaPackageName(args[3]).replaceAll("\\.", "/");
		final File generatedContentDirectory = new File(baseDirectory, args[2]+"/"+StringUtil.getJavaPackageDir(basePackageDirectory));

		if(generatedContentDirectory.exists() && !generatedContentDirectory.isDirectory()) {
			System.err.println("ERROR: Generated content path exists, but is no directory ("+generatedContentDirectory.getAbsolutePath()+")");
			System.exit(4);
		}

		if(args.length>=5) {
			patterns = args[4].split(",");
			for (int i = 0; i < patterns.length; i++) {
				patterns[i] = patterns[i].trim();
			}
		} else {
			patterns = new String[] { "**/*.xhtml" , "**/*.html"};
		}

		if(args.length==6) {
			translations = args[5].split(",");
			translations = args[4].split(",");
			for (int i = 0; i < translations.length; i++) {
				translations[i] = translations[i].trim();
			}
		}

		Log.info(ViewControllerGenerator.class, StringUtils.combineAll("using file patterns: ", StringUtils.arrayAsString(patterns)));
		Log.info(ViewControllerGenerator.class, StringUtils.combineAll("using translations: ", StringUtils.arrayAsString(translations)));


		final String basePackageName = StringUtil.getJavaPackageName(args[3]);
		Context.setContext(new Context());
		Application.setApplication(new Application());
		findFilesAndGenerateControllers(webFileDirectory, generatedContentDirectory, basePackageName, null, true);
		purgeGeneratedFolder(generatedContentDirectory);
	}

	private static void purgeGeneratedFolder(final File generatedContentDirectory) {
		final File[] files = generatedContentDirectory.listFiles();

		if(files != null) {
			for(final File f : files) {
				if(f.isDirectory()) {
					final String dirName = f.getName();
					if(!(".svn".equals(dirName))) {
						purgeGeneratedFolder(f);
					}
				} else {
					final String fileName = f.getName();
					if(fileName.endsWith(".java") && !generatedFiles.contains(f.getAbsolutePath())) {
						f.deleteOnExit();
					}
				}
			}
		}
	}

}
