/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.jsxp.framework.annotations.ApplicationScope;
import org.jsxp.framework.annotations.FlashScope;
import org.jsxp.framework.annotations.RequestScope;
import org.jsxp.framework.annotations.SessionScope;
import org.jsxp.framework.annotations.ViewScope;


/**
 * The Class ViewControllerLifeCycle. Get access to the life cycle of a view Controller by calling
 * {@link ViewController#getLifeCycle()}
 * 
 * @see ViewController
 */
public class ViewControllerLifeCycle {

	/**
	 * The Enum LifeCyclePhase.
	 */
	public static enum LifeCyclePhase {
		CREATED,
		ELEMENTPROCESSING,
		EXECUTING,
		FILL_PREVIOUS_INPUT,
		INITIALIZE_INPUT_PARAMETER,
		INITIALIZING,
		MARK_VALIDATION_ERRORS,
		PREPARING,
		RENDERING,
		TEMPLATING,
		VALIDATING,
		VARIABLE_SUBSTITUTION
	}

	private class ScopePropertyInfo {
		private Method getMethod;
		private Annotation scope;
		private String scopeName;
		private Method setMethod;
		private Class<?> type;
		private ViewController ViewController;

		@Override
		public String toString() {
			final StringBuilder sb = new StringBuilder();
			sb.append("ScopePropertyInfo: ");
			sb.append("ViewController=").append(ViewController);
			sb.append(", scopeName=").append(scopeName);
			sb.append(", scope=").append(scope);
			sb.append(", type=").append(type);
			sb.append(", getMethod=").append(getMethod);
			sb.append(", setMethod=").append(setMethod);
			return sb.toString();
		}
	};

	private static ThreadLocal<HashMap<String, List<ScopePropertyInfo>>> scopePropertyCacheTL = new ThreadLocal<HashMap<String,List<ScopePropertyInfo>>>();

	private boolean finishedCurrentPhase = false;
	private boolean injectCompleted = false;
	private LifeCyclePhase lifeCyclePhase = LifeCyclePhase.CREATED;
	private Map<LifeCyclePhase, ArrayList<PhaseListener>> phaseToPhaseListeners = null;
	private List<ScopePropertyInfo> scopeProperties;
	private final ViewController ViewController;

	/**
	 * Instantiates a new view controller life cycle.
	 * 
	 * @param ViewController the view control
	 */
	ViewControllerLifeCycle(final ViewController ViewController) {
		this.ViewController = ViewController;
		this.phaseToPhaseListeners = new HashMap<LifeCyclePhase, ArrayList<PhaseListener>>();
		findScopeProperties();
	}

	/**
	 * Adds the phase listener.
	 * 
	 * @param phase the phase
	 * @param observer the observer
	 */
	public void addPhaseListener(final LifeCyclePhase phase, final PhaseListener observer) {
		ArrayList<PhaseListener> phaseListener = phaseToPhaseListeners.get(phase);
		if (phaseListener == null) {
			phaseListener = new ArrayList<PhaseListener>();
			phaseToPhaseListeners.put(phase, phaseListener);
		}
		phaseListener.add(0, observer);
	}

	private void findScopeProperties() {
		HashMap<String, List<ScopePropertyInfo>> propertyCache = scopePropertyCacheTL.get();
		if(propertyCache == null) {
			propertyCache = new HashMap<String, List<ScopePropertyInfo>>();
			scopePropertyCacheTL.set(propertyCache);
		}

		final String className = ViewController.getClass().getName();
		if(propertyCache.containsKey(className)) {
			scopeProperties = propertyCache.get(className);
		} else {
			scopeProperties = new LinkedList<ScopePropertyInfo>();
			propertyCache.put(className, scopeProperties);

			final HashMap<String, ScopePropertyInfo> propertyInfos = new HashMap<String, ScopePropertyInfo>();
			final Method[] methods = ViewController.getClass().getMethods();
			for(final Method m : methods) {
				if(m.getName().startsWith("get") || m.getName().startsWith("set")) {
					final String propertyName = m.getName().substring(3);
					if(!propertyInfos.containsKey(propertyName)) {
						propertyInfos.put(propertyName, new ScopePropertyInfo());
					}
					final ScopePropertyInfo current = propertyInfos.get(propertyName);
					current.ViewController = ViewController;
					String scopeName = null;
					Class<?> type = null;
					if(m.getName().startsWith("get")) {
						current.getMethod = m;
						scopeName = m.getReturnType().getName()+"//"+propertyName;
						type = m.getReturnType();
					} else {
						current.setMethod = m;
						if(m.getParameterTypes().length==1) {
							scopeName = m.getParameterTypes()[0].getName()+"//"+propertyName;
							type = m.getParameterTypes()[0];
						}
					}

					if(current.scopeName == null) {
						current.scopeName = scopeName;
					} else if(!current.scopeName.equals(scopeName)) {
						current.scopeName = null;
					}
					if(current.type == null) {
						current.type = type;
					} else if(!current.type.equals(type)){
						current.type = null;
					}

					final Annotation[] annotations = m.getAnnotations();
					for(final Annotation a : annotations) {
						Annotation scope = null;
						if(a instanceof ApplicationScope) {
							scope = a;
						} else if(a instanceof FlashScope) {
							scope = a;
						} else if(a instanceof RequestScope) {
							scope = a;
						} else if(a instanceof SessionScope) {
							scope = a;
						} else if(a instanceof ViewScope) {
							scope = a;
						}

						if(current.scope == null) {
							current.scope = scope;
						} else if(current.scope != scope) {
							current.scope = null;
						}
					}
				}
			}

			for(final ScopePropertyInfo info : propertyInfos.values()) {
				if(info.getMethod != null && info.scope != null && info.scopeName != null
						&& info.setMethod != null && info.type != null && info.ViewController != null) {
					scopeProperties.add(info);
				}
			}
		}
	}

	/**
	 * Finish phase.
	 * 
	 * @param phase the phase
	 */
	final void finishPhase(final LifeCyclePhase phase) {
		if (!lifeCyclePhase.equals(phase)) {
			throw new RuntimeException("not in phase '"+phase+"' but in phase '"+this.lifeCyclePhase+"' in view '" + ViewController.getViewFilename() + "(" + ViewController + ")'");
		}
		restoreScopeProperties();
		finishedCurrentPhase = true;
	}

	/**
	 * Gets the life cycle phase.
	 * 
	 * @return the life cycle phase
	 */
	final public LifeCyclePhase getLifeCyclePhase() {
		return lifeCyclePhase;
	}

	private void injectScopeProperties() {
		injectCompleted = true;
		for(final ScopePropertyInfo info : scopeProperties) {
			Map<String, Object> scope = null;
			if(info.scope instanceof ApplicationScope) {
				scope = Context.getContext().getScopes().getApplicationScope();
			} else if(info.scope instanceof FlashScope) {
				scope = Context.getContext().getScopes().getFlashScope();
			} else if(info.scope instanceof RequestScope) {
				scope = Context.getContext().getScopes().getRequestScope();
			} else if(info.scope instanceof SessionScope) {
				scope = Context.getContext().getScopes().getSessionScope();
			} else if(info.scope instanceof ViewScope) {
				scope = Context.getContext().getScopes().getViewScope();
			}
			if(scope != null) {
				Object value = scope.get(info.scopeName);
				if(value == null) {
					try {
						value = info.type.newInstance();
					} catch (final InstantiationException e) {
						throw new RuntimeException(e);
					} catch (final IllegalAccessException e) {
						throw new RuntimeException(e);
					}
				}
				if(info.type.isInstance(value)) {
					try {
						info.setMethod.invoke(ViewController, value);
					} catch (final IllegalArgumentException e) {
						throw new RuntimeException(e);
					} catch (final IllegalAccessException e) {
						throw new RuntimeException(e);
					} catch (final InvocationTargetException e) {
						throw new RuntimeException(e);
					}
				}
			}
		}
	}

	private void notifyPhaseListenerAfter(final LifeCyclePhase phase) {
		final ArrayList<PhaseListener> phaseListeners = phaseToPhaseListeners.get(phase);
		if (phaseListeners != null) {
			for (final PhaseListener phaseListener : phaseListeners) {
				phaseListener.afterPhase(phase);
			}
		}
	}

	private void notifyPhaseListenerBefore(final LifeCyclePhase phase) {
		final ArrayList<PhaseListener> phaseListeners = phaseToPhaseListeners.get(phase);
		if (phaseListeners != null) {
			for (final PhaseListener phaseListener : phaseListeners) {
				phaseListener.beforePhase(phase);
			}
		}
	}

	/**
	 * Removes the phase listener.
	 * 
	 * @param phase the phase
	 * @param observer the observer
	 */
	public void removePhaseListener(final LifeCyclePhase phase, final PhaseListener observer) {
		final List<PhaseListener> phaseListener = phaseToPhaseListeners.get(phase);
		if (phaseListener != null) {
			phaseListener.remove(observer);
		}
	}

	private void restoreScopeProperties() {
		if(!injectCompleted) {
			injectScopeProperties();
		}
		for(final ScopePropertyInfo info : scopeProperties) {
			Object value = null;
			try {
				value = info.getMethod.invoke(ViewController);
			} catch (final IllegalArgumentException e) {
				throw new RuntimeException(e);
			} catch (final IllegalAccessException e) {
				throw new RuntimeException(e);
			} catch (final InvocationTargetException e) {
				throw new RuntimeException(e);
			}

			Map<String, Object> scope = null;
			if(info.scope instanceof ApplicationScope) {
				scope = Context.getContext().getScopes().getApplicationScope();
			} else if(info.scope instanceof FlashScope) {
				scope = Context.getContext().getScopes().getFlashScope();
			} else if(info.scope instanceof RequestScope) {
				scope = Context.getContext().getScopes().getRequestScope();
			} else if(info.scope instanceof SessionScope) {
				scope = Context.getContext().getScopes().getSessionScope();
			} else if(info.scope instanceof ViewScope) {
				scope = Context.getContext().getScopes().getViewScope();
			}
			if(scope != null) {
				if(value == null) {
					scope.remove(info.scopeName);
				} else {
					scope.put(info.scopeName, value);
				}
			}
		}
	}

	/**
	 * Sets the life cycle phase.
	 * 
	 * @param lifeCyclePhase the new life cycle phase
	 */
	final void setLifeCyclePhase(final LifeCyclePhase lifeCyclePhase) {
		if (!finishedCurrentPhase) {
			throw new RuntimeException("phase not finished yet. traverse from phase '"+this.lifeCyclePhase+"' to phase '"+lifeCyclePhase+"' not allowed in view '" + ViewController.getViewFilename() + "(" + ViewController + ")'");
		}
		final LifeCyclePhase previousPhase = this.lifeCyclePhase;
		traversePhase(previousPhase, lifeCyclePhase);
		this.lifeCyclePhase = lifeCyclePhase;
		this.finishedCurrentPhase = false;
	}

	final private void traversePhase(final LifeCyclePhase previousPhase, final LifeCyclePhase nextPhase) {
		// allowed traversals
		if (previousPhase.equals(LifeCyclePhase.CREATED) && nextPhase.equals(LifeCyclePhase.PREPARING)) {
		} else if (previousPhase.equals(LifeCyclePhase.PREPARING) && nextPhase.equals(LifeCyclePhase.INITIALIZE_INPUT_PARAMETER)) {
		} else if (previousPhase.equals(LifeCyclePhase.INITIALIZE_INPUT_PARAMETER) && nextPhase.equals(LifeCyclePhase.FILL_PREVIOUS_INPUT)) {
		} else if (previousPhase.equals(LifeCyclePhase.FILL_PREVIOUS_INPUT) && nextPhase.equals(LifeCyclePhase.VALIDATING)) {
		} else if (previousPhase.equals(LifeCyclePhase.VALIDATING) && nextPhase.equals(LifeCyclePhase.INITIALIZING)) {
		} else if (previousPhase.equals(LifeCyclePhase.INITIALIZING) && nextPhase.equals(LifeCyclePhase.MARK_VALIDATION_ERRORS)) {
		} else if (previousPhase.equals(LifeCyclePhase.INITIALIZING) && nextPhase.equals(LifeCyclePhase.TEMPLATING)) {
		} else if (previousPhase.equals(LifeCyclePhase.INITIALIZING) && nextPhase.equals(LifeCyclePhase.VARIABLE_SUBSTITUTION)) {
		} else if (previousPhase.equals(LifeCyclePhase.MARK_VALIDATION_ERRORS) && nextPhase.equals(LifeCyclePhase.EXECUTING)) {
		} else if (previousPhase.equals(LifeCyclePhase.MARK_VALIDATION_ERRORS) && nextPhase.equals(LifeCyclePhase.TEMPLATING)) {
		} else if (previousPhase.equals(LifeCyclePhase.EXECUTING) && nextPhase.equals(LifeCyclePhase.TEMPLATING)) {
		} else if (previousPhase.equals(LifeCyclePhase.TEMPLATING) && nextPhase.equals(LifeCyclePhase.ELEMENTPROCESSING)) {
		} else if (previousPhase.equals(LifeCyclePhase.TEMPLATING) && nextPhase.equals(LifeCyclePhase.VARIABLE_SUBSTITUTION)) {
		} else if (previousPhase.equals(LifeCyclePhase.ELEMENTPROCESSING) && nextPhase.equals(LifeCyclePhase.VARIABLE_SUBSTITUTION)) {
		} else if (previousPhase.equals(LifeCyclePhase.VARIABLE_SUBSTITUTION) && nextPhase.equals(LifeCyclePhase.RENDERING)) {
		} else {
			throw new RuntimeException("traverse from phase '"+previousPhase+"' to phase '"+nextPhase+"' not allowed for view '" + ViewController.getViewFilename() + "(" + ViewController + ")'");
		}

		// after methods
		notifyPhaseListenerAfter(previousPhase);
		restoreScopeProperties();

		// before methods
		notifyPhaseListenerBefore(nextPhase);
		restoreScopeProperties();
	}
}
