/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.jsxp.framework.ViewControllerLifeCycle.LifeCyclePhase;
import org.jsxp.framework.elementprocessing.ElementProcessor;
import org.jsxp.framework.util.object.BeanUtils;
import org.jsxp.framework.webroot.jsxpframework.components.DebugMessagesXhtmlController;
import org.jsxp.framework.webroot.jsxpframework.components.DeveloperWarningsXhtmlController;
import org.jsxp.framework.webroot.jsxpframework.components.FlashMessagesXhtmlController;
import org.jsxp.framework.webroot.jsxpframework.components.RedirectViewXhtmlController;
import org.jsxp.framework.webroot.jsxpframework.components.ScopesMessagesXhtmlController;


/**
 * The Class ViewControllerLifeCycleManager. Responsible for managing the life cycle of any {@link ViewController}.
 * Used internally by the jsxp framework.
 * 
 * Here you can find the possible phases of a {@link ViewController}-LifeCyle: {@link LifeCyclePhase}.
 * 
 * To get information and to react on the traversal of those phases you can implement a PhaseListener
 * and add him to {@link ViewController#getLifeCycle()}
 * {@link ViewControllerLifeCycle#addPhaseListener(LifeCyclePhase, PhaseListener)}.
 */
public class ViewControllerLifeCycleManager {

	/**
	 * Instantiates a new view controller life cycle manager.
	 */
	ViewControllerLifeCycleManager() {
	}

	private boolean convertForCharEncoding(final ViewController ViewController) {
		String currentCharEncoding = Context.getContext().getHttpServletRequest().getCharacterEncoding();
		if (currentCharEncoding == null) {
			currentCharEncoding = "ISO-8859-1";
		}
		return !currentCharEncoding.equals(ViewController.getCharacterEncoding());
	}

	private void execute(final ViewController viewController) {
		viewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.EXECUTING);
		Context.getContext().addDebugMessage(
				"execute view control '"
						+ viewController.getViewFilename() + "("
						+ viewController + ")'");
		try {
			viewController.execute();
		} catch (final Exception e) {
			try {
				viewController.handleExecutingException(e, viewController);
			} catch (final RuntimeException ex) {
				Application.getApplication().handleExecutingException(ex, viewController);
			}
		}
		viewController.getLifeCycle().finishPhase(LifeCyclePhase.EXECUTING	);
	}

	private void fillPreviousInput(final String inputParameterPrefix, final ViewController ViewController) {
		ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.FILL_PREVIOUS_INPUT);

		if(!(ViewController instanceof RedirectViewXhtmlController)) {
			Context.getContext().addDebugMessage(
					"initialize previous form input for view controller '"
							+ ViewController.getViewFilename() + "("
							+ ViewController + ")'");
			if(Boolean.TRUE.equals(Context.getContext().isPreviousInputSaved())) {
				final Map<String, Object[]> previousRequestParameters = Context.getContext().getSavedPreviousInput();
				if(previousRequestParameters==null) {
					throw new IllegalStateException("Restore previous input was requested, but the previous request parameters are not available ("+ViewController.getViewFilename()+").");
				}
				final boolean convert = convertForCharEncoding(ViewController);
				try {
					final Object prevValuesBean = ViewController.getPreviousInputParameterBean();
					if(prevValuesBean != null) {
						initializeInputParameter(inputParameterPrefix, previousRequestParameters, ViewController, convert, prevValuesBean);
					}
				} catch (final Exception e) {
					try {
						ViewController.handleInputProcessingException(e, ViewController);
					} catch (final RuntimeException ex) {
						Application.getApplication().handleInputProcessingException(ex, ViewController);
					}
				}
			}
		}

		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.FILL_PREVIOUS_INPUT);
	}

	private ViewController getGlobalViewTemplates(final String uri, final ViewController ViewController) {
		if(!ViewController.isApplicationTemplateEnabled()) {
			return null;
		}
		if(Application.getApplication().getTemplate() != null && uri != null) {
			if(!Application.getApplication().getTemplate().getViewFilename().equals(uri)) {
				final ViewController viewTemplate = Application.getApplication().getViewController(Application.getApplication().getTemplate().getViewFilename());
				return viewTemplate;
			}
		}

		return null;
	}

	private void importSpecialViewControllers(
			final ViewController ViewController) {
		// DeveloperWarnings
		if (Application.getApplication().getProjectStage().useDeveloperWarningsViewController()) {
			try {
				ViewController.importView(new DeveloperWarningsXhtmlController(), "developerWarnings.", false, null);
			} catch (final Exception e) {
				throw new RuntimeException(e);
			}
		}

		// Scopes
		if (Application.getApplication().getProjectStage().useScopeMessagesViewController()) {
			try {
				ViewController.importView(new ScopesMessagesXhtmlController(), "scopeMessages.", false, null);
			} catch (final Exception e) {
				throw new RuntimeException(e);
			}
		}

		// FlashMessages
		try {
			ViewController.importView(new FlashMessagesXhtmlController(), "flashMessages.", false, null);
		} catch(final Exception e) {
			throw new RuntimeException(e);
		}

		// DebugMessages
		if (Application.getApplication().getProjectStage().useDebugMessagesViewController()) {
			try {
				ViewController.importView(new DebugMessagesXhtmlController(), "debugMessages.", false, null);
			} catch (final Exception e) {
				throw new RuntimeException(e);
			}
		}
	}

	private void init(final String uri, final ViewController ViewController, final boolean importSpecialViews) {
		Context.getContext().addDebugMessage(
				"init view control '"
						+ ViewController.getViewFilename() + "("
						+ ViewController + ")'");
		ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.INITIALIZING);

		if (importSpecialViews) {
			importSpecialViewControllers(ViewController);
		}

		try {
			ViewController.init();

			final ViewController globalTemplate = getGlobalViewTemplates(uri, ViewController);
			if (globalTemplate != null) {
				ViewController.setViewTemplate(globalTemplate);
				ViewController.importView(globalTemplate);
			}
		} catch (final Exception e) {
			try {
				ViewController.handleInitException(e, ViewController);
			} catch (final RuntimeException ex) {
				Application.getApplication().handleInitException(ex, ViewController);
			}
		}
		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.INITIALIZING);
	}

	private void initializeInputParameter(final String inputParameterPrefix, final Map<String, Object[]> inputParameter,
			final ViewController ViewController) {

		ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.INITIALIZE_INPUT_PARAMETER);
		final boolean convert = convertForCharEncoding(ViewController);
		try {
			initializeInputParameter(inputParameterPrefix, inputParameter, ViewController, convert);
		} catch (final Exception e) {
			try {
				ViewController.handleInputProcessingException(e, ViewController);
			} catch (final RuntimeException ex) {
				Application.getApplication().handleInputProcessingException(ex, ViewController);
			}
		}
		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.INITIALIZE_INPUT_PARAMETER);
	}

	private void initializeInputParameter(final String inputParameterPrefix, final Map<String, Object[]> requestInputParameter,
			final ViewController ViewController, final boolean convert) throws Exception {
		Context.getContext().addDebugMessage(
				"initialize input parameter for view control '"
						+ ViewController.getViewFilename() + "("
						+ ViewController + ")'");
		final Object inputParameter = ViewController.getInputParameterBean();
		initializeInputParameter(inputParameterPrefix, requestInputParameter, ViewController, convert, inputParameter);
	}

	private void initializeInputParameter(final String inputParameterPrefix, final Map<String, Object[]> requestInputParameter,
			final ViewController ViewController, final boolean convert, final Object beanToFill)
					throws IllegalAccessException, InvocationTargetException, UnsupportedEncodingException
					{
		final Map<String, Object[]> parameter = requestInputParameter;

		if (beanToFill == null || parameter == null || parameter.isEmpty()) {
			return;
		}

		for (final Entry<String, Object[]> entry : parameter.entrySet()) {
			final String key = entry.getKey();
			if (inputParameterPrefix == null
					|| key.startsWith(inputParameterPrefix)) {
				final Object[] values = entry.getValue();

				if (convert) {
					String currentCharEncoding = Context.getContext().getHttpServletRequest().getCharacterEncoding();
					if (currentCharEncoding == null) {
						currentCharEncoding = "ISO-8859-1";
					}
					for(int i = 0; i<values.length; i++) {
						if (values[i] instanceof String) {
							values[i] = new String(((String)values[i]).getBytes(currentCharEncoding),ViewController.getCharacterEncoding());
						}
					}

					entry.setValue(values);
				}
			}

			BeanUtils.setProperties(inputParameterPrefix, parameter, beanToFill);
		}
					}

	@SuppressWarnings("unchecked")
	private void markValidationErrors(final ViewController ViewController) throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
		ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.MARK_VALIDATION_ERRORS);

		final LinkedList<ValidationError> errorsList = (LinkedList<ValidationError>) Context.getContext().getScopes().getFlashAttributes().get(Validator.VALIDATION_ERROR_LIST);
		if(errorsList != null && !errorsList.isEmpty()) {
			if(!(ViewController instanceof RedirectViewXhtmlController)) {
				for(final ValidationError currentError : errorsList) {
					final String message = currentError.getMessage();
					if (message != null) {
						Context.getContext().getFlashMessages().addErrorMessage(message);
					}

					for(final ElementProcessor elementProcessor : currentError.getElementProcessors()) {
						elementProcessor.process(ViewController);
					}
				}
			}
		}

		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.MARK_VALIDATION_ERRORS);
	}

	private void prepare(final ViewController ViewController, final ViewController parent, final String idPrefix) {
		Application.getApplication().lifeCycleStartsForViewController(ViewController);

		Context.getContext().addDebugMessage(
				"prepare view control '"
						+ ViewController.getViewFilename() + "("
						+ ViewController + ")'");
		ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.PREPARING);

		try {
			ViewController.prepare(parent, idPrefix);
		} catch (final Exception e) {
			try {
				ViewController.handlePreparingException(e, ViewController);
			} catch (final RuntimeException ex) {
				if (parent == null) {
					Application.getApplication().handlePreparingException(e, ViewController);
				} else {
					throw ex;
				}
			}
		}
		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.PREPARING);
	}

	private void processElements(final ViewController viewController) {
		Context.getContext().addDebugMessage(
				"handle resources for view control '"
						+ viewController.getViewFilename() + "("
						+ viewController + ")'");

		viewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.ELEMENTPROCESSING);

		List<ElementProcessor> elementProcessors = null;
		if (viewController.isUsedAsTemplate()) {
			ViewController templatedView = viewController;
			do {
				templatedView = templatedView.getTemplatedView();
			} while (templatedView.isUsedAsTemplate());
			elementProcessors = templatedView.elementProcessors();
		} else {
			elementProcessors = viewController.elementProcessors();
		}

		for (final ElementProcessor elementProcessor : elementProcessors) {
			elementProcessor.process(viewController);
		}

		viewController.getLifeCycle().finishPhase(LifeCyclePhase.ELEMENTPROCESSING);
	}

	/**
	 * Process life cycle.
	 * 
	 * @param parameter the parameter
	 * @param uri the uri
	 * @param viewController the view control
	 * @param currentViewFlow the current view flow
	 * @param importSpecialViews the import special views
	 * 
	 * @return the view controller<?>
	 * 
	 * @throws IOException Signals that an I/O exception has occurred.
	 * @throws Exception the exception
	 */
	ViewController processLifeCycle(final Map<String, Object[]> parameter,
			final String uri, final ViewController viewController, final boolean importSpecialViews) throws IOException, Exception {
		prepare(viewController, null, null);
		initializeInputParameter(null, parameter, viewController);
		fillPreviousInput(null, viewController);
		validate(viewController, null);
		init(uri, viewController, importSpecialViews);

		if(!viewController.isSkipExecution()) {
			markValidationErrors(viewController);
			execute(viewController);
		} else {
			Context.getContext().addDebugMessage(
					"execution SKIPPED for view control '"
							+ viewController.getViewFilename() + "("
							+ viewController + ")'");
		}

		if (Context.getContext().isRedirect() && !(viewController instanceof RedirectViewXhtmlController)) {
			return viewController;
		}

		templating(viewController, null);

		ViewController renderViewController = viewController;
		while (renderViewController.getViewTemplate() != null) {
			renderViewController = renderViewController.getViewTemplate();
		}

		processElements(renderViewController);
		substituteVariables(viewController);
		render(viewController, null);
		return renderViewController;
	}

	/**
	 * Process lifecycle imported view.
	 * 
	 * @param inputParameterPrefix the input parameter prefix
	 * @param inputParameterInjection the input parameter injection
	 * @param importedViewController the imported view control
	 * @param parent the parent
	 * @param idPrefix the id prefix
	 */
	void processLifecycleImportedView(final String inputParameterPrefix, final boolean inputParameterInjection, final ViewController importedViewController, final ViewController parent, final String idPrefix) {
		prepare(importedViewController, parent, idPrefix);
		initializeInputParameter(inputParameterPrefix, Context.getContext().getRequestInputParameter(), importedViewController);
		fillPreviousInput(null, importedViewController);
		validate(importedViewController, parent);
		init(null, importedViewController, false);
	}

	private void render(final ViewController ViewController, final ViewController parent) {
		final List<ViewController> importedViews = ViewController
				.getImportedViews();
		for (final ViewController importedView : importedViews) {
			try {
				render(importedView, ViewController);
			} catch (final Exception e) {
				try {
					ViewController.handleRenderingException(e, importedView);
				} catch (final RuntimeException ex) {
					if (parent == null) {
						Application.getApplication().handleInitException(ex, importedView);
					} else {
						throw ex;
					}
				}
			}

		}

		Context.getContext().addDebugMessage(
				"render view control '"
						+ ViewController.getViewFilename() + "("
						+ ViewController + ")'");
		try {
			ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.RENDERING);
		} catch (final Exception e) {
			try {
				ViewController.handleRenderingException(e, ViewController);
			} catch (final RuntimeException ex) {
				if (parent == null) {
					Application.getApplication().handleInitException(ex, ViewController);
				} else {
					throw ex;
				}
			}
		}
		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.RENDERING);
	}

	private void substituteVariables(final ViewController viewController) {
		Context.getContext().addDebugMessage(
				"substitute variables for view control '"
						+ viewController.getViewFilename() + "("
						+ viewController + ")'");
		viewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.VARIABLE_SUBSTITUTION);
		viewController.substituteVariables();
		viewController.getLifeCycle().finishPhase(LifeCyclePhase.VARIABLE_SUBSTITUTION);

		final List<ViewController> importedViews = viewController.getImportedViews();
		for (final ViewController importedView : importedViews) {
			substituteVariables(importedView);
		}
	}

	private void templating(final ViewController ViewController, final ViewController parent) {
		final List<ViewController> importedViews = ViewController
				.getImportedViews();
		for (final ViewController importedView : importedViews) {
			try {
				templating(importedView, ViewController);
			} catch (final Exception e) {
				try {
					ViewController.handleTemplatingException(e, importedView);
				} catch (final RuntimeException ex) {
					if (parent == null) {
						Application.getApplication().handleTemplatingException(ex, importedView);
					} else {
						throw ex;
					}
				}
			}
		}

		ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.TEMPLATING);
		try {
			if (ViewController.getViewTemplate() != null) {
				final ViewController viewTemplate = ViewController.getViewTemplate();
				Context.getContext().addDebugMessage(
						"processing template '" + viewTemplate + "'for view '"
								+ ViewController.getViewFilename() + "("
								+ ViewController + ")'");
				final Element[] templateElements = viewTemplate.getTemplateElements();
				if (templateElements != null) {
					for (final Element e : templateElements) {
						String templateId = e.getId();
						if (viewTemplate.getIdPrefix() != null) {
							templateId = templateId.substring(viewTemplate.getIdPrefix().length());
						}
						final Element toExchange = ViewController.getElement(templateId);
						if (toExchange != null) {
							e.exchange(toExchange);
						} else {
							Context.getContext().addDeveloperWarningMessage(
									"template element '" + e.getId()
									+ "' not found in view '"
									+ ViewController.getViewFilename()
									+ "'");
							e.remove();
						}
					}
				}
			}
		} catch (final Exception e) {
			try {
				ViewController.handleTemplatingException(e, ViewController);
			} catch (final RuntimeException ex) {
				if (parent == null) {
					Application.getApplication().handleTemplatingException(ex, ViewController);
				} else {
					throw ex;
				}
			}
		}
		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.TEMPLATING);
	}

	private void validate(final ViewController ViewController, final ViewController parent) {
		Context.getContext().addDebugMessage(
				"validate view control '"
						+ ViewController.getViewFilename() + "("
						+ ViewController + ")'");
		ViewController.getLifeCycle().setLifeCyclePhase(LifeCyclePhase.VALIDATING);
		try {
			ViewController.validate();
		} catch (final Exception e) {
			try {
				ViewController.handleValidatingException(e, ViewController);
			} catch (final RuntimeException ex) {
				if (parent == null) {
					Application.getApplication().handleValidatingException(ex, ViewController);
				} else {
					throw ex;
				}
			}
		}
		ViewController.getLifeCycle().finishPhase(LifeCyclePhase.VALIDATING);
	}

}
