/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.dom;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import org.jsxp.framework.DuplicateIdException;
import org.jsxp.framework.Element;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.internal.ProxyElement;
import org.jsxp.framework.resources.ViewResourceResolver;
import org.jsxp.framework.util.text.StringUtil;


/**
 * The Class DomElement. Default implementation (DOM based) of jsxp elements.
 * 
 * @see InternalElement
 */
public class DefaultElement implements InternalElement {

	private final HashMap<String, String> attributes = new HashMap<String, String>();
	private final HashSet<Element> children = new HashSet<Element>();
	private final ArrayList<DefaultElement> childrenList = new ArrayList<DefaultElement>();
	private String id = null;
	private String idPrefix = null;
	private String idSuffix = null;
	private final boolean isText;
	private String name = null;
	private String namespaceUri = null;
	private Element parent = null;
	private DefaultElement sourceParent = null;
	private final ViewController sourceViewController;
	private String value = null;
	private final HashMap<String, String> variableValues = new HashMap<String, String>();

	/**
	 * Instantiates a new dom element.
	 *
	 * @param sourceViewController
	 * @param sourceParent
	 * @param id
	 * @param idPrefix
	 * @param namespaceUri
	 * @param name
	 * @param value
	 */
	public DefaultElement(final ViewController sourceViewController, final DefaultElement sourceParent, final String id, final String idPrefix, final String idSuffix, final String namespaceUri, final String name, final String value, final boolean isText) {
		this.sourceParent = sourceParent;
		this.sourceViewController = sourceViewController;
		this.idPrefix = idPrefix;
		this.idSuffix = idSuffix;
		this.id = id;
		this.value = value;
		this.name = name;
		this.namespaceUri = namespaceUri;
		this.isText = isText;

		if (sourceViewController != null) {
			sourceViewController.addElement(this);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addElement(org.jsxp.framework.framework.Element)
	 */
	@Override
	public void addElement(final Element e) {
		addElement(e, null);
	}

	private void addElement(final Element e, final Element insertBefore) {
		if (null == e) {
			throw new NullPointerException("parent element is null");
		}


		final DefaultElement element = getRealElement(e);
		if(insertBefore != null && !(insertBefore instanceof DefaultElement)) {
			throw new IllegalArgumentException("The given element is of wrong class");
		}

		if (this.children.contains(insertBefore)) {
			throw new IllegalArgumentException("the element to insert before is not a child");
		}

		if (e.getParent() != null) {
			e.getParent().removeElement(e);
		}

		final int index = childrenList.indexOf(insertBefore);
		childrenList.add(index, element);
		children.add(element);

		element.setParent(this);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addElements(org.jsxp.framework.framework.Element[])
	 */
	@Override
	public void addElements(final Element[] elements) {
		addElements(elements, null);
	}

	private void addElements(final Element[] elements, final Element insertBefore) {
		for (final Element e : elements) {
			addElement(e, insertBefore);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addText(java.lang.String)
	 */
	@Override
	public void addText(final String text) {
		final DefaultElement textElement = new DefaultElement(sourceViewController, this, null, idPrefix, idSuffix, null, null, null, true);
		textElement.setValue(text);
		addElement(textElement);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addAsElements(java.lang.String)
	 */
	@Override
	public Element addXmlAsElements(final String xml) {
		final ViewParserDom parser = new ViewParserDom();
		final InputStream inputStream = new ByteArrayInputStream(xml.getBytes());
		final InternalElement e = parser.read(null, new ViewResourceResolver(-1, inputStream), this, idPrefix, null, this, null);

		addElements(e.getElements());
		return e;
	}

	private void clearChildren() {
		children.clear();
		childrenList.clear();
	}

	private boolean containsVariableValueParents(final String key, final DefaultElement element, final StringBuffer variable) {
		String lookupKey = key;

		if (key.indexOf('.')>0 && key.startsWith(id+'.') && !variableValues.containsKey(lookupKey)) {
			lookupKey = key.substring(id.length()+1);
		}
		if (variableValues.containsKey(lookupKey)) {
			final String value = variableValues.get(lookupKey);
			if (value != null) {
				variable.append(value);
			}
			return true;
		} else if (parent != null && getRealElement(parent).sourceViewController == sourceViewController) {
			return getRealElement(parent).containsVariableValueParents(key, element, variable);
		} else {
			return false;
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#copy()
	 */
	@Override
	public DefaultElement copy() {
		return copy(null, null);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#copy(java.lang.String)
	 */
	@Override
	public DefaultElement copy(final String idPrefix, final String idSuffix) {
		try {
			final DefaultElement copy = processCopiedElement(this, null, idPrefix, idSuffix, null, sourceViewController);
			return copy;
		} catch (final DuplicateIdException e) {
			throw new RuntimeException(e);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#exchange(org.jsxp.framework.framework.Element)
	 */
	@Override
	public void exchange(final Element e) {
		if (this.getParent() != null) {
			final DefaultElement parent = getRealElement(this.getParent());
			parent.addElement(e, this);
			parent.removeElement(this);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#exchange(org.jsxp.framework.framework.Element[])
	 */
	@Override
	public void exchange(final Element[] elements) {
		if (this.getParent() != null) {
			final DefaultElement parent = getRealElement(this.getParent());
			for (final Element e : elements) {
				parent.addElement(e, this);
			}
			parent.removeElement(this);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#findElementById(java.lang.String)
	 */
	@Override
	public DefaultElement findElementById(final String id) {
		if(id.equals(getId())) {
			return this;
		}
		for(final Element child : getElements()) {
			final DefaultElement result = getRealElement(child).findElementById(id);
			if(result != null) {
				return result;
			}
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#findElementByName(java.lang.String, java.lang.String)
	 */
	@Override
	public DefaultElement findElementByName(final String name, final String namespaceURI) {
		if(name.equals(getName())) {
			if (namespaceURI == null || namespaceURI.equals(getNamespaceURI())) {
				;
			}
			return this;
		}
		for(final Element child : getElements()) {
			final DefaultElement result = getRealElement(child).findElementByName(name, namespaceURI);
			if(result != null) {
				return result;
			}
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getAttribute(java.lang.String)
	 */
	@Override
	public String getAttribute(final String key) {
		return getAttribute(null, key);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getAttribute(java.lang.String, java.lang.String)
	 */
	@Override
	public String getAttribute(final String namespace, final String key) {
		return attributes.get(StringUtil.concateStrings(null, namespace, ":", key));
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getAttributes()
	 */
	@Override
	public Map<String, String> getAttributes() {
		return attributes;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getElements()
	 */
	@Override
	public Element[] getElements() {
		return childrenList.toArray(new Element[0]);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getId()
	 */
	@Override
	public String getId() {
		return id;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getIdPrefix()
	 */
	@Override
	public String getIdPrefix() {
		return idPrefix;
	}

	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.Element#getIdSuffix()
	 */
	@Override
	public String getIdSuffix() {
		return idSuffix;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getName()
	 */
	@Override
	public String getName() {
		return name;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getNamespaceURI()
	 */
	@Override
	public String getNamespaceURI() {
		return namespaceUri;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getParent()
	 */
	@Override
	public Element getParent () {
		return parent;
	}

	private DefaultElement getRealElement(final Element e) {
		if(e instanceof ProxyElement) {
			if (!(((ProxyElement)e).getRealElement() instanceof DefaultElement)) {
				throw new IllegalArgumentException("The given element is from the wrong class");
			}
			return (DefaultElement)((ProxyElement)e).getRealElement();
		}

		if(!(e instanceof DefaultElement)) {
			throw new IllegalArgumentException("The given element is from the wrong class");
		}

		return (DefaultElement)e;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getSourceViewController()
	 */
	@Override
	public ViewController getSourceViewController() {
		return sourceViewController;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getValue()
	 */
	@Override
	public String getValue() {
		return value;
	}

	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.internal.InternalElement#getVariableValue(java.lang.String)
	 */
	@Override
	public String getVariableValue(final String key) {
		final StringBuffer variable = new StringBuffer();

		if (containsVariableValueParents(key, this, variable)) {
			return variable.toString();
		}

		if (sourceViewController.containsVariable(key)) {
			return sourceViewController.getValueForVariable(key);
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#hasAttribute(java.lang.String)
	 */
	@Override
	public boolean hasAttribute(final String key) {
		return hasAttribute(null, key);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#hasAttribute(java.lang.String, java.lang.String)
	 */
	@Override
	public boolean hasAttribute(final String namespace, final String key) {
		return attributes.containsKey(StringUtil.concateStrings(null, namespace, ":", key));
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#isText()
	 */
	@Override
	public boolean isText() {
		return isText;
	}

	/**
	 * Process node.
	 * 
	 * @param contextNode the context node
	 * @param parentElement the parent element
	 * @param idPrefix the id prefix
	 * @param idPrefix the id suffix (not used in child elements)
	 * @param idElements the id elements
	 * @param sourceViewController the source view controller
	 * 
	 * @return the dom element
	 * 
	 * @throws DuplicateIdException the duplicate id exception
	 */
	DefaultElement processCopiedElement(final DefaultElement original, final InternalElement parentElement, final String idPrefix, final String idSuffix, final Map<String, InternalElement> idElements, final ViewController sourceViewController) throws DuplicateIdException {

		final ViewController sourceViewControllerLocal = original.sourceViewController;
		final DefaultElement sourceParentElement = original.sourceParent;

		String id = null;
		String idNode = original.id;
		String idPrefixLocal = idPrefix;
		if (idNode != null) {
			id = idNode;
			idPrefixLocal = original.idPrefix;
			idPrefixLocal = StringUtil.concateStrings(null, idPrefix, idPrefixLocal);

			if (idSuffix != null) {
				idNode = StringUtil.concateStrings(null, idPrefixLocal, id, "$", idSuffix);
			} else {
				idNode = StringUtil.concateStrings(null, idPrefixLocal, id);
			}
			id = idNode;
		}

		final DefaultElement contextElement = new DefaultElement(sourceViewControllerLocal, sourceParentElement, id, idPrefixLocal, idSuffix, original.namespaceUri, original.name, original.value, original.isText);

		if (idNode != null && idElements != null) {
			// usual id elements from source file
			if (idElements.containsKey(id)) {
				//throw new DuplicateIdException("id '"+id+"' already definded");
			} else {
				idElements.put(id, contextElement);
			}
		}

		if (parentElement != null) {
			parentElement.addElement(contextElement);
		}

		for (final DefaultElement childNode : original.childrenList) {
			processCopiedElement(childNode, contextElement, idPrefix, null, idElements, sourceViewController);
		}

		return contextElement;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#remove()
	 */
	@Override
	public void remove() {
		if (this.getParent() != null) {
			this.getParent().removeElement(this);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeAttribute(java.lang.String)
	 */
	@Override
	public void removeAttribute(final String name) {
		removeAttribute(null, name);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeAttribute(java.lang.String, java.lang.String)
	 */
	@Override
	public void removeAttribute(final String namespace, final String name) {
		attributes.remove(StringUtil.concateStrings(null, namespace, ":", name));
	}

	private void removeChild(final DefaultElement element) {
		children.remove(element);
		childrenList.remove(element);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeElement(org.jsxp.framework.framework.Element)
	 */
	@Override
	public void removeElement(final Element e) {
		final DefaultElement element = getRealElement(e);

		element.setParent(null);
		removeChild(element);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeElements()
	 */
	@Override
	public void removeElements() {
		for (final Element e : childrenList) {
			final DefaultElement c = getRealElement(e);
			c.setParent(null);
		}
		clearChildren();
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setAttribute(java.lang.String, java.lang.String)
	 */
	@Override
	public void setAttribute(final String key, final String value) {
		setAttribute(null, key, value);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setAttribute(java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
	public void setAttribute(final String namespace, final String key, final String value) {
		attributes.put(StringUtil.concateStrings(null, namespace, ":", key), value);
	}

	/**
	 * Sets the identifier of this element. This is the id you can define for elements in the
	 * design view (e.g. by setting the jsxp:id attribute). it can be null for anonymous elements.
	 * The framework calls this method automatically to set the element id.
	 * 
	 * @param id the new id
	 */
	void setId(final String id) {
		this.id = id;
	}

	/**
	 * Sets the parent element. Needed to structure the elements in a hierarchical order.
	 * 
	 * @param parent the new parent
	 */
	void setParent(final Element parent) {
		this.parent = parent;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setValue(java.lang.String)
	 */
	@Override
	public void setValue(final String value) {
		if (getElements() == null || getElements().length == 0) {
			this.value = value;
		} else {
			removeElements();
			addText(value);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setValues(java.lang.String[])
	 */
	@Override
	public void setValues(final String[] values) {
		if (getElements() == null || getElements().length == 0) {
			final StringBuilder buffer = new StringBuilder();
			for (final String value : values) {
				buffer.append(value);
			}
			setValue(buffer.toString());
		} else {
			int i = 0;
			for (final Element child : getElements()) {
				if (child.isText()) {
					child.setValue(values[i]);
					i++;
					if (i >= values.length) {
						break;
					}
				}
			}
			if (i < values.length) {
				for (; i< values.length; i++) {
					addText(values[i]);
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setVariable(java.lang.String, java.lang.String)
	 */
	@Override
	public void setVariable(final String name, final String value) {
		variableValues.put(name, value);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#fillParameters(java.lang.Object)
	 */
	@Override
	public void setVariablesWithBean(final Object parameterBean) {
		// this method is overridden in the proxy element objects
		return;
	}

	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.internal.InternalElement#substituteVariables()
	 */
	@Override
	public void substituteVariables() {
		if(this.isText()) {
			final String text = this.getValue();
			String newText = StringUtil.replaceVariables(text, this);
			if (newText == null) {
				newText = "";
			}
			setValue(newText);
		} else {
			if(attributes != null) {
				for (final String key : attributes.keySet()) {
					final String attributeText = attributes.get(key);
					String newText = StringUtil.replaceVariables(attributeText, this);
					if (newText == null) {
						newText = "";
					}

					attributes.put(key, newText);
				}
			}
		}
	}

}
