/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.dom;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import org.jsxp.framework.Application;
import org.jsxp.framework.DuplicateIdException;
import org.jsxp.framework.Element;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.ViewParser;
import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.internal.ProxyElement;
import org.jsxp.framework.resources.ViewResourceResolver;
import org.jsxp.framework.util.text.StringUtil;
import org.w3c.dom.Attr;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Text;


/**
 * The Class DomElement. Default implementation (DOM based) of jsxp elements.
 * 
 * @see InternalElement
 */
public class ElementDom implements InternalElement {
	
	private static ViewParser parser = Application.getApplication().getViewParser();

	private HashSet<Element> children = new HashSet<Element>();
	private ArrayList<Element> childrenList = new ArrayList<Element>();
	private Node node = null;
	private Element parent = null;
	@SuppressWarnings("unused")
	private ElementDom sourceParent = null;
	private HashMap<String, String> variableValues = new HashMap<String, String>();
	private ViewController sourceViewController;
	private String idPrefix = null;
	private String idSuffix = null;
	private String id = null;
	
	/**
	 * Instantiates a new dom element.
	 * 
	 * @param sourceViewController the source view controller
	 * @param sourceParent the source parent
	 * @param node the node
	 * @param id the id
	 * @param idPrefix the id prefix
	 */
	public ElementDom(ViewController sourceViewController, ElementDom sourceParent, Node node,  String id, String idPrefix, String idSuffix) {
		this.node = node;
		this.sourceParent = sourceParent;
		this.sourceViewController = sourceViewController;
		
		this.idPrefix = idPrefix;
		this.idSuffix = idSuffix;
		this.id = id;
		if (sourceViewController != null) {
			sourceViewController.addElement(this);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getSourceViewController()
	 */
	public ViewController getSourceViewController() {
		return sourceViewController;
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getParent()
	 */
	public Element getParent () {
		return parent;
	}
	
	/**
	 * Sets the parent element. Needed to structure the elements in a hierarchical order.
	 * 
	 * @param parent the new parent
	 */
	void setParent(Element parent) {
		this.parent = parent;
	}
	
	/**
	 * Gets the node (the real DOM node this implementation is based on).
	 * 
	 * @return the node
	 */
	Node getNode() {
		return node;
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getIdPrefix()
	 */
	public String getIdPrefix() {
		return idPrefix;
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getId()
	 */
	public String getId() {
		return id;
	}
	
	/**
	 * Sets the identifier of this element. This is the id you can define for elements in the
	 * design view (e.g. by setting the jsxp:id attribute). it can be null for anonymous elements.
	 * The framework calls this method automatically to set the element id.
	 * 
	 * @param id the new id
	 */
	void setId(String id) {
		this.id = id;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getElements()
	 */
	public Element[] getElements() {
		return (Element[])childrenList.toArray(new Element[0]);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addText(java.lang.String)
	 */
	public void addText(String text) {
		Text textNode = getNode().getOwnerDocument().createTextNode(text);
		ElementDom textElement = new ElementDom(sourceViewController, this, textNode, null, idPrefix, idSuffix);
		addElement(textElement);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addElement(org.jsxp.framework.framework.Element)
	 */
	public void addElement(Element e) {
		addElement(e, null);
	}
	
	private void addElement(Element e, Element insertBefore) {
		if (null == e) {
			throw new NullPointerException("parent element is null");
		}
		
		if(e instanceof ProxyElement) {
			if (!(((ProxyElement)e).getRealElement() instanceof ElementDom)) {
				throw new IllegalArgumentException("The given element must be a DomElement");
			}
		} else if(!(e instanceof ElementDom)) {
			throw new IllegalArgumentException("The given element must be a DomElement");
		}
		
		if(insertBefore != null && !(insertBefore instanceof ElementDom)) {
			throw new IllegalArgumentException("The given element must be a DomElement");
		}
		 
		ElementDom domElement = getDomElement(e);
		ElementDom elementInsertBefore = (ElementDom) insertBefore;
		
		if (this.children.contains(e)) {
			return;
		}
		
		if (e.getParent() != null) {
			e.getParent().removeElement(e);
		}
		
		addChild(e);
		domElement.setParent(this);
		
		if (domElement.getNode().getParentNode() != this.getNode()) {
			Node nodeInsertBefore = null;
			if (elementInsertBefore != null) {
				nodeInsertBefore = elementInsertBefore.getNode();
			}
			getNode().insertBefore(domElement.getNode(), nodeInsertBefore);
		}	
		
	}
	
	private void addChild(Element e) {
		children.add(e);
		childrenList.add(e);
	}

	private void removeChild(ElementDom element) {
		children.remove(element);
		childrenList.remove(element);
	}
	
	private void clearChildren() {
		children.clear();
		childrenList.clear();
	}

	private ElementDom getDomElement(Element e) {
		if(e instanceof ProxyElement) {
			if (!(((ProxyElement)e).getRealElement() instanceof ElementDom)) {
				throw new IllegalArgumentException("The given element must be a DomElement");
			}
			return (ElementDom)((ProxyElement)e).getRealElement();
		} 
		
		if(!(e instanceof ElementDom)) {
			throw new IllegalArgumentException("The given element must be a DomElement");
		}
		
		return (ElementDom)e;
	}
	
	private void addElements(Element[] elements, Element insertBefore) {
		for (Element e : elements) {
			addElement(e, insertBefore);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addElements(org.jsxp.framework.framework.Element[])
	 */
	public void addElements(Element[] elements) {
		addElements(elements, null);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#remove()
	 */
	public void remove() {
		if (this.getParent() != null) {
			this.getParent().removeElement(this);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#exchange(org.jsxp.framework.framework.Element)
	 */
	public void exchange(Element e) {
		if (this.getParent() != null) {
			ElementDom parent = getDomElement(this.getParent());
			parent.addElement(e, this);
			parent.removeElement(this);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#exchange(org.jsxp.framework.framework.Element[])
	 */
	public void exchange(Element[] elements) {
		if (this.getParent() != null) {
			ElementDom parent = getDomElement(this.getParent());
			for (Element e : elements) {
				parent.addElement(e, this);
			}
			parent.removeElement(this);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeElement(org.jsxp.framework.framework.Element)
	 */
	public void removeElement(Element e) {
		if(e instanceof ProxyElement) {
			e = ((ProxyElement)e).getRealElement();
		}
		if(!(e instanceof ElementDom)) {
			throw new IllegalArgumentException("The given element must be a DomElement");
		}
		ElementDom element = (ElementDom) e;
		
		element.setParent(null);
		getNode().removeChild(element.getNode());
		removeChild(element);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeElements()
	 */
	public void removeElements() {
		for (Element e : childrenList) {
			ElementDom c = getDomElement(e);
			c.setParent(null);
			getNode().removeChild(c.getNode());
		}
		clearChildren();
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setAttribute(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void setAttribute(String namespace, String key, String value) {
		((org.w3c.dom.Element)getNode()).setAttributeNS(namespace, key, value);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setAttribute(java.lang.String, java.lang.String)
	 */
	public void setAttribute(String key, String value) {
		((org.w3c.dom.Element)getNode()).setAttributeNS(null, key, value);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getAttribute(java.lang.String)
	 */
	public String getAttribute(String key) {
		return ((org.w3c.dom.Element)getNode()).getAttribute(key);
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getAttribute(java.lang.String, java.lang.String)
	 */
	public String getAttribute(String namespace, String key) {
		return ((org.w3c.dom.Element)getNode()).getAttributeNS(namespace, key);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#hasAttribute(java.lang.String, java.lang.String)
	 */
	public boolean hasAttribute(String namespace, String key) {
		return ((org.w3c.dom.Element)getNode()).hasAttributeNS(namespace, key);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#hasAttribute(java.lang.String)
	 */
	public boolean hasAttribute(String key) {
		return ((org.w3c.dom.Element)getNode()).hasAttribute(key);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setValue(java.lang.String)
	 */
	public void setValue(String value) {
		if(getNode() instanceof Text) {
			((Text)getNode()).setData(value);
		} else {
			removeElements();
			addText(value);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setValues(java.lang.String[])
	 */
	public void setValues(String[] values) {
		if(getNode() instanceof Text) {
			StringBuilder buffer = new StringBuilder();
			for (String value : values) {
				buffer.append(value);	
			}
			((Text)getNode()).setData(buffer.toString());
		} else {
			int i = 0;
			for (Element child : getElements()) {
				if (child.isText()) {
					child.setValue(values[i]);
					i++;
					if (i >= values.length) {
						break;
					}
				}
			}
			if (i < values.length) {
				for (; i< values.length; i++) {
					addText(values[i]);					
				}
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getValue()
	 */
	public String getValue() {
		if(getNode() instanceof Text) {
			return ((Text)getNode()).getData();
		} else {
			return ((org.w3c.dom.Element)getNode()).getTextContent();
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getName()
	 */
	public String getName() {
		if (node.getNodeType() == Node.ELEMENT_NODE) {
			return ((org.w3c.dom.Element)getNode()).getNodeName();
		} else {
			return null;
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getNamespaceURI()
	 */
	public String getNamespaceURI() {
		return ((org.w3c.dom.Element)getNode()).getNamespaceURI();
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#copy(java.lang.String)
	 */
	public ElementDom copy(String idPrefix, String idSuffix) {
		try {
			Node n = getNode().cloneNode(true);
			if(!(parser instanceof ViewParserDom)) {
				throw new IllegalStateException("The parser for a DomElement must be a DomParser");
			}
			
			ElementDom c = ((ViewParserDom)parser).processNode(n, null, idPrefix, idSuffix, null, sourceViewController);
			c.sourceParent = this;
			return c;
		} catch (DuplicateIdException e) {
			throw new RuntimeException(e);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#findElementByName(java.lang.String, java.lang.String)
	 */
	public ElementDom findElementByName(String name, String namespaceURI) {
		if(name.equals(getName())) {
			if (namespaceURI == null || namespaceURI.equals(getNamespaceURI()));
			return this;
		}
		for(Element child : getElements()) {
			ElementDom result = getDomElement(child).findElementByName(name, namespaceURI);
			if(result != null) {
				return result;
			}
		}
		return null;
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#findElementById(java.lang.String)
	 */
	public ElementDom findElementById(String id) {
		if(id.equals(getId())) {
			return this;
		}
		for(Element child : getElements()) {
			ElementDom result = getDomElement(child).findElementById(id);
			if(result != null) {
				return result;
			}
		}
		return null;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#copy()
	 */
	public ElementDom copy() {
		return copy(null, null);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#isText()
	 */
	public boolean isText() {
		return node instanceof Text;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getAttributes()
	 */
	public Map<String, String> getAttributes() {
		HashMap<String, String> attributes = new HashMap<String, String>();
		NamedNodeMap attr = node.getAttributes();
		if(attr != null) {
			for(int i=0; i<attr.getLength(); i++) {
				Attr attribute = (Attr) attr.item(i);
				if (attribute.getNamespaceURI() == null) {
					attributes.put(attribute.getName(), attribute.getValue());
				}
			}
		}
		return attributes;
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeAttribute(java.lang.String)
	 */
	public void removeAttribute(String name) {
		node.getAttributes().removeNamedItem(name);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#removeAttribute(java.lang.String, java.lang.String)
	 */
	public void removeAttribute(String namespace, String name) {
		node.getAttributes().removeNamedItemNS(namespace, name);
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#getAttributes(java.lang.String)
	 */
	public Map<String, String> getAttributes(String namespace) {
		HashMap<String, String> attributes = new HashMap<String, String>();
		NamedNodeMap attr = node.getAttributes();
		for(int i=0; i<attr.getLength(); i++) {
			Attr attribute = (Attr) attr.item(i);
			if (attribute.getNamespaceURI().equals(namespace)) {
				attributes.put(attribute.getName(), attribute.getValue());
			}
		}
		return attributes;
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#fillParameters(java.lang.Object)
	 */
	public void setVariablesWithBean(Object parameterBean) {
		// this method is overridden in the proxy element objects
		return;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#setVariable(java.lang.String, java.lang.String)
	 */
	public void setVariable(String name, String value) {
		variableValues.put(name, value);
	}	
	
	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.internal.InternalElement#getVariableValue(java.lang.String)
	 */
	public String getVariableValue(String key) {
		StringBuffer variable = new StringBuffer();
		
		if (containsVariableValueParents(key, this, variable)) {
			return variable.toString();
		}
		
		if (sourceViewController.containsVariable(key)) {
			return sourceViewController.getValueForVariable(key);
		}
		
		return null;
	}
		
	private boolean containsVariableValueParents(String key, ElementDom element, StringBuffer variable) {
		String lookupKey = key;
		
		if (key.indexOf('.')>0 && key.startsWith(id+'.') && !variableValues.containsKey(lookupKey)) {
			lookupKey = key.substring(id.length()+1);
		}		
		if (variableValues.containsKey(lookupKey)) {
			String value = variableValues.get(lookupKey);
			if (value != null) {
				variable.append(value);
			}
			return true;
		} else if (parent != null && getDomElement(parent).sourceViewController == sourceViewController) {
			return getDomElement(parent).containsVariableValueParents(key, element, variable);
		} else {
			return false;
		}		
	}
	
	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.internal.InternalElement#substituteVariables()
	 */
	public void substituteVariables() {
		if(this.isText()) {
			String text = this.getValue();
			String newText = StringUtil.replaceVariables(text, this);
			if (newText == null) {
				newText = "";
			}
			setValue(newText);
		} else {
			NamedNodeMap attributes = node.getAttributes();
			if(attributes != null) {
				for(int i=0; i<attributes.getLength(); i++) {
					Attr attribute = (Attr) attributes.item(i);
					String attributeText = attribute.getValue();
					String newText = StringUtil.replaceVariables(attributeText, this);
					if (newText == null) {
						newText = "";
					}
					
					attribute.setValue(newText);
				}
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Element#addAsElements(java.lang.String)
	 */
	public Element addXmlAsElements(String xml) {
		ViewParserDom parser = new ViewParserDom();
		InputStream inputStream = new ByteArrayInputStream(xml.getBytes());
		InternalElement e = parser.read(null, new ViewResourceResolver(-1, inputStream), this, idPrefix, null, this, null);
		
		addElement(e);//.getElements());	
		return e;
	}

	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.Element#getIdSuffix()
	 */
	public String getIdSuffix() {
		return idSuffix;
	}
	
}
