/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.dom;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;

import org.jsxp.framework.Context;
import org.jsxp.framework.DuplicateIdException;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.ViewParser;
import org.jsxp.framework.dom.adapter.XMLTransformer;
import org.jsxp.framework.internal.InternalElement;
import org.jsxp.framework.internal.ProxyElement;
import org.jsxp.framework.resources.ResourceResolver;
import org.jsxp.framework.util.adapter.AdapterFinder;
import org.jsxp.framework.util.text.StringUtil;
import org.w3c.dom.Document;
import org.w3c.dom.EntityReference;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.UserDataHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * Default implementation of {@link ViewParser}. Used by the jsxp framework to parse a xml-based design view file and build
 * a jsxp element representation (based on the {@link ElementDom} implementation).
 */
public class ViewParserDom extends XMLProcessing implements ViewParser {

	private final Map<String, Document> filenameDOMCache = new Hashtable<String, Document>();
	private final Map<String, Long> filenameModifiedTimestamp = new Hashtable<String, Long>();
	private String idAttributeName = "id";
	private String idNamespace = "http://www.jsxp.org/spec/1.0";

	/**
	 * Instantiates a new dom parser.
	 */
	public ViewParserDom() {
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Parser#getIdAttributeName()
	 */
	@Override
	public String getIdAttributeName() {
		return idAttributeName;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Parser#getIdNamespace()
	 */
	@Override
	public String getIdNamespace() {
		return idNamespace;
	}

	private Object getNodeUserData(final Node node, final String key, final Object defaultValue) {
		Object value;
		value = node.getUserData(key);
		if (value == null) {
			value = defaultValue;
			setNodeUserData(node, key, defaultValue);
		}
		return value;
	}

	/**
	 * Process node.
	 * 
	 * @param contextNode the context node
	 * @param parentElement the parent element
	 * @param idPrefix the id prefix
	 * @param idPrefix the id suffix (not used in child elements)
	 * @param idElements the id elements
	 * @param sourceViewController the source view controller
	 * 
	 * @return the dom element
	 * 
	 * @throws DuplicateIdException the duplicate id exception
	 */
	ElementDom processNode(Node contextNode, final InternalElement parentElement, final String idPrefix, final String idSuffix, final Map<String, InternalElement> idElements, final ViewController sourceViewController) throws DuplicateIdException {
		Node idNode = null;

		// fix for entity references
		if (contextNode instanceof EntityReference) {
			final Node fixedEntityNode = contextNode.getOwnerDocument().createTextNode("&"+contextNode.getNodeName()+";");
			final org.w3c.dom.ProcessingInstruction pi =
					contextNode.getOwnerDocument().createProcessingInstruction(Result.PI_DISABLE_OUTPUT_ESCAPING,"yes");
			contextNode.getParentNode().insertBefore(pi, contextNode);
			contextNode.getParentNode().replaceChild(fixedEntityNode, contextNode);
			contextNode = fixedEntityNode;
		}

		if (contextNode.hasAttributes()) {
			final NamedNodeMap attributes = contextNode.getAttributes();
			idNode = attributes.getNamedItemNS(idNamespace, idAttributeName);
		}

		final ViewController sourceViewControllerLocal = (ViewController) getNodeUserData(contextNode, "sourceViewController", sourceViewController);
		final ElementDom sourceParentElement = (ElementDom) getNodeUserData(contextNode, "sourceParent", parentElement);

		String id = null;
		String idPrefixLocal = idPrefix;
		if (idNode != null) {
			id = (String) getNodeUserData(idNode, "id", idNode.getNodeValue());
			idPrefixLocal = (String) getNodeUserData(idNode, "idPrefix", null);
			idPrefixLocal = StringUtil.concateStrings(null, idPrefix, idPrefixLocal);
			setNodeUserData(idNode, "idPrefix", idPrefixLocal);
			setNodeUserData(idNode, "idSuffix", idSuffix);
			if (idSuffix != null) {
				idNode.setNodeValue(StringUtil.concateStrings(null, idPrefixLocal, id, "$", idSuffix));
			} else {
				idNode.setNodeValue(StringUtil.concateStrings(null, idPrefixLocal, id));
			}
			id = idNode.getNodeValue();
		}

		final ElementDom contextElement = new ElementDom(sourceViewControllerLocal, sourceParentElement, contextNode, id, idPrefixLocal, idSuffix);

		if (idNode != null && idElements != null) {
			// usual id elements from source file
			if (idElements.containsKey(id)) {
				//throw new DuplicateIdException("id '"+id+"' already definded");
			} else {
				idElements.put(id, contextElement);
			}
		}

		if (parentElement != null) {
			parentElement.addElement(contextElement);
		}

		final NodeList nodes = contextNode.getChildNodes();
		for (int i=0; i<nodes.getLength(); i++) {
			final Node childNode = nodes.item(i);
			processNode(childNode, contextElement, idPrefix, idSuffix, idElements, sourceViewController);
		}
		return contextElement;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Parser#parseXML(java.lang.String, org.jsxp.framework.framework.ResourceResolver, java.lang.Object, java.lang.String, java.util.Map, org.jsxp.framework.framework.internal.InternalElement, org.jsxp.framework.framework.ViewController)
	 */
	@Override
	public InternalElement read(final String uri, final ResourceResolver resolver, final Object contextObjectForResolver, final String idPrefix,
			final Map<String, InternalElement> idElements, final InternalElement rootElement, final ViewController sourceViewController) {
		try {
			Document document = null;
			Node rootNode = null;

			final long modifiedTimeStamp = resolver.getResourceModifiedTimeStamp(uri, contextObjectForResolver);
			if (uri!=null && filenameModifiedTimestamp.containsKey(uri)
					&& filenameModifiedTimestamp.get(uri) == modifiedTimeStamp
					&& filenameDOMCache.containsKey(uri)) {
				Context.getContext().addDebugMessage("cache resolved for view '"+uri+"'");
				document = filenameDOMCache.get(uri);
			} else {
				Context.getContext().addDebugMessage("no cache resolved for view '"+uri+"'");
				final InputStream inputStream = resolver.getResourceInputStream(uri, contextObjectForResolver);
				if(inputStream == null) {
					throw new IllegalStateException("Could not resolve view file name \""+uri+"\" to an input stream.");
				}
				document = parseDocument(inputStream);
				if (uri != null) {
					filenameModifiedTimestamp.put(uri, modifiedTimeStamp);
					filenameDOMCache.put(uri, document);
				}
			}

			if (rootElement != null) {
				if(!(rootElement instanceof ElementDom)) {
					throw new IllegalArgumentException("Root element must be a DomElement");
				}
				final Document rootDocument = ((ElementDom)rootElement).getNode().getOwnerDocument();
				rootNode = rootDocument.importNode(document.getDocumentElement(), true);
				document = rootDocument;
			} else {
				document = (Document) document.cloneNode(true);
				rootNode = document.getDocumentElement();
			}

			final InternalElement newRootElement = processNode(rootNode, null, idPrefix, null, idElements, sourceViewController);
			return newRootElement;
		} catch (final ParserConfigurationException e) {
			Context.getContext().addDebugMessage(e.getLocalizedMessage());
			e.printStackTrace();
			throw new IllegalStateException("Could not complete XML parsing: ", e);
		} catch (final SAXParseException e) {
			Context.getContext().addDebugMessage(e.getLocalizedMessage());
			e.printStackTrace();
			throw new IllegalStateException("Could not complete XML parsing: ", e);
		} catch (final SAXException e) {
			Context.getContext().addDebugMessage(e.getLocalizedMessage());
			e.printStackTrace();
			throw new IllegalStateException("Could not complete XML parsing: ", e);
		} catch (final IOException e) {
			Context.getContext().addDebugMessage(e.getLocalizedMessage());
			e.printStackTrace();
			throw new IllegalStateException("Could not complete XML parsing: ", e);
		} catch (final DuplicateIdException e) {
			Context.getContext().addDebugMessage(e.getLocalizedMessage());
			e.printStackTrace();
			throw new IllegalStateException("Could not complete XML parsing: ", e);
		}
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Parser#setIdAttributeName(java.lang.String)
	 */
	@Override
	public void setIdAttributeName(final String idAttributeName) {
		this.idAttributeName = idAttributeName;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Parser#setIdNamespace(java.lang.String)
	 */
	@Override
	public void setIdNamespace(final String idNamespace) {
		this.idNamespace = idNamespace;
	}

	private void setNodeUserData(final Node node, final String key, final Object value) {
		node.setUserData(key, value, new UserDataHandler() {
			@Override
			public void handle(final short operation, final String key, final Object data, final Node src, final Node dst) {
				final Object value = src.getUserData(key);
				dst.setUserData(key, value, this);
			}
		});
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.Parser#codeXML(org.jsxp.framework.framework.internal.InternalElement, java.io.OutputStream)
	 */
	@Override
	public void write(InternalElement elem, final OutputStream os, final String contentType) {
		if(elem instanceof ProxyElement) {
			elem = ((ProxyElement)elem).getRealElement();
		}
		if(!(elem instanceof ElementDom)) {
			throw new IllegalArgumentException("Only DomElement instances are allowed for coding XML with this class.");
		}

		final ElementDom element = (ElementDom) elem;
		final Node node = element.getNode();

		final XMLTransformer xmlTransformer = AdapterFinder.find(XMLTransformer.class, null);
		xmlTransformer.write(node, os, contentType);
	}

}
