/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.dom;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;

import javax.xml.XMLConstants;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.dom.DOMSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.jsxp.framework.dom.adapter.XMLTransformer;
import org.jsxp.framework.util.adapter.AdapterFinder;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;


/**
 * The Class XMLProcessing. Used by the {@link ViewParserDom} implementation to process a xml file.
 */
public class XMLProcessing {

	/**
	 * Gets the node.
	 * 
	 * @param node the node
	 * @param expression the expression
	 * 
	 * @return the node
	 * 
	 * @throws XPathExpressionException the x path expression exception
	 */
	protected static Node getNode(final Node node, final String expression)
			throws XPathExpressionException {
		final XPath xpath = XPathFactory.newInstance().newXPath();
		final Node resultNode = (Node) xpath.evaluate(expression, node,
				XPathConstants.NODE);
		return resultNode;
	}

	/**
	 * Gets the nodes.
	 * 
	 * @param document the document
	 * @param expression the expression
	 * 
	 * @return the nodes
	 * 
	 * @throws XPathExpressionException the x path expression exception
	 */
	protected static NodeList getNodes(final Document document, final String expression)
			throws XPathExpressionException {
		final XPath xpath = XPathFactory.newInstance().newXPath();
		final NodeList nodes = (NodeList) xpath.evaluate(expression, document,
				XPathConstants.NODESET);
		return nodes;
	}

	/**
	 * Parses the document.
	 * 
	 * @param inputStream the input stream
	 * 
	 * @return the document
	 * 
	 * @throws ParserConfigurationException the parser configuration exception
	 * @throws SAXException the SAX exception
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	protected static Document parseDocument(final InputStream inputStream) throws ParserConfigurationException, SAXException, IOException {
		final XMLTransformer xmlTransformer = AdapterFinder.find(XMLTransformer.class, null);
		return xmlTransformer.read(inputStream);
	}

	/**
	 * Prints the file.
	 * 
	 * @param filename the filename
	 * 
	 * @throws FileNotFoundException the file not found exception
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	protected static void printFile(final String filename)
			throws FileNotFoundException, IOException {
		final FileReader fr = new FileReader(filename);
		final BufferedReader br = new BufferedReader(fr);
		while (br.ready()) {
			System.out.println(br.readLine());
		}
		br.close();
	}

	/**
	 * Validate.
	 * 
	 * @param document the document
	 * 
	 * @throws MalformedURLException the malformed url exception
	 * @throws SAXException the SAX exception
	 * @throws IOException Signals that an I/O exception has occurred.
	 */
	public static void validate(final Document document) throws MalformedURLException,
	SAXException, IOException {
		// Create a SchemaFactory capable of understanding WXS schemas.
		final SchemaFactory factory =
				SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

		// Load a WXS schema, represented by a Schema instance.
		//Source schemaFile = new StreamSource(new File("mySchema.xsd"));
		final URL url = new URL("http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd");

		final Schema schema = factory.newSchema(url);

		// Create a Validator object, which can be used to validate
		// an instance document.
		final Validator validator = schema.newValidator();

		// Validate the DOM tree.
		validator.validate(new DOMSource(document));
	}

	/**
	 * Instantiates a new xML processing.
	 */
	public XMLProcessing() {
		super();
	}

}