/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.elementprocessing.processors;

import java.util.HashMap;

import org.jsxp.framework.Context;
import org.jsxp.framework.Element;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.elementprocessing.ElementProcessor;
import org.jsxp.framework.elementprocessing.ElementProcessorCommand;
import org.jsxp.framework.util.text.StringUtils;
import org.jsxp.framework.webroot.jsxpframework.components.ajax.AjaxXhtmlController;


/**
 * This processor will enable Elements with Ajax behaviour.
 * <br/>
 * <br/>
 * It an be configured for partial request, so a request will be made
 * for example when the element is clicked and the normal lifecycle of
 * view controller will be used on server.
 * <br/>
 * <br/>
 * It can also be used with a {@link ElementProcessorCommand}. This is
 * a command that will be called and executed on the server, when the
 * Ajax behaviour is used. This is a much easier way of implementing server side
 * java behaviour for ajax requests.
 * <br/>
 * <br/>
 * For advanced configuration set the different properties.
 */
public class AjaxBehaviour implements ElementProcessor {

	private String body = null;
	private String callbackFunction = null;
	private String contentType = null;
	private Element element = null;
	private ElementProcessorCommand elementProcessorCommand = null;
	private HashMap<String, ElementProcessorCommand> elementProcessorCommands = null;
	private String event = "onClick";
	private int maxWaitingTimeMS = 5000;
	private String method = "post";
	private String[] updateIds;
	private String uri = "#";

	/**
	 * Instantiates a new resource handler.
	 * 
	 * @param element
	 * @param elementProcessorCommand
	 * @param updateIds the jsxp id to update on the client. leave blank if all should be updated
	 */
	public AjaxBehaviour(final Element element, final ElementProcessorCommand elementProcessorCommand, final String... updateIds) {
		this.element = element;
		this.updateIds = updateIds;
		this.elementProcessorCommand = elementProcessorCommand;
	}

	/**
	 * Instantiates a new resource handler. As default the event is onClick.
	 * @param element the element to process
	 * @param updateIds the jsxp id to update on the client. leave blank if all should be updated
	 */
	public AjaxBehaviour(final Element element, final String... updateIds) {
		this.element = element;
		this.updateIds = updateIds;
	}

	/**
	 * Instantiates a new resource handler.
	 * 
	 * @param element
	 * @param elementProcessorCommand
	 * @param event like "onClick" and so on
	 * @param updateIds the jsxp id to update on the client. leave blank if all should be updated
	 */
	public AjaxBehaviour(final String event, final Element element, final ElementProcessorCommand elementProcessorCommand, final String... updateIds) {
		this.element = element;
		this.updateIds = updateIds;
		this.elementProcessorCommand = elementProcessorCommand;
		this.event = event;
	}

	/**
	 * Instantiates a new resource handler.
	 * @param element the element to process
	 * @param event like "onClick" and so on
	 * @param updateIds the jsxp id to update on the client. leave blank if all should be updated
	 */
	public AjaxBehaviour(final String event, final Element element, final String... updateIds) {
		this.element = element;
		this.updateIds = updateIds;
		this.event = event;
	}

	public String getBody() {
		return body;
	}

	public String getCallbackFunction() {
		return callbackFunction;
	}

	public String getContentType() {
		return contentType;
	}

	public String getEvent() {
		return event;
	}

	public int getMaxWaitingTimeMS() {
		return maxWaitingTimeMS;
	}

	public String getMethod() {
		return method;
	}

	public String[] getUpdateIds() {
		return updateIds;
	}

	public String getUri() {
		return uri;
	}

	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.elementprocessing.ElementProcessor#process(org.jsxp.framework.ViewController)
	 */
	@Override
	@SuppressWarnings("unchecked")
	public void process(final ViewController renderViewController) {
		if (element == null) {
			return;
		}

		try {
			renderViewController.importView(new AjaxXhtmlController());
		} catch (final Exception e) {
			throw new RuntimeException(e);
		}

		String commandId = null;
		if (elementProcessorCommand != null) {
			// get the map for commands from session
			elementProcessorCommands = (HashMap<String, ElementProcessorCommand>) Context.getContext().getScopes().getSessionAttributes().get(ElementProcessorCommand.class.getName());
			if (elementProcessorCommands == null) {
				elementProcessorCommands = new HashMap<String, ElementProcessorCommand>();
				Context.getContext().getScopes().getSessionAttributes().put(ElementProcessorCommand.class.getName(), elementProcessorCommands);
			}

			commandId = elementProcessorCommand.toString();
			elementProcessorCommands.put(commandId, elementProcessorCommand);
		}

		// function(method, uri, body, contentType, maxWaitingTime, callbackFunction, updateIds)
		final StringBuffer sb = new StringBuffer();
		sb.append("org.jsxp.js.createAJAXSupport().doRequest(");
		sb.append("'").append(method).append("'");
		sb.append(", ");
		sb.append("'").append(uri).append("'");
		sb.append(", ");
		if (commandId != null) { sb.append("'").append(commandId).append("'"); }
		else { sb.append("null"); }
		sb.append(", ");
		if (body != null) {	sb.append("'").append(body).append("'"); }
		else { sb.append("null"); }
		sb.append(", ");
		if (contentType != null) {	sb.append("'").append(contentType).append("'"); }
		else { sb.append("null"); }
		sb.append(", ");
		sb.append(maxWaitingTimeMS);
		sb.append(", ");
		if (callbackFunction != null) {	sb.append("'").append(callbackFunction).append("'"); }
		else { sb.append("null"); }

		sb.append(", ");
		sb.append("'").append(StringUtils.combine(",", false, (Object[])updateIds)).append("'");
		sb.append("); return false;");

		element.setAttribute(event, sb.toString());
	}

	public void setBody(final String body) {
		this.body = body;
	}

	public void setCallbackFunction(final String callbackFunction) {
		this.callbackFunction = callbackFunction;
	}

	public void setContentType(final String contentType) {
		this.contentType = contentType;
	}

	public void setEvent(final String event) {
		this.event = event;
	}

	public void setMaxWaitingTimeMS(final int maxWaitingTimeMS) {
		this.maxWaitingTimeMS = maxWaitingTimeMS;
	}

	public void setMethod(final String method) {
		this.method = method;
	}

	public void setUpdateIds(final String[] updateIds) {
		this.updateIds = updateIds;
	}

	public void setUri(final String uri) {
		this.uri = uri;
	}
}
