/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.elementprocessing.processors;

import java.util.HashSet;
import java.util.Set;

import org.jsxp.framework.Context;
import org.jsxp.framework.Element;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.elementprocessing.ElementProcessor;


/**
 * This
 * processor renders a resource that a view controller uses as a link in the HTML-head. This
 * resource handler can be used for stylesheets or javascript (text/css, text/javascript resource type) files.
 */
public class HeadLink implements ElementProcessor {

	private Set<String> alreadyProcessedKeys = null;
	private Element element = null;
	private String key = null;

	/**
	 * Instantiates a new resource handler.
	 * @param element the element to process
	 * @param key used to check what keys are already processed so we can prevent to add stuff in head twices (most time you can use the link uri)
	 */
	@SuppressWarnings("unchecked")
	public HeadLink(final Element element, final String key) {
		this.element = element;
		this.key = key;

		// used to check what keys are already processed so we can prevent to add stuff in head twices
		alreadyProcessedKeys = (Set<String>) Context.getContext().getScopes().getRequestAttributes().get(this.getClass().getName());
		if (alreadyProcessedKeys == null) {
			alreadyProcessedKeys = new HashSet<String>();
			Context.getContext().getScopes().getRequestAttributes().put(this.getClass().getName(), alreadyProcessedKeys);
		}
	}


	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.elementprocessing.ElementProcessor#process(org.jsxp.framework.ViewController)
	 */
	@Override
	public void process(final ViewController renderViewController) {
		final Element target = renderViewController.getRootElement().findElementByName("head", null);

		if (target == null) {
			Context.getContext().addDeveloperWarningMessage("no head html element found in view controller: "+renderViewController);
			return;
		}

		// check that an element is not added twice to the head
		if (alreadyProcessedKeys.contains(key)) {
			return;
		}
		alreadyProcessedKeys.add(key);

		target.addElement(element);
	}
}
