/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.elementprocessing.processors;

import java.util.ResourceBundle;

import org.jsxp.framework.Context;
import org.jsxp.framework.Element;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.elementprocessing.ElementProcessor;
import org.jsxp.framework.util.text.StringUtil;


/**
 * Defines how resource are handled by the jsxp framework. This processor will exchange the
 * value of an element to content from a resource bundle. The resource bundle will be
 * resolved by the uri.
 * <br/><br/>
 * The content will be resolved from the resource bundle by the element ID as a key (or a
 * provided key) and the current locale from context.
 */
public class ResourceBundleValue implements ElementProcessor {

	private Element element = null;
	private String key = null;
	private String uri = null;

	/**
	 * Instantiates a new resource handler.
	 * @param element the element to process
	 * @param uri uri to resolve the resource bundle
	 */
	public ResourceBundleValue(final Element element, final String uri) {
		this.element = element;
		this.uri = uri;
	}

	/**
	 * Instantiates a new resource handler.
	 * @param element the element to process
	 * @param uri uri to resolve the resource bundle
	 * @param key the key of the object in the resource bundle
	 */
	public ResourceBundleValue(final Element element, final String uri, final String key) {
		this.element = element;
		this.uri = uri;
		this.key = key;
	}

	/*
	 * (non-Javadoc)
	 * @see org.jsxp.framework.resources.ElementHandler#processElement(org.jsxp.framework.resources.Resource)
	 */
	@Override
	public void process(final ViewController renderViewController) {
		if (key == null) {
			key = element.getId();
		}

		final ResourceBundle resourceBundle = ResourceBundle.getBundle(uri, Context.getContext().getCurrentLocale());
		Object value = null;

		try {
			value = resourceBundle.getObject(key+"[]");
		} catch (final Exception e) {
			// ignore and try without array
		}

		if (value != null) {
			if (value instanceof String) {
				final String[] values = StringUtil.split((String)value, '|');
				value = values;
			}
		} else {
			value = resourceBundle.getObject(key);
		}

		if (value instanceof String) {
			element.setValue((String) value);
		} else if (value instanceof String[]) {
			element.setValues((String[]) value);
		} else if (value != null){
			element.setValue(value.toString());
		}
	}

}
