/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.webroot.jsxpframework.components;

import org.jsxp.framework.ElementTemplate;
import org.jsxp.framework.PhaseListener;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.ViewControllerLifeCycle.LifeCyclePhase;
import org.jsxp.framework.webroot.generated.jsxpframework.components.PagerXhtmlControllerGenerated;


/**
 * The Class PagerXhtmlController. {@link ViewController} for the basic jsxp pager component.
 * You can build pageable list in your view with this pager. If used on a page this pager
 * will automatically process a given "page" input parameter to calculate the current
 * position of the pager. Also it can be used for the result output and will render a pager
 * (with links for the user to browse through the pageable list.
 * You can configure the pager by using the different setters.
 */
public class PagerXhtmlController extends PagerXhtmlControllerGenerated {

	/**
	 * The Class InputParams.
	 */
	public class InputParams {
		private String page;

		/**
		 * Gets the page.
		 * 
		 * @return the page
		 */
		public String getPage() {
			return page;
		}

		/**
		 * Sets the page.
		 * 
		 * @param page the new page
		 */
		public void setPage(final String page) {
			this.page = page;
		}
	}

	private String backText = "Back";

	private int count = 100;
	private int currentPage;
	private int firstPage = 1;
	private String firstText = "First";
	private final InputParams input = new InputParams();
	private int lastPage = 1;
	private String lastText = "Last";
	private String nextText = "Next";
	private int pageSize = 10;

	private String query = "";

	/**
	 * Gets the back text.
	 * 
	 * @return the back text
	 */
	public String getBackText() {
		return backText;
	}

	/**
	 * Gets the count.
	 * 
	 * @return the count
	 */
	public int getCount() {
		return count;
	}

	/**
	 * Gets the current page.
	 * 
	 * @return the current page
	 */
	public int getCurrentPage() {
		return currentPage;
	}

	/**
	 * Gets the first text.
	 * 
	 * @return the first text
	 */
	public String getFirstText() {
		return firstText;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.ViewController#getInputParameter()
	 */
	@Override
	public InputParams getInputParameterBean() {
		return input;
	}

	/**
	 * Gets the last element.
	 * 
	 * @return the last element
	 */
	public int getLastElement() {
		int lastElement = getStartElement() + pageSize;
		if (lastElement > count) {
			lastElement = count;
		}
		return lastElement;
	}

	/**
	 * Gets the last text.
	 * 
	 * @return the last text
	 */
	public String getLastText() {
		return lastText;
	}

	/**
	 * Gets the next text.
	 * 
	 * @return the next text
	 */
	public String getNextText() {
		return nextText;
	}

	/**
	 * Gets the page size.
	 * 
	 * @return the page size
	 */
	public int getPageSize() {
		return pageSize;
	}

	/**
	 * Gets the query.
	 * 
	 * @return the query
	 */
	public String getQuery() {
		return query;
	}

	/**
	 * Gets the start element.
	 * 
	 * @return the start element
	 */
	public int getStartElement() {
		return (currentPage-1)*pageSize;
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.ViewController#init()
	 */
	@Override
	public void init() {
		getLifeCycle().addPhaseListener(LifeCyclePhase.VARIABLE_SUBSTITUTION, new PhaseListener() {

			@Override
			public void afterPhase(final LifeCyclePhase phase) {
			}

			@Override
			public void beforePhase(final LifeCyclePhase phase) {
				processVariables();
			}
		});

		initPager();
	}

	private void initPager() {
		setNeedToSaveScopesToSession(false);

		currentPage = 1;
		if(input.page != null && input.page.matches("\\d+")) {
			currentPage=Integer.parseInt(input.page);
		}

		firstPage = 1;
		lastPage =  (int) (Math.ceil(((double)count)/pageSize));
		if(currentPage > lastPage) {
			currentPage = lastPage;
		}
		if(currentPage < firstPage) {
			currentPage = firstPage;
		}
	}

	private void processVariables() {
		initPager();

		getElementLinkPrevious().setVariablePreviousText(backText);
		getElementLinkFirst().setVariableFirstText(firstText);
		getElementLinkNext().setVariableNextText(nextText);
		getElementLinkLast().setVariableLastText(lastText);

		int firstShownPage = Math.max(firstPage, currentPage-2);
		final int lastShownPage = Math.min(firstShownPage+4, lastPage);
		if(firstShownPage-lastShownPage < 4) {
			firstShownPage = Math.max(firstPage, lastShownPage-4);
		}

		getElementLinkFirst().setVariableFirst(Long.toString(firstPage));
		getElementLinkPrevious().setVariablePrevious(Long.toString(Math.max(1, currentPage-1)));
		getElementLinkNext().setVariableNext(Long.toString(Math.min(lastPage, currentPage+1)));
		getElementLinkLast().setVariableLast(Long.toString(lastPage));
		getElementSimplePager().setVariableQuery(query);

		final ElementTemplate<PagerActivePage> activePageTemplate = getElementPagerActivePage().createTemplate();
		final ElementTemplate<PagerInactivePage> inactivePageTemplate = getElementPagerInactivePage().createTemplate();
		getElementPagesList().removeElements();
		for(long i = firstShownPage; i<=lastShownPage; i++) {
			if(i==currentPage) {
				final PagerActivePage page = activePageTemplate.createElement();
				page.setVariablePageText(Long.toString(i));
				page.setVariablePage(Long.toString(i));
				getElementPagesList().addElement(page);
			} else {
				final PagerInactivePage page = inactivePageTemplate.createElement();
				page.setVariablePageText(Long.toString(i));
				page.setVariablePage(Long.toString(i));
				getElementPagesList().addElement(page);
			}
		}

	}

	/**
	 * Sets the back text.
	 * 
	 * @param backText the new back text
	 */
	public void setBackText(final String backText) {
		this.backText = backText;
	}

	/**
	 * Sets the count.
	 * 
	 * @param last the new count
	 */
	public void setCount(final int last) {
		this.count = last;
	}

	/**
	 * Sets the first text.
	 * 
	 * @param firstText the new first text
	 */
	public void setFirstText(final String firstText) {
		this.firstText = firstText;
	}

	/**
	 * Sets the last text.
	 * 
	 * @param lastText the new last text
	 */
	public void setLastText(final String lastText) {
		this.lastText = lastText;
	}

	/**
	 * Sets the next text.
	 * 
	 * @param nextText the new next text
	 */
	public void setNextText(final String nextText) {
		this.nextText = nextText;
	}

	/**
	 * Sets the page size.
	 * 
	 * @param pageSize the new page size
	 */
	public void setPageSize(final int pageSize) {
		this.pageSize = pageSize;
	}

	/**
	 * Sets the query.
	 * 
	 * @param query the new query
	 */
	public void setQuery(final String query) {
		this.query = query;
	}
}
