/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.webroot.jsxpframework.components;

import java.util.Map;

import org.jsxp.framework.Context;
import org.jsxp.framework.Element;
import org.jsxp.framework.PhaseListener;
import org.jsxp.framework.ViewController;
import org.jsxp.framework.ViewControllerLifeCycle.LifeCyclePhase;
import org.jsxp.framework.webroot.generated.jsxpframework.components.ScopesMessagesXhtmlControllerGenerated;


/**
 * The Class ScopesMessagesXhtmlController. Standard {@link ViewController} for
 * show all content from all scopes in different message containers on the resulting view.
 * You can embedded this component to your view to show the scopes with their content.
 * This component uses the {@link MessagesXhtmlController} to render the scopes.
 */
public class ScopesMessagesXhtmlController extends ScopesMessagesXhtmlControllerGenerated {

	private class ScopeMessagesPhaseListenerRendering implements PhaseListener {

		/* (non-Javadoc)
		 * @see org.jsxp.framework.framework.PhaseListener#afterPhase(org.jsxp.framework.framework.ViewControllerLifeCycle.LifeCyclePhase)
		 */
		@Override
		public void afterPhase(final LifeCyclePhase phase) {
		}

		/* (non-Javadoc)
		 * @see org.jsxp.framework.framework.PhaseListener#beforePhase(org.jsxp.framework.framework.ViewControllerLifeCycle.LifeCyclePhase)
		 */
		@Override
		public void beforePhase(final LifeCyclePhase phase) {
			if (!Context.getContext().getUpdateIds().isEmpty()) {
				Context.getContext().getUpdateIds().add(getElementScopes().getId());
			}

			ViewController ViewController = getParent();
			while (ViewController.getViewTemplate() != null) {
				ViewController = ViewController.getViewTemplate();
			}
			if(!Context.getContext().getDebugMessages().isEmpty()) {
				final Element rootElement = ViewController.getRootElement();
				Element element = rootElement.findElementById("jsxp:scopes");

				if(element == null) {
					element = rootElement.findElementByName("body", null);
				}

				if(element != null) {
					element.addElement(getElementScopes());
				}
			}
		}
	}
	private class ScopeMessagesPhaseListenerVariableSubstitution implements PhaseListener {

		/* (non-Javadoc)
		 * @see org.jsxp.framework.framework.PhaseListener#afterPhase(org.jsxp.framework.framework.ViewControllerLifeCycle.LifeCyclePhase)
		 */
		@Override
		public void afterPhase(final LifeCyclePhase phase) {
		}

		/* (non-Javadoc)
		 * @see org.jsxp.framework.framework.PhaseListener#beforePhase(org.jsxp.framework.framework.ViewControllerLifeCycle.LifeCyclePhase)
		 */
		@Override
		public void beforePhase(final LifeCyclePhase phase) {
			processScope(Context.getContext().getScopes().getApplicationScope(), getElementApplicationScope(), applicationScopeMessagesViewController);
			processScope(Context.getContext().getScopes().getSessionScope(), getElementSessionScope(), sessionScopeMessagesViewController);
			processScope(Context.getContext().getScopes().getRequestScope(), getElementRequestScope(), requestScopeMessagesViewController);
			processScope(Context.getContext().getScopes().getViewScope(), getElementViewScope(), viewScopeMessagesViewController);
			processScope(Context.getContext().getScopes().getFlashScope(), getElementFlashScope(), flashScopeMessagesViewController);
		}

		private void processScope(final Map<String, Object> scope, final Element target, final MessagesXhtmlController view) {
			if (scope.keySet() != null && !scope.keySet().isEmpty()) {
				for(final String k : scope.keySet()) {
					final Object value = scope.get(k);
					view.getMessages().add(k+": "+value);
				}
				target.exchange(view.getElementMessagesComponent());
			}
		}
	}
	private MessagesXhtmlController applicationScopeMessagesViewController = null;
	private MessagesXhtmlController flashScopeMessagesViewController = null;
	private MessagesXhtmlController requestScopeMessagesViewController = null;

	private MessagesXhtmlController sessionScopeMessagesViewController = null;

	private MessagesXhtmlController viewScopeMessagesViewController = null;

	/**
	 * Instantiates a new scopes messages xhtml controller.
	 */
	public ScopesMessagesXhtmlController() {
		setNeedToSaveScopesToSession(false);
		getLifeCycle().addPhaseListener(LifeCyclePhase.VARIABLE_SUBSTITUTION, new ScopeMessagesPhaseListenerVariableSubstitution());
		getLifeCycle().addPhaseListener(LifeCyclePhase.RENDERING, new ScopeMessagesPhaseListenerRendering());
	}

	/* (non-Javadoc)
	 * @see org.jsxp.framework.framework.ViewController#init()
	 */
	@Override
	protected void init() throws Exception {
		applicationScopeMessagesViewController = importView(new MessagesXhtmlController());
		applicationScopeMessagesViewController.getElementHeadline().setValue("Application scope");
		sessionScopeMessagesViewController = importView(new MessagesXhtmlController());
		sessionScopeMessagesViewController.getElementHeadline().setValue("Session scope");
		requestScopeMessagesViewController = importView(new MessagesXhtmlController());
		requestScopeMessagesViewController.getElementHeadline().setValue("Request scope");
		viewScopeMessagesViewController = importView(new MessagesXhtmlController());
		viewScopeMessagesViewController.getElementHeadline().setValue("View scope");
		flashScopeMessagesViewController  = importView(new MessagesXhtmlController());
		flashScopeMessagesViewController.getElementHeadline().setValue("Flash scope");
	}

}
