/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.logging;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.logging.Level;

/**
 * Simple generator logger which logs all messages for a class.
 * Can be used by defining a Logger for each using class or by just using
 * the static methods and giving them a object or class as parameter.
 * If you use it in the second way, you don't need to define a logger
 * for each class any more.
 * @author Robert Mencl
 */
public class Log
{
	protected static final PrintWriter ERR = new PrintWriter(System.err);

	private static Map<Class<?>, Log> instances = new HashMap<Class<?>, Log>();
	protected static List<LogEntry> log = new ArrayList<LogEntry>();

	protected static final PrintWriter OUT = new PrintWriter(System.out);

	/**
	 * Clears the log entry protocol
	 */
	public static void clearLogEntryProtocol()
	{
		log.clear();
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 */
	public static void debug(final Object thiss, final String message)
	{
		debug(thiss, message, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param params
	 */
	public static void debug(final Object thiss, final String message, final Object... params)
	{
		insertMessage(thiss, LogLevel.DEBUG, message, null, params);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 */
	public static void error(final Object thiss, final String message)
	{
		error(thiss, message, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param params
	 */
	public static void error(final Object thiss, final String message, final Object... params)
	{
		insertMessage(thiss, Level.SEVERE, message, null, params);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param t
	 */
	public static void error(final Object thiss, final String message, final Throwable t)
	{
		error(thiss, message, t, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param t
	 * @param params
	 */
	public static void error(final Object thiss, final String message, final Throwable t, final Object... params)
	{
		insertMessage(thiss, Level.SEVERE, message, t, params);
	}

	/**
	 * adds a log enty
	 * @param <T>
	 * @param thiss
	 * @param message
	 * @param t
	 * @param raiseError
	 * @throws T
	 */
	public static <T extends Throwable> void errorAndRaise(final Object thiss, final String message, final T t, final boolean raiseError) throws T
	{
		errorAndRaise(thiss, message, t, raiseError, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param <T>
	 * @param thiss
	 * @param message
	 * @param t
	 * @param raiseError
	 * @param params
	 * @throws T
	 */
	public static <T extends Throwable> void errorAndRaise(final Object thiss, final String message, final T t, final boolean raiseError, final Object... params) throws T
	{
		insertMessage(thiss, Level.SEVERE, message, t, params);
		if (raiseError)
		{
			throw t;
		}
	}

	public static void flush()
	{
		ERR.flush();
		OUT.flush();
	}

	/**
	 * @param class for that a log object is needed
	 * @return a log object for the given class
	 */
	synchronized public static Log getInstance(final Class<?> clazz)
	{
		Log instance = instances.get(clazz);

		if (instance == null)
		{
			// create an instance:
			instance = new Log(clazz);
			// take the instance from the map
			instance = instances.get(clazz);
			if (instance == null)
			{
				ERR.println("DID NOT find the correct instance of class" + clazz.getCanonicalName() + " within generator map.");
			}
		}

		return instance;
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 */
	public static void info(final Object thiss, final String message)
	{
		info(thiss, message, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param params
	 */
	public static void info(final Object thiss, final String message, final Object... params)
	{
		insertMessage(thiss, Level.INFO, message, null, params);
	}

	/**
	 * adds a log enty
	 * @param source
	 * @param level
	 * @param message
	 * @param t
	 * @param params
	 */
	public static void insertMessage(final Object source, final Level level, final String message, final Throwable t, final Object... params)
	{
		if (source instanceof Class<?>)
		{
			getInstance((Class<?>)source).insertMessage(level, message, null, params);
		}
		else
		{
			getInstance(source.getClass()).insertMessage(level, message, null, params);
		}
	}
	/**
	 * Creates a collection of {@link LogEntry} objects
	 * from the log entry protocol.
	 * @param minLevel the minimum level of the included log entries
	 * @param maxLevel the maximum level of the included log entries
	 * @param unique If unique is set to true, than only one log entry
	 *  will be included if log entries
	 * 	have equal string representations ({@link LogEntry#toString()})
	 * @return warnings and error messages as {@link LogEntry} objects
	 * 
	 */
	public static Collection<LogEntry> messagesFromLogEntryProtocol(final Level minLevel, final Level maxLevel, final boolean unique)
	{
		final SortedSet<String> uniqueMessages = new TreeSet<String>();

		final ArrayList<LogEntry> result = new ArrayList<LogEntry>();

		for (final LogEntry entry : log)
		{
			if ((minLevel == null || entry.getLevel().intValue() >= minLevel.intValue())
					&& (maxLevel == null || entry.getLevel().intValue() <= maxLevel.intValue()))
			{
				if (!unique)
				{
					result.add(entry);
				}
				else if (!uniqueMessages.contains(entry.toString()))
				{
					result.add(entry);
					//
					uniqueMessages.add(entry.toString());
				}
			}
		}

		return result;
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 */
	public static void warn(final Object thiss, final String message)
	{
		warn(thiss, message, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param params
	 */
	public static void warn(final Object thiss, final String message, final Object... params)
	{
		insertMessage(thiss, Level.WARNING, message, null, params);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param t
	 */
	public static void warn(final Object thiss, final String message, final Throwable t)
	{
		warn(thiss, message, t, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param thiss
	 * @param message
	 * @param t
	 * @param params
	 */
	public static void warn(final Object thiss, final String message, final Throwable t, final Object... params)
	{
		insertMessage(thiss, Level.WARNING, message, t, params);
	}

	/**
	 * Creates a collection of {@link LogEntry} objects
	 * from the log entry protocol. Warning and error will
	 * be included.
	 * @param unique If unique is set to true, than only one log entry
	 *  will be included if log entries
	 * 	have equal string representations ({@link LogEntry#toString()})
	 * @return warnings and error messages as {@link LogEntry} objects
	 * 
	 */
	public static Collection<LogEntry> warningsAndErrorMessagesFromLogEntryProtocol(final boolean unique)
	{
		return messagesFromLogEntryProtocol(Level.WARNING, null, unique);
	}

	protected Class<?> clazz = null;

	private Log(final Class<?> clazz)
	{
		instances.put(clazz,this);
		this.clazz = clazz;
	}

	/**
	 * adds a log enty
	 * @param message
	 */
	public void debug(final String message)
	{
		debug(message, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param params
	 */
	public void debug(final String message, final Object... params)
	{
		insertMessage(LogLevel.DEBUG, message, null, params);
	}

	/**
	 * adds a log enty
	 * @param message
	 */
	public void error(final String message)
	{
		error(message, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param params
	 */
	public void error(final String message, final Object... params)
	{
		error(message, (Throwable)null, params);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param t
	 */
	public void error(final String message, final Throwable t)
	{
		error(message, t, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param t
	 * @param params
	 */
	public void error(final String message, final Throwable t, final Object... params)
	{
		insertMessage(Level.SEVERE, message, t, params);
	}

	/**
	 * adds a log enty
	 * @param <T>
	 * @param message
	 * @param t
	 * @param raiseError
	 * @throws T
	 */
	public <T extends Throwable> void errorAndRaise(final String message, final T t, final boolean raiseError) throws T
	{
		errorAndRaise(message, t, raiseError, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param <T>
	 * @param message
	 * @param t
	 * @param raiseError
	 * @param params
	 * @throws T
	 */
	public <T extends Throwable> void errorAndRaise(final String message, final T t, final boolean raiseError, final Object... params) throws T
	{
		insertMessage(Level.SEVERE, message, t, params);
		if (raiseError)
		{
			throw t;
		}
	}

	/**
	 * adds a log enty
	 * @param message
	 */
	public void info(final String message)
	{
		info(message, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param params
	 */
	public void info(final String message, final Object... params)
	{
		insertMessage(Level.INFO, message, null, params);
	}

	/**
	 * adds a log enty
	 * @param level
	 * @param message
	 * @param t
	 * @param params
	 */
	public void insertMessage(final Level level, final String message, final Throwable t, final Object... params)
	{
		final LogEntry entry = new LogEntry(level, clazz, message, t, params);
		log.add(entry);

		if (level.intValue() <= Level.INFO.intValue()) { OUT.println(entry); OUT.flush(); }
		else                                           { ERR.println(entry); ERR.flush(); }
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param t
	 * @param params
	 */
	public void warn(final String message)
	{
		warn(message, (Throwable)null, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param params
	 */
	public void warn(final String message, final Object... params)
	{
		warn(message, (Throwable)null, params);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param t
	 */
	public void warn(final String message, final Throwable t)
	{
		warn(message, t, (Object[]) null);
	}

	/**
	 * adds a log enty
	 * @param message
	 * @param t
	 * @param params
	 */
	public void warn(final String message, final Throwable t, final Object... params)
	{
		insertMessage(Level.WARNING, message, t, params);
	}
}
