/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.logging;

import java.util.logging.Level;

/**
 * Simple log entry.<br/><br/>
 * Supports parameters that are stored. But the {@link #toString()} method
 * will only use the {@link Object#toString()} values from the parameters
 * from the time of the {@link LogEntry} creation.
 * @author Oliver Szymanski
 */
public class LogEntry
{
	protected Class<?>  clazz;
	protected Level     level;
	protected String    message;
	protected Object[]  params;
	protected Throwable t;
	protected String    toStringCache;

	/**
	 * Creates
	 * @param level the log level
	 * @param clazz the logging class
	 * @param message the log message
	 * @param t an optional throwable
	 * @param params optional parameter objects that are used to create the complete log message in this constructor and are stored for later analysis.
	 */
	public LogEntry(final Level level, final Class<?> clazz, final String message, final Throwable t, final Object... params)
	{
		this.level = level;
		this.clazz = clazz;
		this.message = message;
		this.t = t;
		this.params = params;

		createToStringCache(level, clazz, message, params);
	}

	/**
	 * Create the value for calls to the {@link #toString()} method at the
	 * time of creation of the {@link LogEntry}
	 * @param level
	 * @param clazz
	 * @param message
	 * @param params
	 */
	private void createToStringCache(final Level level, final Class<?> clazz,
			final String message, final Object... params) {
		if (params != null && params.length > 0)
		{
			final StringBuffer sb = new StringBuffer();
			if (message != null)
			{
				sb.append(message);
				if (params != null && params.length > 0) { sb.append("\n"); }
			}

			boolean first = true;
			for (final Object object : params)
			{
				if (!first) { sb.append("\n"); }
				else        { first = false;   }

				sb.append(String.valueOf(object));
			}

			toStringCache = sb.toString();
		}
		else
		{
			toStringCache = message;
		}
		toStringCache = "[" + level + "]" + " {" + clazz.getSimpleName() + "}: " + toStringCache;
	}

	/**
	 * 
	 * @return the log level
	 */
	public Level getLevel()
	{
		return this.level;
	}

	/**
	 * 
	 * @return the parameter objects.
	 * the state of this objects may have been
	 * changed when this method is called.
	 * 
	 */
	public Object[] getParams()
	{
		return params;
	}

	@Override
	public String toString()
	{
		return toStringCache;
	}
}
