/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.util.adapter;

import org.jsxp.framework.logging.Log;

/**
 * Use to find an adapter class. Helpful if you just want to define an interface
 * and leave the actual implementation to be configurable.
 */
public class AdapterFinder
{
	/**
	 * Finds a class that fits the implementation, either by the given
	 * implementationQualifiedClassName or if that name is null
	 * by trying to find a class factory that matches the given class name and package
	 * with an additional package prefix of "com.jsxp.framework.adapter.impl." and the class factory
	 * name suffix "Factory". If no factory is found it tries to find an adapter class same way
	 * as with the factory without the suffix. If still no adapter was found and
	 * the desired class is not an interface it tries to create an instane of that class
	 * as a default implementation.
	 * <br/><br/>
	 * If an implementing class is found and it has a default constructor that
	 * is called and the new instance is returned.
	 * <br/><br/>
	 * 
	 * Raises no error if no class/object was found/created
	 * @param <T>
	 * @param <X>
	 * @param adapterClassOrInterface
	 * @param implementationQualifiedClassName
	 * @return
	 */
	public static <T, X extends T> T find(final Class<X> adapterClassOrInterface,final String implementationQualifiedClassName)
	{
		try
		{
			return find(adapterClassOrInterface, implementationQualifiedClassName, false);
		}
		catch (final Exception e)
		{
			return null;
		}
	}

	/**
	 * Finds a class that fits the implementation, either by the given
	 * implementationQualifiedClassName or if that name is null
	 * by trying to find a class factory that matches the given class name and package
	 * with an additional package prefix of "com.jsxp.framework.adapter.impl." and the class factory
	 * name suffix "Factory". If no factory is found it tries to find an adapter class same way
	 * as with the factory without the suffix. If still no adapter was found and
	 * the desired class is not an interface it tries to create an instane of that class
	 * as a default implementation.
	 * <br/><br/>
	 * If an implementing class is found and it has a default constructor that
	 * is called and the new instance is returned.
	 * <br/><br/>
	 * 
	 * @param <T>
	 * @param <X>
	 * @param clazz
	 * @param implementationQualifiedClassName
	 * @param raiseError
	 * @return
	 * @throws Exception
	 */
	@SuppressWarnings("unchecked")
	public static <T, X extends T> T find(final Class<X> adapterClassOrInterface, String implementationQualifiedClassName, final boolean raiseError) throws Exception
	{
		if (implementationQualifiedClassName == null)
		{
			// try to use a adapter factory
			try
			{
				final String factoryImplementationQualifiedClassName =
						"com.jsxp.framework.adapter.impl." + adapterClassOrInterface.getName() + "Factory";
				Log.debug(AdapterFinder.class, "looking for adapter factory: "+factoryImplementationQualifiedClassName);
				final Class<X> clazz = (Class<X>) Class.forName(factoryImplementationQualifiedClassName);
				final AdapterFactory factory = (AdapterFactory) clazz.newInstance();
				final T implementationObject = (T)factory.createInstance();
				Log.debug(AdapterFinder.class, "using adapter: "+implementationObject);
				return implementationObject;
			}
			catch (final ClassNotFoundException e)
			{
				Log.debug(AdapterFinder.class, "did not found an adapter factory "+e.getMessage());
			}
			catch (final Exception e)
			{
				Log.debug(AdapterFinder.class, "exception occurred while using adapter factory");
				Log.errorAndRaise(AdapterFinder.class, "could not create adapter factory: ", e, raiseError, implementationQualifiedClassName);
				return null;
			}

			// try to use a default implementation
			implementationQualifiedClassName = "com.jsxp.framework.adapter.impl." + adapterClassOrInterface.getName();
		}

		try
		{
			final Class<X> clazz = (Class<X>) Class.forName(implementationQualifiedClassName);
			final T implementationObject = clazz.newInstance();
			Log.debug(AdapterFinder.class, "using adapter: "+implementationObject);
			return implementationObject;
		}
		catch (final ClassNotFoundException e){ }
		catch (final Exception e)
		{
			Log.errorAndRaise(AdapterFinder.class, "could not create adapter: ", e, raiseError, implementationQualifiedClassName);
			return null;
		}

		if (!adapterClassOrInterface.isInterface())
		{
			try {
				final T implementationObject = adapterClassOrInterface.newInstance();
				Log.debug(AdapterFinder.class, "using adapter: "+implementationObject);
				return implementationObject;
			} catch (final Exception e)
			{
				Log.errorAndRaise(AdapterFinder.class, "could not create adapter: ", e, raiseError, implementationQualifiedClassName);
				return null;
			}
		}

		return null;
	}
}
