/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.util.object;

import java.io.UnsupportedEncodingException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Map;
import java.util.Map.Entry;

import org.jsxp.framework.logging.Log;
import org.jsxp.framework.util.text.StringUtils;


/**
 * Class to get values from properties from bean objects (simple java objects) and maps.
 * Support an hierarchical approach.
 * 
 * @author Oliver Szymanski
 *
 */
public class BeanUtils
{
	/**
	 * 
	 * @param property the property name
	 * @return the method name corresponding property (with a get prefix)
	 */
	private static String getGetterMethodName(final String property) {
		if (property == null || property.length() == 0)
		{
			return null;
		}

		final StringBuffer methodName = new StringBuffer("get");
		methodName.append(property.toUpperCase().substring(0,1));
		if (property.length() > 1)
		{
			methodName.append(property.substring(1));
		}

		return methodName.toString();
	}

	/**
	 * Gets the java property name.
	 * 
	 * @param id the id
	 * 
	 * @return the java property name
	 */
	public static String getJavaPropertyName(String id) {
		id = StringUtils.toFirstUpperCase(id);
		id = id.replace("-", "_");
		id = id.replace(".", "_");
		return id;
	}

	/**
	 * Like the method {@link #getKeyPathValue(String, Object)} but this will
	 * start to find a value from the last keyValues object to the first parameter
	 * until one is found (or no one is found)
	 * @param keyExpression the property/key expression to get a value for
	 * @param keyValues the objects that holds the value
	 * @return the value if found one or null
	 */
	public static Object getKeyPathValue(final String keyExpression, final Object... keyValues)
	{
		for (int i = keyValues.length-1; i >= 0; i--) {
			final Object result = getKeyPathValue(keyExpression, keyValues[i]);
			if (result != null) { return result; }
		}

		return null;
	}

	/**
	 * Will interpret the key as a combination of key part (separated by '.').
	 * Than it uses the method {@link #getValue(String, Object)} to get a
	 * value for each part starting at the beginning and with the last
	 * value as new keyValue object.
	 * @param keyExpression the property/key expression to get a value for
	 * @param keyValue the object that holds the value
	 * @return the value if found one or null
	 */
	public static Object getKeyPathValue(String keyExpression, final Object keyValue)
	{
		if (keyExpression.equals("."))
		{
			return keyValue;
		}
		else if (keyExpression.startsWith("."))
		{
			keyExpression = keyExpression.substring(1);
		}

		final String[] keySplit = keyExpression.split("\\.");
		if (keySplit.length > 0)
		{
			final Object value = getValue(keySplit[0], keyValue);
			if (keySplit.length==1)
			{
				return value;
			}
			else
			{
				return getKeyPathValue(keyExpression.substring(keySplit[0].length()+1), value);
			}
		}
		else
		{
			return null;
		}
	}

	/**
	 * 
	 * @param property the property name
	 * @return the method name corresponding property (without a get prefix)
	 */
	private static String getMethodName(final String property) {
		return StringUtils.toFirstLowerCase(property);
	}

	/**
	 * tries to get a value for a property (key) from the keyValue object.
	 * If keyValue is a map it looks directly for the key.
	 * If not it tries to find a corresponding method (with and without get prefix) for the property,
	 * calls this method and returns the result.
	 * @param key the property/key to get a value for
	 * @param keyValue the object that holds the value
	 * @return the value if found one or null
	 */
	@SuppressWarnings("unchecked")
	public static Object getValue(final String key, final Object keyValue)
	{
		if (keyValue instanceof Map<?, ?> && ((Map<String, Object>)keyValue).containsKey(key))
		{
			return ((Map<String, Object>)keyValue).get(key);
		}
		else
		{
			Method getterMethod;
			try
			{
				getterMethod = keyValue.getClass().getMethod(getGetterMethodName(key));
				return getterMethod.invoke(keyValue);
			}
			catch (final Exception e1)
			{
				try
				{
					getterMethod = keyValue.getClass().getMethod(getMethodName(key));
					return getterMethod.invoke(keyValue);
				}
				catch (final Exception e2) {
				}
			}

			return null;
		}
	}

	public static void setProperties(final String keyWithPrefix, final Map<String, Object[]> parameter, final Object beanToFill)
			throws IllegalAccessException, InvocationTargetException, UnsupportedEncodingException
			{
		if (beanToFill == null || parameter == null || parameter.isEmpty()) {
			return;
		}

		for (final Entry<String, Object[]> entry : parameter.entrySet()) {
			String key = entry.getKey();
			if (keyWithPrefix == null
					|| key.startsWith(keyWithPrefix)) {
				final Object[] values = entry.getValue();

				if (keyWithPrefix != null) {
					key = key.substring(keyWithPrefix.length());
				}

				boolean methodCalled = false;
				final Method[] methods = beanToFill.getClass().getMethods();
				for (final Method method : methods) {
					if (method.getName().equals("set" + getJavaPropertyName(key)) &&
							method.getParameterTypes().length == 1
							) {
						try {
							if (values.length == 1) {
								if (method.getParameterTypes()[0].isArray()) {
									method.invoke(beanToFill, new Object[] { values });
								} else {
									method.invoke(beanToFill, new Object[] { values[0] });
								}
							} else if (values.length > 1) {
								if (method.getParameterTypes()[0].isArray()) {
									method.invoke(beanToFill, new Object[] { values });
								} else {
									Log.error(BeanUtils.class, "-> input parameter '"+key+" was an array but property in parameter class '"+beanToFill.getClass()+"' was not an array");
								}
							}
							methodCalled = true;
							break;
						} catch (final IllegalArgumentException e) {
							Log.error(BeanUtils.class, "-> input parameter '"+key+" could not be assigned to property in parameter class '"+beanToFill.getClass()+"' "+e.getMessage());
						}
					}
				}

				if (!methodCalled) {
					Log.error(BeanUtils.class, "-> input parameter '"+key+": no matching property found in parameter class '"+beanToFill.getClass()+"'.");
				}
			}
		}
			}

	/**
	 * Constructor
	 */
	private BeanUtils() { }
}
