/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.util.text;

import java.lang.reflect.Method;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.jsxp.framework.Context;
import org.jsxp.framework.internal.InternalElement;


/**
 * The Class StringUtil. Utility class for strings.
 */
public class StringUtil {

	/**
	 * Capitalize first letter.
	 * 
	 * @param str the str
	 * 
	 * @return the string
	 */
	public static String capitalizeFirstLetter(final String str) {
		if (str == null || str.length()==0) { return str; }
		else if (str.length() == 1) { return str.toUpperCase(); }
		else {
			final StringBuffer result = new StringBuffer(str.length());
			result.append(str.substring(0,1).toUpperCase());
			result.append(str.substring(1));
			return result.toString();
		}
	}

	/**
	 * Compare.
	 * 
	 * @param a the a
	 * @param b the b
	 * 
	 * @return true, if successful
	 */
	public static boolean compare(final String a, final String b) {
		if (a == null) {
			return (a == b);
		} else {
			return a.equals(b);
		}
	}

	/**
	 * Concate strings.
	 * 
	 * @param delimiter the delimiter
	 * @param args the args
	 * 
	 * @return the string
	 */
	public static String concateStrings(final Character delimiter, final Collection<String> args) {
		final StringBuilder result = new StringBuilder();
		final boolean appendDelimiter = delimiter!=null;
		for (final String s : args) {
			if (s != null) {
				result.append(s);
				if (appendDelimiter) {
					result.append(delimiter);
				}
			}
		}
		return result.toString();
	}

	/**
	 * Concate strings.
	 * 
	 * @param delimiter the delimiter
	 * @param args the args
	 * 
	 * @return the string
	 */
	public static String concateStrings(final Character delimiter, final String ... args) {
		final StringBuilder result = new StringBuilder();
		final boolean appendDelimiter = delimiter!=null;
		for (final String s : args) {
			if (s != null) {
				result.append(s);
				if (appendDelimiter) {
					result.append(delimiter);
				}
			}
		}
		return result.toString();
	}

	/**
	 * Delete suffix.
	 * 
	 * @param str the str
	 * 
	 * @return the string
	 */
	public static String deleteSuffix(final String str) {
		if (str == null || str.length()==0) {
			return str;
		}

		final int i = str.lastIndexOf('.');
		if (i > 0) {
			return str.substring(0, i);
		}

		return str;
	}

	/**
	 * Gets the java class name.
	 * 
	 * @param javaName the java name
	 * 
	 * @return the java class name
	 */
	public static String getJavaClassName(String javaName) {
		javaName = capitalizeFirstLetter(javaName);
		//javaName = StringUtil.deleteSuffix(javaName);

		javaName = replaceOccurrences(javaName, "-");
		javaName = replaceOccurrences(javaName, "_");
		javaName = replaceOccurrences(javaName, ".");
		javaName = replaceOccurrences(javaName, " ");

		return javaName;
	}

	/**
	 * Gets the java package dir.
	 * 
	 * @param javaName the java name
	 * 
	 * @return the java package dir
	 */
	public static String getJavaPackageDir(String javaName) {
		javaName = lowerFirstLetter(javaName);
		javaName = replaceOccurrences(javaName, "-");
		javaName = replaceOccurrences(javaName, " ");
		javaName = javaName.replace('.', '/');
		javaName = javaName.replace('\\', '/');
		return javaName;
	}

	/**
	 * Gets the java package name.
	 * 
	 * @param javaName the java name
	 * 
	 * @return the java package name
	 */
	public static String getJavaPackageName(String javaName) {
		javaName = lowerFirstLetter(javaName);
		javaName = replaceOccurrences(javaName, "-");
		javaName = replaceOccurrences(javaName, " ");
		javaName = javaName.replace('/', '.');
		javaName = javaName.replace('\\', '.');
		return javaName;
	}

	/**
	 * Gets the java property name.
	 * 
	 * @param id the id
	 * 
	 * @return the java property name
	 */
	public static String getJavaPropertyName(final String id) {
		String idCapitalLetter = id.substring(0,1).toUpperCase()+id.substring(1);

		idCapitalLetter = replaceOccurrences(idCapitalLetter, "-");
		idCapitalLetter = replaceOccurrences(idCapitalLetter, "_");
		idCapitalLetter = replaceOccurrences(idCapitalLetter, ".");

		return idCapitalLetter;
	}

	/**
	 * Gets the property value.
	 * 
	 * @param parameterBean the parameter bean
	 * @param propertyName the property name
	 * 
	 * @return the property value
	 */
	public static String getPropertyValue(final Object parameterBean, final String propertyName) {
		final String getterName = "get"+getJavaPropertyName(propertyName);

		try {
			final Method getMethod = parameterBean.getClass().getMethod(getterName);
			final String value = getMethod.invoke(parameterBean).toString();
			return value;
		} catch (final Exception e) {
			Context.getContext().addDeveloperWarningMessage("Could not access property \""+propertyName+" in bean "+parameterBean+": "+e.getMessage());
		}
		return null;
	}

	/**
	 * Checks if is empty.
	 * 
	 * @param s the s
	 * 
	 * @return true, if is empty
	 */
	public static boolean isEmpty(final String s) {
		return s == null || "".equals(s);
	}

	/**
	 * Lower first letter.
	 * 
	 * @param str the str
	 * 
	 * @return the string
	 */
	public static String lowerFirstLetter(final String str) {
		if (str == null || str.length()==0) { return str; }
		else if (str.length() == 1) { return str.toLowerCase(); }
		else {
			final StringBuffer result = new StringBuffer(str.length());
			result.append(str.substring(0,1).toLowerCase());
			result.append(str.substring(1));
			return result.toString();
		}
	}

	/**
	 * Replace occurrences.
	 * 
	 * @param name the name
	 * @param toReplace the to replace
	 * 
	 * @return the string
	 */
	public static String replaceOccurrences(final String name, final String toReplace) {
		String retVal = name;
		while(retVal.indexOf(toReplace) != -1) {
			final int index = retVal.indexOf(toReplace);
			final String before = retVal.substring(0, index);
			final String upperCaseChar = retVal.substring(index+1, index+2).toUpperCase();
			final String after = retVal.substring(index+2);
			retVal = before+upperCaseChar+after;
		}
		return retVal;
	}

	/**
	 * Replace prefix in list.
	 * 
	 * @param list the list
	 * @param toReplace the to replace
	 * @param replaceWith the replace with
	 */
	public static void replacePrefixInList(final List<String> list, final String toReplace, final String replaceWith) {
		if (list == null || list.isEmpty()) {
			return;
		}

		for (int i = 0; i<list.size(); i++) {
			final String str = list.get(i);
			if (str.indexOf(toReplace)==0) {
				list.set(i, str.replaceFirst(toReplace, replaceWith));
			}
		}
	}

	/**
	 * Replace variables.
	 * 
	 * @param originalText the original text
	 * @param element the element
	 * 
	 * @return the string
	 */
	public static String replaceVariables(final String originalText, final InternalElement element) {
		if(originalText==null) {
			return null;
		}
		int previousIndex = 0;
		int index = originalText.indexOf("$(");
		final StringBuilder newText = new StringBuilder();

		while(index != -1 && index >= previousIndex) {
			if(index == 0 || originalText.charAt(index-1)!='$') {
				final int endIndex = originalText.indexOf(")", index);
				if(endIndex > index) {
					final String propertyName = originalText.substring(index+2, endIndex);
					final String before = originalText.substring(previousIndex, index);
					final String value = element.getVariableValue(propertyName);
					if(!StringUtil.isEmpty(value)) {
						newText.append(before).append(value);
						previousIndex = endIndex+1;
					} else {
						final String orig = originalText.substring(previousIndex, index);
						newText.append(orig);
						previousIndex=endIndex+1;
					}
				} else {
					Context.getContext().addDeveloperWarningMessage("Template substitution: \"$(\" found, but there is no corresponding \")\"");
					final String orig = originalText.substring(previousIndex, index+1);
					newText.append(orig);
					previousIndex = index+1;
				}
			} else {
				if(index > 0) {
					final String orig = originalText.substring(previousIndex, index);
					newText.append(orig);
				}
				previousIndex = index+1;
			}
			index = originalText.indexOf("$(", previousIndex);
		}
		if(previousIndex == -1) {
			return originalText;
		} else {
			newText.append(originalText.substring(previousIndex));
		}
		return newText.toString();
	}

	/**
	 * Split.
	 * 
	 * @param toSplit the to split
	 * @param delimiter the delimiter
	 * 
	 * @return the string[]
	 */
	public static String[] split(final String toSplit, final char delimiter) {
		final List<String> list = new LinkedList<String>();
		int index = toSplit.indexOf(delimiter, 0);
		int indexStart = 0;
		while (index > 0) {
			if (toSplit.charAt(index-1) != '\\') {
				String str = toSplit.substring(indexStart, index);
				str = str.replaceAll("[\\\\][|]", "|");
				list.add(str);
				indexStart = ++index;
				index = toSplit.indexOf(delimiter, indexStart);
			} else {
				index++;
				index = toSplit.indexOf(delimiter, index);
			}
		}
		String str = toSplit.substring(indexStart);
		str = str.replaceAll("[\\\\][|]", "|");
		list.add(str);
		return list.toArray(new String[] {});
	}
}
