/*   Copyright 2012 Oliver Szymanski & David Tanzer

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
package org.jsxp.framework.util.text;

/**
 * Some needful string utilities.
 * 
 * @author Oliver Szymanski
 * @author Robert Mencl
 *
 */
public class StringUtils
{
	/**
	 * Empty String
	 */
	public final static String EMPTY_STRING = "";

	/**
	 * Formats an object array as a string
	 * @param array the objects whos {@link Object#toString()}
	 * @return the {@link String}
	 */
	public static String arrayAsString(final Object[] array)
	{
		if (array == null) { return String.valueOf(array); }

		final StringBuffer sb = new StringBuffer("[");
		for (int i = 0; i < array.length; i++)
		{
			if (array[i] instanceof Object[])
			{
				sb.append(arrayAsString(array));
			}
			else
			{
				if (i > 0) { sb.append(", "); }
				sb.append(String.valueOf(array[i]));
			}
		}

		sb.append("]");
		return sb.toString();
	}

	/**
	 * Combines string representation of objects
	 * @param separator is used between the objects to combine in the combined result
	 * @param avoidDuplicateSeparator if true then separator is only used between part where a separator is not already there at the end of the previous and the beginning of the next part
	 * @param str the objects who's {@link Object#toString()} return values will be combined
	 * @return the combined {@link String}
	 */
	public static String combine(final String separator, final boolean avoidDuplicateSeparator, final Object...str)
	{
		final StringBuffer sb = new StringBuffer();
		for (int i = 0; i < str.length; i++)
		{
			final String currentPart = String.valueOf(str[i]);
			sb.append(currentPart);

			if (separator != null && i != str.length-1 && currentPart != null && currentPart.length() > 0)
			{
				if (!avoidDuplicateSeparator)
				{
					sb.append(separator);
				}
				else
				{
					final String nextPart = String.valueOf(str[i+1]);
					if (!currentPart.endsWith(separator) && !nextPart.startsWith(separator))
					{
						sb.append(separator);
					}
				}
			}
		}
		return sb.toString();
	}

	/**
	 * Combines string representation of objects
	 * @param str the objects whos {@link Object#toString()} return values will be combined
	 * @return the combined {@link String}
	 */
	public static String combineAll(final Object...str)
	{
		if (str == null) { return String.valueOf(str); }

		final StringBuffer sb = new StringBuffer();
		for (int i = 0; i < str.length; i++)
		{
			sb.append(String.valueOf(str[i]));
		}
		return sb.toString();
	}

	/**
	 * Combines string representation of objects
	 * @param str the objects whos {@link Object#toString()} return values will be combined
	 * @return the combined {@link String}
	 */
	public static String combineAll(final String[] str)
	{
		return combineAll((Object[])str);
	}

	/**
	 * Counts the occurrence of a char in a {@link String}
	 * @param content
	 * @param countChar
	 * @return
	 */
	public static int countChar(final String content, final char countChar)
	{
		int start;
		if ((start = content.indexOf(countChar)) < 0) { return 0; }

		int count = 0;
		for(int i = start; i < content.length(); i++)
		{
			if (content.charAt(i) == countChar) { count++; }
		}

		return count;
	}

	/**
	 * Returns the substring till the first occurrence of a search string
	 * in a source string. returns the source string, if no occurrence is there.
	 * returns null if the source string is null.
	 * @param source
	 * @param search
	 * @return
	 */
	public static String substringTillFirstOccurence(final String source, final String search)
	{
		if (source == null || search == null)
		{
			return source;
		}

		final int pos = source.indexOf(search);

		if (pos <= 0) {
			return source;
		}

		return source.substring(0, pos);
	}

	/**
	 * Returns the substring till the last occurrence of a search string
	 * in a source string. returns the source string, if no occurrence is there.
	 * returns null if the source string is null.
	 * @param source
	 * @param search
	 * @return
	 */
	public static String substringTillLastOccurence(final String source, final String search)
	{
		if (source == null || search == null)
		{
			return source;
		}

		final int pos = source.lastIndexOf(search);

		if (pos <= 0) {
			return source;
		}

		return source.substring(0, pos);
	}

	/**
	 * 
	 * 
	 * @param s the string to process
	 * @return Returns the same string but with the first character as lower case character.
	 */
	public static String toFirstLowerCase(final String s)
	{
		if (null == s || s.length()==0) { return s; }

		final char ch = s.charAt(0);

		if (! Character.isLowerCase(ch))
		{
			final char lower = Character.toLowerCase(ch);
			final String mod = s.substring(1);
			return "" + lower + mod;
		}

		return s;
	}

	/**
	 * 
	 * 
	 * @param s the string to process
	 * @return Returns the same string but with the first character as upper case character.
	 */
	public static String toFirstUpperCase(final String s)
	{
		if (null == s || s.length()==0) { return s; }

		final char ch = s.charAt(0);

		if (! Character.isUpperCase(ch))
		{
			final char upper = Character.toUpperCase(ch);
			final String mod = s.substring(1);
			return "" + upper + mod;
		}

		return s;
	}

	/**
	 * Private constructor because this class is just a collection of static methods
	 */
	private StringUtils() { }


}
